/**
 * @file Debug.h
 *
 * Freddy's debug declarations
 * 
 * @author Moricz, Ferenc
 * @date november 4., 2001.
 *
 */

#ifndef INCLUDE_NSDEBUG
#define INCLUDE_NSDEBUG
#include "Debug.h"

#ifndef DEBUGLEVEL
#define DEBUGLEVEL 3
#endif

#include <qstring.h>

#include <vector>
using std::vector;

namespace nsdebug {
	class Rule
	{
	  public:
		inline Rule(void) { sClass = ""; iLevel = DEBUGLEVEL; iLevel = 10; }
		inline Rule(char *sRule, int iLevel = -1) { sClass = sRule; iLevel = iLevel; }
		QString sClass;
		int iLevel;
	};

	class Class
	{
	  friend class Debug;
	  public:
		Class(const char *sName, Class *pParent = NULL);
		~Class();

		void setLevelOfSubTree(int iLevel = 10);
		int iLevel;

		inline QString *getFullName(void) { return &sFullName; }

	  protected:
		QString sName, sFullName;
		int iDepth;

		/* the pointers which creates the tree-structure */
		Class *pParent;
		vector<Class *> oVectChilds;
	};

	class Debug
	{
	  public:
		Debug();
		Debug::~Debug();

		static bool isAllowed(Class &oItem);

		Class *getClass(const QString &sName, int *pLevel = NULL, bool bCreate = false);

		void Debug::printRules(void);
		void printTree(Class *pRoot);

		//Class *insertItem(char *sName);

		/** Adds a new rule to the debug rule list
		  * @param sRule the rule to be added
		  * You can append new rules to the accept and deny rule list.
		  * To add a accept rule prepend "+" or "" to the string.
		  * To add a remove rule prepend a "-" to the string.
		  */
		void addRule(const char *sRule, int iLevel = -1);
		void removeRule(int iNumber);
		void applyRules(void);

		inline bool isPrintable(Class *pClass, int iLevel)
		{ if (iLevel <= pClass->iLevel) return true;
			else { iMissed ++; return false; } }
		void error(Class *pClass, int iLevel, char *pFormat, ...);
		void warning(Class *pClass, int iLevel, char *pFormat, ...);
		void message(Class *pClass, int iLevel, char *pFormat, ...);
		void print(Class *pClass, int iLevel, char *pFormat, ...);

		static Debug *pMain;

	  protected:
		int iMissed;
		int iDebugLevel;
		vector<Rule> oVectRules;
		Class oTreeRoot;
	};
};


/*
 * System type detection
 *
 * Result: any of the following symbols will be defined:
 *    _SYSTEM_WINDOWS_
 *    _SYSTEM_LINUX_
 */

#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#define _SYSTEM_WINDOWS_
#else
#define _SYSTEM_LINUX_
#define _SYSTEM_VXWORKS_
#endif

/*
 * OS specific settings
 */
#if defined(_SYSTEM_WINDOWS_)
#pragma warning( disable: 4660 )
#pragma warning( disable: 4786 )
#endif

/* 
 * Debug macros
 *
 * LINUX version 
 */


#if defined(_SYSTEM_LINUX_) || defined (_SYSTEM_VXWORKS_)

#include <stdio.h>

#if DEBUGLEVEL>=4
#define debug_begin(class) debug_print("4:" class, "<<< %s [\n", __PRETTY_FUNCTION__)
#define debug_end(class) debug_print("4:" class, ">>> %s() ]\n", __PRETTY_FUNCTION__)
#else
#define debug_begin()
#define debug_end()
#endif


#define debug_add_rule(rule) nsdebug::Debug::pMain->addRule(rule)
#define debug_print_rules() nsdebug::Debug::pMain->printRules()
#define debug_print_tree() nsdebug::Debug::pMain->printTree(NULL)

/*
 * LINUX - ERROR
 */
#if DEBUGLEVEL>=1
#define debug_error(sClass, sFormat, args...) \
	{ \
		static int iLevel = 1; \
		static nsdebug::Class *pDebugClass = nsdebug::Debug::pMain->getClass(sClass, &iLevel, true); \
		nsdebug::Debug::pMain->error(pDebugClass, iLevel, sFormat, ##args); \
	}
#else
#define debug_error(sClass, sFormat, args...)
#endif

/*
 * LINUX - WARNING
 */
#if DEBUGLEVEL>=2
#define debug_warning(sClass, sFormat, args...) \
	{ \
		static int iLevel = 2; \
		static nsdebug::Class *pDebugClass = nsdebug::Debug::pMain->getClass(sClass, &iLevel, true); \
		nsdebug::Debug::pMain->warning(pDebugClass, iLevel, sFormat, ##args); \
	}
#else
#define debug_warning(sClass, sFormat, args...)
#endif



/*
 * LINUX - MESSAGE
 */
#if DEBUGLEVEL>=3 || defined(DEBUG_DYNAMIC)
#define debug_message(sClass, sFormat, args...) \
	{ \
		static int iLevel = 3; \
		static nsdebug::Class *pDebugClass = nsdebug::Debug::pMain->getClass(sClass, &iLevel, true); \
		if (nsdebug::Debug::pMain->isPrintable(pDebugClass, iLevel)) \
			nsdebug::Debug::pMain->message(pDebugClass, iLevel, sFormat, ##args); \
	}

/*
 * LINUX - PRINT
 */
#define debug_print(sClass, sFormat, args...) \
	{ \
		static int iLevel = 3; \
		static nsdebug::Class *pDebugClass = nsdebug::Debug::pMain->getClass(sClass, &iLevel, true); \
		if (nsdebug::Debug::pMain->isPrintable(pDebugClass, iLevel)) \
			nsdebug::Debug::pMain->print(pDebugClass, iLevel, sFormat, ##args); \
	}
#else
#define debug_message(text, args...)
#define debug_print(text, args...)
#endif

/** Short names for the debug functions */
#if !defined(dmesg) && !defined(dwarn) && !defined(derr)
#define dbeg debug_begin
#define dend debug_end
#define derr debug_error
#define dwarn debug_warning
#define dmesg debug_message
#define dprint debug_print
#endif

#if !defined(DEBUG_MESSAGE) && !defined(DEBUG_WARNING) && !defined(DEBUG_ERROR)
#define DEBUG_BEGIN debug_begin
#define DEBUG_END debug_end
#define DEBUG_ERROR(format, args...) debug_error(NULL, format, ##args)
#define DEBUG_WARNING(format, args...) debug_warning(NULL, format, ##args)
#define DEBUG_MESSAGE(format, args...) debug_message(NULL, format, ##args)
#define DEBUG_PRINT(format, args...) debug_print(NULL, format, ##args)
#endif

/* 
 * Debug macros
 *
 * WINDOWS version 
 */
#elif defined(_SYSTEM_WINDOWS_)
#undef DEBUGLEVEL
#define DEBUGLEVEL 5

#define dbegin()
#define dbeg()
#define dend()

#if DEBUGLEVEL>=3
#define dmesg fflush(stdout), printf("\n"), printf
#define dprint fflush(stdout), printf
#else 
#define dmesg
#define dprint
#endif

#if DEBUGLEVEL>=2
#define dwarn fflush(stdout), printf("\n"), printf
#define derr fflush(stdout), printf("\n"), printf
#else
#define dwarn
#define derr
#endif

#if DEBUGLEVEL>=1
#define DEBUG_ERROR fflush(stderr), printf("\n"), printf
#else
#define DEBUG_ERROR
#endif

#define debug_add_rule()
#define debug_print_rules()
#define debug_print_tree()

#undef DEBUG_ERROR
#undef DEBUG_WARNING
#undef DEBUG_MESSAGE
#undef DEBUG_PRINT
#define DEBUG_ERROR derr
#define DEBUG_WARNING dwarn
#define DEBUG_MESSAGE dmesg
#define DEBUG_PRINT dprint

#endif

#endif  // INCLUDE_NSDEBUG

