/**
 * @file UndoRedo.h
 *
 * @brief Undo/Redo function, declarations
 * 
 * @author Moricz, Ferenc
 * $Revision: 1.13 $
 * $Date: 2004/02/22 00:12:48 $
 *
 */

#ifndef FREDDY_UNDOREDO
#define FREDDY_UNDOREDO

class EditBuffer;

#include <qwidget.h>

#include <vector>
using namespace std;

/*
 * This class represents an atomic text operation
 * Atomic operations:
 *    - insertstring [reversible] - inserts a string to a line
 *    - deletestring [reversible] - deletes a string from a line
 *    - createline [reversible] - creates a new, empty line
 *    - deleteline [reversible] - deletes a line
 *    - movecursor [reversible] - moves the cursor to a new position
 *    - scroll [not reversible] - scrolls the buffer to a new position
 *    - setselection [not reversible] - changes the current selection
 */
class AtomicTextOperation
{                              
  public:
	enum OperationCode { INSERTSTRING, DELETESTRING,
		CREATELINE, DELETELINE, MOVECURSOR, SCROLL, SETSELECTION};

	AtomicTextOperation(OperationCode eCode, QString sText="", int iColumn=-1, int iLine=-1, int iAttrib1=-1, int iAttrib2=-1);
	OperationCode getOperationCode(void) { return eOperationCode; }

	void execDo(EditBuffer *pBuffer);
	void execUndo(EditBuffer *pBuffer);

  protected:
	int iLine, iColumn;            /* the line and column where the operation is done */
	OperationCode eOperationCode;  /* the type of the operation */
	QString sText;                 /* the text which is changed with the operation */
	int iAttrib1, iAttrib2;        /* options */
};


class TextOperation
{
  public:
	TextOperation() { eStatus = TextOperation::TEXT_UNKNOWN; }

	QString *sOperationName;                   /* the name of the operation (eg. TYPE) */
	void addAtomicOperation(AtomicTextOperation *pOperation);
	vector<AtomicTextOperation *> ovAtomicOps; /* the list of the atomic operations */

	enum RESULT_EFFECT { TEXT_CHANGED, TEXT_UNCHANGED, TEXT_UNKNOWN };
	RESULT_EFFECT eStatus;
};


class UndoRedo
{
  public:
	UndoRedo();

	void createOperation(const QString &sName);
	void continueOperation(void);
	void closeOperation(void);
	void doAtomicOperation(AtomicTextOperation::OperationCode, QString ="", int iColumn=-1, int iLine=-1, int iAttrib1=-1, int iAttrib2=-1);

	inline void undoLastOperation(bool bMicroUndo = false) { execute_UndoRedoOperation(true, bMicroUndo); };
	inline void redoLastOperation(bool bMicroUndo = false) { execute_UndoRedoOperation(false, bMicroUndo); };

	void setSavedPosition(void);
	bool isAtSavedPosition(void);
	bool isChanged(void);

	EditBuffer *pBuffer;
  protected:
	vector<TextOperation *> ovPrevOperationLst, ovNextOperationLst;
	bool bOperationOpened;
	TextOperation *pSavedPosition;

	void stripOpName(QString &sOpName);
	void execute_UndoRedoOperation(bool bUndo = true, bool bMicro = false);
};

#endif    /* FREDDY_UNDOREDO */

