/**
 * @file console.cpp
 * @brief console module - implementation
 *
 * @full Console window module for Freddy
 * 
 * @author Ferenc Moricz
 * Created:   2001/12/21
 * $Revision: 1.6 $
 * $Date: 2003/10/12 01:30:24 $
 *
 */

#include "console.h"
#include "console_icons.h"
#include "Editor.h"

#include <errno.h>
#include <fcntl.h>

#include <qhbox.h>
#include <qmessagebox.h>
#include <qmultilineedit.h>

static ConsoleWindow *console;
FILE *pipe_fp;
QTimer *console_timer;
console_feeder *feeder = NULL;
Module *pMod;

int handle_setnonblocking(int handle, int set)
{
#ifdef _SYSTEM_LINUX_
	int ioret, flags;

#if defined(O_NONBLOCK)
	if (-1 == (flags = fcntl(handle, F_GETFL, 0))) flags = 0;
	if (set) flags |= O_NONBLOCK; else flags &= ~O_NONBLOCK;
	ioret = fcntl(handle, F_SETFL, flags);
#else
	flags = set;
	ioret = ioctl(handle, FIOBIO, &flags);
#endif

	return (ioret != -1);
#else
	return 0;
#endif
}


Module *module_init(Module *pModule)
{
	pModule->qsName = new QString("console");
	return pMod = pModule;
}

#ifdef _SYSTEM_WINDOWS_
#include <io.h>
#include <stdlib.h>
#include <stdio.h> 
#include <process.h>
#include <windows.h>

#define BUFSIZE 4096 
 
HANDLE hChildStdinRd, hChildStdinWr, hChildStdinWrDup, 
   hChildStdoutRd, hChildStdoutWr, hChildStdoutRdDup, 
   hInputFile, hSaveStdin, hSaveStdout; 
 
BOOL CreateChildProcess(VOID); 
VOID WriteToPipe(VOID); 
VOID ReadFromPipe(VOID); 
VOID ErrorExit(LPTSTR); 
VOID ErrMsg(LPTSTR, BOOL); 
#endif

QString *module_function(Module *pModule, const QString &qsFunction,
            QString *qsParam1, QString *qsParam2, int iParam1, int iParam2)
{
	if (qsFunction == "3")
	{
		if ( console != NULL )
		{
			console->close( true );
			console = NULL;
		} else
		{
			QFrame *container = (QFrame *)pMod->pQuery("cb.get.editor.downcontainer", NULL, NULL, NULL);
			console = new ConsoleWindow(container);
			console->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Minimum));
			console->setMinimumHeight(20);
			console->setMaximumHeight(200);
			console->show();

			QString *sFFamily = (QString *)pMod->pQuery("cb.get.properties.string",
			         "module.console.font.family", NULL, NULL);
			int iFSize = (int)pMod->pQuery("cb.get.properties.number",
			         "module.console.font.size", (char *)10, NULL);
			QFont fnt=*new QFont( (sFFamily ? sFFamily->latin1() : "fixed"), iFSize, QFont::Normal );
			fnt.setFixedPitch( true );
			console->text()->setFont( fnt );
		}
	} else
	if (qsFunction == "1")
	{
		if (console == NULL) return NULL;

		if (feeder == NULL)
		{
			feeder = new console_feeder();
			feeder->running = false;
		}
             
		if (! feeder->running)      /*? and if it is running */
		{
			/* saving the current values of the input fields */
			QString sTemp = console->getsymbol->text();
			pMod->pQuery("cb.set.properties.string", "module.console.make.symbol", (const char *)&sTemp, NULL);
			sTemp = console->getpath->text();
			pMod->pQuery("cb.set.properties.string", "module.console.make.path", (const char *)&sTemp, NULL);

			console->iCompileState = 0;
			console->iCompileNumWarn = console->iCompileNumErr = 0;
			console->iSignSpinning = 2;
			console->displayCurrentPixmap();
			console->pSignTimer->start(200);
			console->updateToolTip();

			/* create the command's string */
			QString path( NULL );
#ifdef _SYSTEM_LINUX_
			path.append( "make " );
			path
				.append( console->getsymbol->text() );
			path.append( " -C " );
			path.append( console->getpath->text() );
			path.append( " 2>&1" );
			console->text()->clear();
			console->text()->append( path );
#else
			path.append( "dir" );
#endif

			/* create the process & set up the pipes */
#ifdef _SYSTEM_LINUX_
			if ((feeder->pipe_fp = popen(path.latin1(), "r")) == NULL)
			{
				perror("popen error");
				QApplication::exit(1);
				return NULL;
			}
#else
			   SECURITY_ATTRIBUTES saAttr; 
			   BOOL fSuccess; 
			 
			// Set the bInheritHandle flag so pipe handles are inherited. 
			 
			   saAttr.nLength = sizeof(SECURITY_ATTRIBUTES); 
			   saAttr.bInheritHandle = TRUE; 
			   saAttr.lpSecurityDescriptor = NULL; 
			 
			   // The steps for redirecting child process's STDOUT: 
			   //     1. Save current STDOUT, to be restored later. 
			   //     2. Create anonymous pipe to be STDOUT for child process. 
			   //     3. Set STDOUT of the parent process to be write handle to 
			   //        the pipe, so it is inherited by the child process. 
			   //     4. Create a noninheritable duplicate of the read handle and
			   //        close the inheritable read handle. 
			 
			// Save the handle to the current STDOUT. 
			 
			   hSaveStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
			 
			// Create a pipe for the child process's STDOUT. 
			 
			   if (! CreatePipe(&hChildStdoutRd, &hChildStdoutWr, &saAttr, 0)) 
			      ErrorExit("Stdout pipe creation failed\n"); 
			 
			// Set a write handle to the pipe to be STDOUT. 
			 
			   if (! SetStdHandle(STD_OUTPUT_HANDLE, hChildStdoutWr)) 
			      ErrorExit("Redirecting STDOUT failed"); 
			 
			// Create noninheritable read handle and close the inheritable read 
			// handle. 

			    fSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdoutRd,
			        GetCurrentProcess(), &hChildStdoutRdDup , 0,
			        FALSE,
			        DUPLICATE_SAME_ACCESS);
			    if( !fSuccess )
			        ErrorExit("DuplicateHandle failed");
			    CloseHandle(hChildStdoutRd);

			   // The steps for redirecting child process's STDIN: 
			   //     1.  Save current STDIN, to be restored later. 
			   //     2.  Create anonymous pipe to be STDIN for child process. 
			   //     3.  Set STDIN of the parent to be the read handle to the 
			   //         pipe, so it is inherited by the child process. 
			   //     4.  Create a noninheritable duplicate of the write handle, 
			   //         and close the inheritable write handle. 
			 
			// Save the handle to the current STDIN. 
			 
			   hSaveStdin = GetStdHandle(STD_INPUT_HANDLE); 
			 
			// Create a pipe for the child process's STDIN. 
			 
			   if (! CreatePipe(&hChildStdinRd, &hChildStdinWr, &saAttr, 0)) 
			      ErrorExit("Stdin pipe creation failed\n"); 
			 
			// Set a read handle to the pipe to be STDIN. 
			 
			   if (! SetStdHandle(STD_INPUT_HANDLE, hChildStdinRd)) 
			      ErrorExit("Redirecting Stdin failed"); 
			 
			// Duplicate the write handle to the pipe so it is not inherited. 
			 
			   fSuccess = DuplicateHandle(GetCurrentProcess(), hChildStdinWr, 
			      GetCurrentProcess(), &hChildStdinWrDup, 0, 
			      FALSE,                  // not inherited 
			      DUPLICATE_SAME_ACCESS); 
			   if (! fSuccess) 
			      ErrorExit("DuplicateHandle failed"); 
			 
			   CloseHandle(hChildStdinWr); 
			 
			// Now create the child process. 
			 
			   if (! CreateChildProcess()) 
			      ErrorExit("Create process failed"); 
			 
			// After process creation, restore the saved STDIN and STDOUT. 
			 
			   if (! SetStdHandle(STD_INPUT_HANDLE, hSaveStdin)) 
			      ErrorExit("Re-redirecting Stdin failed\n"); 
			 
			   if (! SetStdHandle(STD_OUTPUT_HANDLE, hSaveStdout)) 
			      ErrorExit("Re-redirecting Stdout failed\n"); 
			 
			// Get a handle to the parent's input file. 
			 
			   if (argc > 1) 
			      hInputFile = CreateFile(argv[1], GENERIC_READ, 0, NULL, 
			         OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL); 
			   else 
			      hInputFile = hSaveStdin; 
			 
			   if (hInputFile == INVALID_HANDLE_VALUE) 
			      ErrorExit("no input file\n"); 
			 
			// Write to pipe that is the standard input for a child process. 
			 
			   WriteToPipe(); 
			 
			// Read from pipe that is the standard output for child process. 
			 
			   ReadFromPipe(); 

#if 0
			HANDLE hProcess;
			int hStdOut;
			int hStdOutPipe[2];

			// Create the pipe
			if(_pipe(hStdOutPipe, 512, O_BINARY | O_NOINHERIT) == -1)
				return   NULL;

			// Duplicate stdout handle (next line will close original)
			hStdOut = _dup(_fileno(stdout));

			// Duplicate write end of pipe to stdout handle
			if(_dup2(hStdOutPipe[1], _fileno(stdout)) != 0)
			{
				if (errno == EBADF) printf("EBADF\n");
				else if (errno == EMFILE) printf("EMFILE\n");
				else printf("Other error\n");
				fflush(stdout);
				return   NULL;
			}

			// Close original write end of pipe
			close(hStdOutPipe[1]);

			// Spawn process
			hProcess = (HANDLE)spawnvp(P_NOWAIT, "make", (const char* const*)&"make");

			// Duplicate copy of original stdout back into stdout
			if(_dup2(hStdOut, _fileno(stdout)) != 0)
				return NULL;

			// Close duplicate copy of original stdout
			close(hStdOut);

			//read(

/*			if(hProcess)
			{
				int nOutRead;
				while   (nExitCode == STILL_ACTIVE)
				{
					nOutRead = read(hStdOutPipe[READ_HANDLE], szBuffer, OUT_BUFF_SIZE);
					if(nOutRead)
					{
						nOutRead = Filter(szBuffer, nOutRead, BEEP_CHAR);
						fwrite(szBuffer, 1, nOutRead, stdout);
					}

					if(!GetExitCodeProcess(hProcess,(unsigned long*)&nExitCode))
						return NULL;
				}
			}

			pipe_fp = _fdopen( hStdOutPipe[0], "r" );
			if (pipe_fp == NULL)
			{
				DEBUG_ERROR("_fdopen convert error!");
				return NULL;
			}
*/
#endif
#endif
			handle_setnonblocking( fileno(feeder->pipe_fp), 1);

			console->text()->dirs.clear();
			console->text()->dirs.append( new QString(console->getpath->text()) );
			console_timer = new QTimer( feeder, "consoletimer" );
			QObject::connect( console_timer, SIGNAL(timeout( void )),
					                feeder, SLOT( pthread_slot( void ) ) );
			QObject::connect( console->text(), SIGNAL(doubleClicked(int, int)),
					          console->text(), SLOT( doubleClickEvent(int, int)) );
			console_timer->start( 100 );
			console->text()->errorline = -1;
			feeder->running = true;
		}
	}
	return NULL;
}

#if defined(_SYSTEM_WINDOWS_)
BOOL CreateChildProcess() 
{ 
   PROCESS_INFORMATION piProcInfo; 
   STARTUPINFO siStartInfo; 
 
// Set up members of the PROCESS_INFORMATION structure. 
 
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );
 
// Set up members of the STARTUPINFO structure. 
 
   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO); 
 
// Create the child process. 
 
   return CreateProcess(NULL, 
      "child",       // command line 
      NULL,          // process security attributes 
      NULL,          // primary thread security attributes 
      TRUE,          // handles are inherited 
      0,             // creation flags 
      NULL,          // use parent's environment 
      NULL,          // use parent's current directory 
      &siStartInfo,  // STARTUPINFO pointer 
      &piProcInfo);  // receives PROCESS_INFORMATION 
}
 
VOID WriteToPipe(VOID) 
{ 
   DWORD dwRead, dwWritten; 
   CHAR chBuf[BUFSIZE]; 
 
// Read from a file and write its contents to a pipe. 
 
   for (;;) 
   { 
      if (! ReadFile(hInputFile, chBuf, BUFSIZE, &dwRead, NULL) || 
         dwRead == 0) break; 
      if (! WriteFile(hChildStdinWrDup, chBuf, dwRead, 
         &dwWritten, NULL)) break; 
   } 
 
// Close the pipe handle so the child process stops reading. 
 
   if (! CloseHandle(hChildStdinWrDup)) 
      ErrorExit("Close pipe failed\n"); 
} 
 
VOID ReadFromPipe(VOID) 
{ 
   DWORD dwRead, dwWritten; 
   CHAR chBuf[BUFSIZE]; 
   HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE); 

// Close the write end of the pipe before reading from the 
// read end of the pipe. 
 
   if (!CloseHandle(hChildStdoutWr)) 
      ErrorExit("Closing handle failed"); 
 
// Read output from the child process, and write to parent's STDOUT. 
 
   for (;;) 
   { 
      if( !ReadFile( hChildStdoutRdDup, chBuf, BUFSIZE, &dwRead, 
         NULL) || dwRead == 0) break; 
      if (! WriteFile(hSaveStdout, chBuf, dwRead, &dwWritten, NULL)) 
         break; 
   } 
} 
 
VOID ErrorExit (LPTSTR lpszMessage) 
{ 
   fprintf(stderr, "%s\n", lpszMessage); 
   ExitProcess(0); 
} 
#endif

/*
 * pthread_slot()
 *
 * Read the output of the started process
 */
void console_feeder::pthread_slot( void )
{
	int c, i;

	/* read from the process while it is not finished or suspended */
	for( int linen = 0; !feof(pipe_fp); linen ++ )
	{
		/* get a line from the process */
		for(i = 0; (c = fgetc(pipe_fp))!= '\n' && !feof(pipe_fp); i++)
		{
			if (c == -1)        /* it is suspended now */
				return;
			else linebuffer.append( c );
		}
		processLine(linebuffer);
		console->text()->append(linebuffer);
		if (console->text()->errorline != -1)     /*!OP go to the first error */
			console->text()->setCursorPosition(console->text()->errorline, 0);
		linebuffer.truncate( 0 );
	}

	if ( feof(pipe_fp) )        /* the process is finished */
	{
		console_timer->stop();
		console_timer = NULL;
#ifdef _SYSTEM_LINUX_
		pclose(pipe_fp);
#else
		_pclose(pipe_fp);
#endif
		console->pSignTimer->stop();
		running = false;
		console->iSignSpinning = 0;
		console->displayCurrentPixmap();
	}
}

void console_feeder::processLine(QString &line)
{
	console->signTimeout();
	if (line.find("Entering directory") != -1)
	{
		QString *path = NULL;
#if QT_VERSION > 280
		path = new QString(line.section(' ', 3, 3));
#endif
		path->remove(0, 1);
		path->remove(path->length() - 1, 1);
		console->text()->dirs.append( path );
//!		printf("Section: %s\n", path->latin1());
	} else if (line.find("Leaving directory") != -1)
	{
		QString *path = new QString(*console->text()->dirs.at(0));
		console->text()->dirs.append( path );
	} else
	{
		QString *path = new QString(*console->text()->dirs.last());
		console->text()->dirs.append( path );
	}
	if ((line.find(':') != -1) && (line.find(':', line.find(':')+1) != -1))
	{
		if (console->text()->errorline == -1)
			console->text()->errorline = console->text()->numLines();
		if (line.find("rror") != -1)
		{
			console->iCompileNumErr ++;
			console->iCompileState = 2;
			console->displayCurrentPixmap();
			console->updateToolTip();
		}
		if (line.find(": warning: ") != -1)
		{
			console->iCompileNumWarn ++;
			if (console->iCompileState <= 1)
			{
				console->iCompileState = 1;
				console->displayCurrentPixmap();
			}
			console->updateToolTip();
		}
	}
}

#include <string.h>
#include <stdlib.h>
/**
 * Manipulates the colors of an XPM text image
 *
 * This function multiplies the components of the picture's colors by
 * fRed, fGreen, fBlue.
 */
char **ConsoleWindow::changePixTextColors( const char *sPixText[], float fRed, float fGreen, float fBlue )
{
	char **sReturn;
	int iColumns, iRows, iColors, iCharsPerPixel;
	float fMult[3] = { fRed, fGreen, fBlue };
	int i;

	sscanf(sPixText[0], "%d %d %d %d", &iColumns, &iRows, &iColors, &iCharsPerPixel);
	sReturn = (char **)malloc( (iColors + iColumns + 1) * sizeof (char *));
	for(i = 0; i < (iColors + iColumns + 1); i++)
		sReturn[i] = strdup(sPixText[i]);
	for(i = 1; i <= iColors; i++)
	{
		const char *sStr = sPixText[i], *sHash;
		int iValue;
		char *sDStr;
		if ((sHash = strstr( sStr + iCharsPerPixel, "#")) != NULL)
		{
			sDStr = sReturn[i] + (sHash - sStr) + 1;
			sStr = sHash + 1;
			for(int j = 0; j < 3; j++)
			{
				sscanf(sStr + j*2, "%02x", &iValue);
				iValue = (int)((float)iValue * fMult[j]);
				if (iValue > 255) iValue = 255;
				sprintf(sDStr + j*2, "%02x", iValue);
			}
		}
	}
	return sReturn;
}

void ConsoleWindow::purgePixText( char *sPixText[] )
{
	int iColumns, iRows, iColors, iCharsPerPixel;
	sscanf(sPixText[0], "%d %d %d %d", &iColumns, &iRows, &iColors, &iCharsPerPixel);
	for(int i = 0; i < (iColors + iColumns + 1); i++)
		free(sPixText[i]);
	free(sPixText);
}

void ConsoleWindow::displayCurrentPixmap( bool bRepaint )
{
	static int iPCompileState = -1, iPSignSpinning = -1;
	if (bRepaint || (iPCompileState != iCompileState)
	             || (iPSignSpinning != iSignSpinning))
		pStatusSigner->setPixmap(*pSignPix[iCompileState][iSignSpinning]);
	iPCompileState = iCompileState;
	iPSignSpinning = iSignSpinning;
}

#include <qtooltip.h>
QString sToolTipText;
void ConsoleWindow::updateToolTip( void )
{
	sToolTipText = "There was <b>";
	sToolTipText.append(QString::number(iCompileNumErr));
	sToolTipText.append((console->iCompileNumErr == 1 ? "</b> error, <b>" : "</b> errors, <b>"));
	sToolTipText.append(QString::number(iCompileNumWarn));
	sToolTipText.append((console->iCompileNumWarn == 1 ? "</b> warnings." : "</b> warnings."));
	QToolTip::remove(pStatusSigner);
	QToolTip::add(pStatusSigner, sToolTipText);
}

#include <qhbox.h>
#include <qlayout.h>
ConsoleWindow::ConsoleWindow( QWidget *parent ) : QVBox( parent )
{
	QHBox *h = new QHBox( this );
	h->setMaximumWidth(400);
	QHBox *titlebox = new QHBox( h );
	titlebox->setFrameStyle( Box );
	titlebox->setLineWidth( 1 );
	titlebox->setMidLineWidth( 2 );
	titlebox->setMargin( 2 );
	QLabel *title = new QLabel("<b>  Console  </b>", titlebox);
	title->setTextFormat( Qt::RichText );
	titlebox->show();

	/* status signer tool */
	for(int i = 0; i < 3; i++)
		for( int j = 0; j < 4; j++)
		{
			const char **pPixText = NULL;
			char **pNew = NULL;
			if (j == 0) pPixText = console_sign_green_xpm;
			else if (j == 1) pPixText = console_sign_green_working1_xpm;
			else if (j == 2) pPixText = console_sign_green_working2_xpm;
			else if (j == 3) pPixText = console_sign_green_working3_xpm;
			/* convert colors to orange (value 1) or red (value 2) if needed */
			if (i == 1) pNew = changePixTextColors( pPixText, 3.5F, .8F, .3F);
			else if (i == 2) pNew = changePixTextColors( pPixText, 4, .4F, .2F);
			pSignPix[i][j] = new QPixmap( (pNew ? (const char **)pNew : pPixText) );
			if (pNew != NULL) purgePixText ( pNew );
		}
	iSignSpinning = iCompileState = 0;
	new QLabel(" ", h);
	pStatusSigner = new QLabel(" ", h);
	displayCurrentPixmap();
	pStatusSigner->show();
	pSignTimer = new QTimer( this, "consolesigntimer" );
	QObject::connect( pSignTimer, SIGNAL(timeout( void )),
						this, SLOT( signTimeout( void ) ) );

	(new QLabel("     Make ", h))->show();
	getsymbol = new QLineEdit( h );
	getsymbol->setMaximumWidth( 80 );
	getsymbol->show();
	(new QLabel(" in ", h))->show();
	getpath = new QLineEdit( h );
	getpath->setMaximumWidth( 200 );
	getpath->show();

	QString *sSymb = (QString *)pMod->pQuery("cb.get.properties.string", 
		        "module.console.make.symbol", NULL, NULL);
	if (sSymb != NULL) getsymbol->setText( *sSymb );

	QString *sPath = (QString *)pMod->pQuery("cb.get.properties.string", 
		        "module.console.make.path", NULL, NULL);
	if (sPath != NULL) getpath->setText( *sPath );

	multiline = new ConsoleText( this );
	setFrameStyle( QFrame::StyledPanel | QFrame::Sunken ); 
	setMargin(2);
}

void ConsoleWindow::signTimeout()
{
	if (iSignSpinning)
	{
		iSignSpinning += 1;
		if (iSignSpinning >= 4) iSignSpinning = 1;
	}
	displayCurrentPixmap();
}

ConsoleText::ConsoleText( QWidget *parent ) : QMultiLineEdit( parent )
{
	QObject::connect( this, SIGNAL( cursorPositionChanged( int, int ) ),
			          this, SLOT( cursorMoved( int, int ) ) );
	const QColor *color;

	QPalette *palette = new QPalette( this->palette() );
	QColorGroup *qg = new QColorGroup( this->palette().active() );
	color = (QColor *)pMod->pQuery("cb.get.properties.color", 
		        "module.console.color.text", NULL, NULL);
	if (color == NULL) color = new QColor( 245, 245, 245 );
	qg->setColor( QColorGroup::Text, *color );
	color = (QColor *)pMod->pQuery("cb.get.properties.color", 
		        "module.console.color.background", NULL, NULL);
	if (color == NULL) color = new QColor( 70, 70, 70 );
	qg->setColor( QColorGroup::Base, *color );
	palette->setActive( *qg );
	palette->setInactive( *qg );
	setPalette( *palette );
	setMinimumHeight(10);
}

void ConsoleText::cursorMoved( int y, int x )
{
	cursorline = y;
}

void ConsoleText::doubleClickEvent( int iPar, int iCol )
{
	const QString &line = textLine(cursorline);
	const QString *path = dirs.at(cursorline+1);
	if (path == NULL) return;           /*? why is it necessary */
#if QT_VERSION > 280
//	e->accept();
#endif
	DEBUG_PRINT("Cursor line: %d - %s\n", cursorline, line.latin1());
	DEBUG_PRINT("path: %s\n", path->latin1());                     

	/* find the line in a buffer where this error/warning occured */
#if QT_VERSION > 280
	if ((line.find(':') != -1) && (line.find(':', line.find(':')) != -1))
	{
		const QString &file = line.section(':', 0, 0);
		const QString &num = line.section(':', 1, 1);

		/* convert the line number (string) to int */
		bool bOK;
		int iLineNum = num.toInt(&bOK);
		if (bOK)
		{
			QString sLineNum;
			sLineNum.setNum(iLineNum - 1);
			QString *filepath = new QString(*path);
			filepath->append('/');
			filepath->append(file);
			if (pMod->pQuery("cb.show.editor.filebypath", filepath->latin1(), NULL, NULL))
				pMod->pQuery("cb.set.buffer.cursor.position", "0", sLineNum.latin1(), NULL);

			/* file is not opened - ask user to open it or not */
			else
			{
				QString message( "This file (" );
				message.append( file );
				message.append( ") is not opened. <b>Open</b> it?" );
				QMessageBox mb("Freddy", message,
				               QMessageBox::Warning, QMessageBox::Yes|QMessageBox::Default,
				               QMessageBox::No|QMessageBox::Escape, QMessageBox::NoButton,
				               NULL, "Open it?", true);
				mb.setFocus();
				if( mb.exec() == QMessageBox::Yes )
				{
					pMod->pQuery("cb.editor.openfile", filepath->latin1(), NULL, NULL);
					pMod->pQuery("cb.set.buffer.cursor.position", "0", sLineNum.latin1(), NULL);
				}
			}
		}
	}
#endif
}

