/**
 * @file Command.cpp
 * @brief Command handling - Menu and shortcuts (definitions)
 * 
 * @author Ferenc Moricz
 * $Revision: 1.24 $
 * $Date: 2004/02/22 00:12:49 $
 */

#include <qpopupmenu.h>
#include <qmenubar.h>

#include "Editor.h"
#include "Command.h"

#include <stdio.h>
#include <stdlib.h>

static char *shortsymbols[] = {
       "<insert>", "<delete>", "<home>", "<end>",
       "<pageup>", "<pagedown>",
       "<left>", "<right>", "<up>", "<down>",
       "<printscreen>", "<scrolllock>", "<pause>",
       "<f1>", "<f2>", "<f3>", "<f4>", "<f5>", "<f6>", 
       "<f7>", "<f8>", "<f9>", "<f10>", "<f11>", "<f12>", NULL
};

static int shortvalues[] = {
       Qt::Key_Insert, Qt::Key_Delete, Qt::Key_Home, Qt::Key_End,
       Qt::Key_PageUp, Qt::Key_PageDown,
       Qt::Key_Left, Qt::Key_Right, Qt::Key_Up, Qt::Key_Down,
       Qt::Key_Print, Qt::Key_ScrollLock, Qt::Key_Pause,
       Qt::Key_F1, Qt::Key_F2, Qt::Key_F3, Qt::Key_F4, Qt::Key_F5, Qt::Key_F6,
       Qt::Key_F7, Qt::Key_F8, Qt::Key_F9, Qt::Key_F10, Qt::Key_F11, Qt::Key_F12
};

// . loadShortCuts(..)
/**
 * @brief Load shortcuts and process them
 * @param file the file to load from
 * @return true - if the loading was successful\n false - otherwise
 */
bool Command::loadShortCuts( const char * file, QDict<int> *comlist )
{
	if (file == NULL) return false;

	bool ret = shortcut.loadConfig( file );

	vector<PropertyItem, allocator<PropertyItem> > *shortstag = &shortcut["editor.shortcuts"];
	if (shortstag == NULL) return false;

	/* load shortcuts */
	for(int i=0; i < (int)shortstag->size(); i++)
	{
		int key = 0;
		QString tag = (*shortstag)[i].getString()->lower();
		const char *tagc = tag.latin1();

		/* modifiers: ctrl, alt, shift */
		while(true)
		{
			if ((tag.length() >= 6) && !memcmp(tagc, "<ctrl>", 6))
			{
				key |= Qt::CTRL;
				tagc += 6;
			}
			else if ((tag.length() >= 5) && !memcmp(tagc, "<alt>", 5))
			{
				key |= Qt::ALT;
				tagc += 5;
			}
			else if ((tag.length() >= 7) && !memcmp(tagc, "<shift>", 7))
			{
				key |= Qt::SHIFT;
				tagc += 7;
			}
			else break;
			if (tagc[0] == '-') tagc ++;
		}

		for(int i = 0; shortsymbols[i]!=NULL; i++)  // non-printable character
		if (strstr(tagc, shortsymbols[i]))
		{
			key |= shortvalues[i];
			break;
		}
		if (!(key&0xffff))                          // printable character
		{
			if ((*tagc >= 'a') && (*tagc <= 'z')) key |= *tagc - 'a' + Qt::Key_A;
			else if (*tagc >= 20) key |= *tagc;
		}

		/* the command */
		while((*tagc != ' ') && (*tagc != '\0')) tagc++;  // go until the next
		tagc++;

		if (key & 0xffff)                           // insert shortcut => dict
		{
			shortcut_command.insert(key, new QString(tagc));

			if (comlist != NULL)
				comlist->insert(*new QString(tagc), new int(key));
		}
	}

	return ret;
}


CommandMenuItem::CommandMenuItem(QPopupMenu *pParent, int iMenuID, QString *sCommand)
{
	{
		this->pParent = pParent;
		this->iMenuID=iMenuID;
		this->sCommand = sCommand;
	}
}

#if QT_VERSION > 280
#include <qstylefactory.h>
#endif
// . loadMenu(.)
/**
 * @brief Redefined function: load menu and process them
 * @param file the file to load from
 * @return true - if the loading was successful\n false - otherwise
 */
extern const char * freddy_xpm[];
QPopupMenu *Command::loadMenu( const char * file, const char *sMenuString, QDict<int> *comlist )
{
	static int iMenuID = 0;
	QPopupMenu *pRetMenu = NULL;

	if (file == NULL) return false;

	if (menu.size() == 0)                      /* read file if needed */
		menu.loadConfig( file );

	vector<PropertyItem, allocator<PropertyItem> > *menutag = &menu[sMenuString];
	if (menutag == NULL) return false;

	if (!strcmp(sMenuString, "editor.menu"))
	{
		QMenuBar *pMenuBar = Editor::editor->pMainWindow->menuBar();
		if (pMenuBar) pMenuBar->clear();
	}
	else pRetMenu = new QPopupMenu();

	/* load menu */
	char mstr[160];
	QDict<QPopupMenu> menudict;
	QPopupMenu *pmenu;
	for(int i = 0; i<(int)menutag->size(); i++)  // collect menus
	{
		QString tag = *(*menutag)[i].getString();
		const char *tagc = tag.latin1();

		pmenu = NULL;
		int mtbeg, mtend, *key=NULL;

		/* process tag - create menu tree (e.g "File/Open editor.fileopen") */
		for(mtbeg = mtend = 0; tagc[mtend] != '\0'; mtend++)
		{
			if (tagc[mtend] == '/')             // found a menu point
			{
				mstr[mtend] = '\0';
				if (menudict.find(mstr)==NULL)  // create new popup
				{
					QPopupMenu *m = new QPopupMenu( Editor::editor->pMainWindow );
					if (pmenu == NULL)                  // add to menubar
					{
						if (!pRetMenu)
						{
							Editor::editor->pMainWindow->menuBar()->insertItem(tr(mstr+mtbeg), m);
						}
						else pRetMenu->insertItem(tr(mstr+mtbeg), m);
					}
					else pmenu->insertItem(tr(mstr+mtbeg), m);  // or to last
					pmenu = m;
					menudict.insert(mstr, m);
				} else pmenu = menudict.find(mstr);
				mtbeg = mtend + 1;
			} else if ((tagc[mtend]==' ') || (tagc[mtend]=='\t')) break;
			if (tagc[mtend] != '_') mstr[mtend] = tagc[mtend];
			else mstr[mtend] = ' ';
		}

		/* create a menu */
		mstr[mtend] = '\0';

		/* create session list in the menu */
		if (strstr(tagc+mtend, "menu.sessionlist") != NULL)
		{
			Editor::editor->sessionmenu = pmenu;
			Editor::editor->sessionmenuindex = pmenu->count();
			QObject::connect(Editor::editor->sessionmenu, SIGNAL(activated(int)),
			                 Editor::editor, SLOT( sessionMenu(int) ));
		}
		/* create available style list in the menu */
		if (strstr(tagc+mtend, "menu.list.styles") != NULL)
		{
			if (pmenu != NULL)
			{
				// get the list of the available styles
#ifdef _SYSTEM_LINUX_
				dmesg("5:command", "Available library's path: %s\n", QApplication::libraryPaths().join(", ").latin1());
				dmesg("5:command", "Available styles: %s\n", QStyleFactory::keys().join( ", " ).latin1());
				QStringList SStyleList = QStyleFactory::keys();
				SStyleList.sort();
				// loop style = stylelist[1..n]
				for ( QStringList::Iterator it = SStyleList.begin();
						it != SStyleList.end(); ++it )
				{
					pmenu->insertItem((*it).latin1(),
						this, SLOT(menuEvent(int)), 0, i + iMenuID);
					QString *sCommand = new QString("editor.styleset ");
					sCommand->append(*it);
					menu_command[i + iMenuID] =
					      new CommandMenuItem(pmenu, i + iMenuID++, sCommand);
				}
#else
			pmenu->insertItem("Windows",
						this, SLOT(menuEvent(int)), 0, i + iMenuID);
			pmenu->insertItem("CDE",
						this, SLOT(menuEvent(int)), 0, i + iMenuID);
			pmenu->insertItem("Motif",
						this, SLOT(menuEvent(int)), 0, i + iMenuID);
			pmenu->insertItem("Platinum",
						this, SLOT(menuEvent(int)), 0, i + iMenuID);
#endif
			pStyleListMenu = pmenu;
			}
		}
		else if (menudict.find(mstr) == NULL)   // create the menu
		{
			while( ((tagc[mtend]==' ')||(tagc[mtend]=='\t')) && 
			       (tagc[mtend]!='\0') ) mtend++;       // skip white spaces

			if (mstr[mtbeg]=='\0')                      // separator
			{
				if (pmenu != NULL) pmenu->insertSeparator();
				else Editor::editor->pMainWindow->menuBar()->insertSeparator();
			}
			else
			{
				QPopupMenu *pParentMenu = NULL;
				/* get the shortcut for this command */
				if (comlist != NULL) key = comlist->find(tagc+mtend);

				/* finally create the menuitem */
				if (pmenu == NULL)
				{
					if (!pRetMenu)
					{
						Editor::editor->pMainWindow->menuBar()->insertItem(tr(mstr+mtbeg),
						this, SLOT(menuEvent(int)), (key ? *key:0), i + iMenuID);
					}
					else
					{
						pRetMenu->insertItem(tr(mstr+mtbeg),
							this, SLOT(menuEvent(int)), (key ? *key:0), i + iMenuID);
						pRetMenu->setAccel( key ? *key:0, i + iMenuID );
						pParentMenu = pRetMenu;
					}
				}
				else {
					pmenu->insertItem(tr(mstr+mtbeg),
						this, SLOT(menuEvent(int)), (key ? *key:0), i + iMenuID);
					pParentMenu = pmenu;
				}

				menu_command[ i + iMenuID ] =
					new CommandMenuItem(pParentMenu,
						i + iMenuID, new QString(tagc+mtend));
			}
		} else
		{
			DEBUG_WARNING("Invalid (already used) menu: %s", mstr);
			continue;
		}
	}
	iMenuID += menutag->size();

//	if (!strcmp(sMenuString, "editor.menu"))
//	{
//		Editor::editor->pMainWindow->menuBar()->insertItem(QPixmap(freddy_xpm), NULL, "");
//	}

	return pRetMenu;
}

// .command(.)
/**
 * @brief Run a command
 * @param c the command
 * @return the result
 *
 * This function finds the appropriate class to run the command. It looks
 * at the beginning of the command (\em editor. , \em buffer.) and uses
 * the right object.\n
 * It breaks out the parameters from the command.
 */
char *Command::command( const char * comstr )
{
	bool bCheck = false;
	char *pReturn = NULL;

	EditSession *session;
	EditBuffer *buffer;

	if (comstr == NULL) return NULL;
                                /* get actual session and buffer */
	session = (EditSession *)Editor::editor->getStack()->visibleWidget();
	if (session != NULL) buffer = (EditBuffer *)session->currentPage();
	else buffer = NULL;

	char *cp[5], *c, *dup = strdup(comstr);

	/* is this a 'command check' command? */
	if (strstr(dup, "CHECK") == dup)
	{
		bCheck = true;
		cp[0] = dup + 6;
	} else cp[0] = dup;

	bool anothercommand;

	// search for commands and execute them
	for(int n = -1; ; )
	{
		// new command - init argument variables
		if ( n==-1 )
		{
			n = 0;
			cp[1] = cp[2] = cp[3] = NULL;
		}

		// search next token
		while((*cp[n]==' ') || (*cp[n]=='\t')) cp[n]++;
		for(c = cp[n]; (*c!=' ') && (*c!='\t') && (*c!=';') && (*c!='\0'); c++);

		// run a command if it is ready
		if ((anothercommand = (*c == ';')) || (*c == '\0'))
		{
			if (c == cp[n]) cp[n--] = NULL; // last ';' is separated 
			else *c = '\0';                 // close this token

			// run the command
			if (n >= 0)
			{
				char *comname = strchr(cp[0], '.') + 1;
				if (comname != NULL)
				{
					if (!memcmp(cp[0], "buffer.", 7))   // buffer command
					{
						if (buffer != NULL)
						{
							if (bCheck)
							{
								QString sQuery = "CHECK ";
								sQuery += comname;
								pReturn = buffer->command(sQuery.latin1(), cp[1], cp[2], cp[3]);
							}
							else buffer->command(comname, cp[1], cp[2], cp[3]);
						}
					} else
					if ((!memcmp(cp[0], "editor.", 7)) ||
					    (!memcmp(cp[0], "debug.", 6)))   // editor command
					{
						QString sQuery;
						if (bCheck)
						{
							sQuery = "CHECK ";
							sQuery += comname;
						} else sQuery = comname;
						pReturn = Editor::editor->command(sQuery.latin1(), cp[1], cp[2], cp[3]);
					} else
					if ( (!memcmp(cp[0], "plugin.", 7))
						 ||(!memcmp(cp[0], "module.", 7)) )
						EditBuffer::module->command(comname, cp[1]);
					else
					if (strstr(cp[0], "menu.") != cp[0])
						DEBUG_WARNING("Invalid command: %s", comstr);
				}
			}

			if (!anothercommand) break; // end of string - all is done
			else {
				n = -1;                 // go to the next command
				cp[0] = c+1;
			}
		} else if (n < 4)               // search for the next token 
		{
			*c = '\0';
			n++;
			cp[n] = c + 1;
		}
	}
	free(dup);

	if (!bCheck) updateMenuItems();

	return pReturn;
}

// .shortEvent(.)
/**
 * @brief Event function to filter shortcuts from key events
 * @param e the key event
 * @return true - if shortcut found\n false - otherwise
 */
bool Command::shortEvent( QKeyEvent *e )
{
	bool bRETURN = false;

	long key = e->key();

	if (e->state() & Qt::ControlButton) key |= Qt::CTRL;
	if (e->state() & Qt::AltButton) key |= Qt::ALT;
	if (e->state() & Qt::ShiftButton) key |= Qt::SHIFT;

	/* do the command interface have an entry for this event ? */
	if (shortcut_command.find(key))
	{
		command(shortcut_command.find(key)->latin1());
		bRETURN = true;
	}

	return bRETURN;
}

// addLabel(..)
/**
 * @brief i18n - Registers a label to update its text automatically
 * @param pLabel the label to add
 * @param pText the original text of the label
 */
void Command::addLabel(QLabel *pLabel, QString *pText)
{
	oLabels.push_back(pLabel);
	oLabelTexts.push_back(pText);
}
// updateLabels()
/**
 * @brief i18n - Update the text of all of the registered labels
 */
void Command::updateLabels()
{
	for(int i = 0; i < (int)oLabels.size(); i++)
	{
		oLabels[i]->setText(tr(oLabelTexts[i]->latin1()));
		oLabels[i]->update();
	}
}

// updateMenuItems()
/**
 * @brief Update the registrated menu items' status (disabled/enabled)
 */
void Command::updateMenuItems()
{
	for(map<int, CommandMenuItem *>::iterator iMi = menu_command.begin();
		iMi != menu_command.end(); iMi++)
	{
		QString sQuery = "CHECK ";
		void *pReturn;
		sQuery += *((*(*iMi).second).sCommand);
		if ((*iMi).second->pParent)
		{
			QPopupMenu *pMenu = (*iMi).second->pParent;
			pReturn = command(sQuery.latin1());
			pMenu->setItemEnabled((*iMi).second->iMenuID, (pReturn?true:false));
		}
	}
}

void Command::menuEvent( int id )
{
	if (menu_command.find(id) != menu_command.end())
	{
		QString *pEventFound = menu_command[id]->sCommand;

		command(pEventFound->latin1());
	}
}

