/**
 * @file EditBuffer.cpp
 * @brief The main editor widget
 * 
 * @author Ferenc Moricz
 * $Revision: 1.110 $
 * $Date: 2004/02/22 00:12:49 $
 */

/**
 * @class EditBuffer
 * @brief The widget, which provides the area to edit text. It is
 * the main, most important widget.
 *
 * <b>Commands:</b>\n\n
 * <b>scrollto <to></b> - scroll to the line number specified by \em <to> \n
 * <b>scrolltocursor</b> - scroll to the line where the cursor is visible \n
 * <b>scrollbar</b> - show/hide scrollbar as necessary <p>
 * <b>setcursor <x> <y></b> - set the cursor's position <p>
 * <b>selectcopy</b> - copy the selected area to the clipboard \n
 * <b>selectpaste</b> - paste the text which is in the clipboard
 *                      (new selection) <p>
 * <b>find</b> - find in text (a text form will appear to get the text) <p>
 * <b>checkfile</b> - check the file if it is older than the stored version
 */

#include "EditBuffer.h"
#include "EditBufferFind.h"
#include "Editor.h"

#include <qcursor.h>
#include <qfont.h>
#include <qgrid.h>
#include <qhbox.h>
#include <qlabel.h>
#include <qpainter.h>
#include <qtooltip.h>

#include <stdio.h>
#include <stdlib.h>

extern char FREDDY_VERSION[], FREDDY_BUILD[];

ModuleLoader *EditBuffer::module;
QPixmap *EditBufferLineBar::pm;
EditBufferLineBar::EditBufferLineBar(QWidget *parent, EditBuffer *buffer)
    : QFrame(parent, "", WResizeNoErase | Qt::WRepaintNoErase), oNumberFontMt(QFont())
{
	this->buffer = buffer;

	reconfigure();

	QPalette *palette=new QPalette( this->palette() );
	QColorGroup *qg=new QColorGroup( palette->active() );
	qg->setColor( QColorGroup::Background, QColor(70, 70, 70) );
	qg->setColor( QColorGroup::Base, QColor(70, 70, 70) );
	qg->setColor( QColorGroup::Foreground, QColor(190, 190, 190) );
	palette->setActive( *qg );
	palette->setInactive( *qg );
	setPalette( *palette );

	setSizePolicy( QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum) );
	setFrameStyle( QFrame::Panel | QFrame::Sunken );
	setLineWidth( 2 );
	setMidLineWidth( 3 );

/* WIN setBackgroundMode(Qt::NoBackground, Qt::NoBackground); */
	dbuf = EditProperties::pMain->getBoolValue( "buffer.doublebuffering", FALSE);
	if (dbuf) pm = new QPixmap( 1, 1 );
}

QSize EditBufferLineBar::sizeHint () const
{
	return QSize( iBarWidth, 10 );
}

void EditBufferLineBar::resizeEvent( QResizeEvent *e )
{
	if (dbuf && (( e->size().width()>pm->width()) ||
	             ( e->size().height()>pm->height())) )
	pm->resize(e->size());
}

void EditBufferLineBar::paintEvent(QPaintEvent *pe)
{
	QPainter p;

	dmesg("8:buffer", "Paint event");

	/// Display line numbers
	if (dbuf) p.begin(pm, this);
	else p.begin( this );
	p.setPen( foregroundColor() );
	p.fillRect( 2, 2, width()-2, height()-2, backgroundColor() );

	for(int i = 0; i<buffer->text->nLines; i++)
	{
//		if ( (i < buffer->text->repaintfrom) ||
//		     (i > buffer->text->repaintto) ) continue;
		/* draw the line numbers vertical aligned to the center of the buffer's text */
		QString sNumber = QString::number(buffer->text->firstVisible+i+1);
		int iBottom = ((2*i+1)*buffer->text->fontheight)/2  + fontascent/2 - 1;

		p.drawText( QRect(4, iBottom - fontascent, iBarWidth - 10, iBottom), AlignRight,
			QString::number(buffer->text->firstVisible+i+1) );
	}
	p.end();
                                /* bitblt back the buffer */
	if (dbuf) for(int i = 0; i<buffer->text->nLines; i++) 
	{
		if ( (i < buffer->text->repaintfrom) ||
		     (i > buffer->text->repaintto) ) continue;
		bitBlt(this, 0, i * buffer->text->fontheight,
		         pm, 0, i * buffer->text->fontheight,
		       width(), buffer->text->fontheight);
	}

	QFrame::paintEvent( pe);
	return;
}

// EditBufferLineBar :: reconfigure()
/**
 * @brief reconfigures the EditBufferLineBar object
 */
void EditBufferLineBar::reconfigure( void )
{
	/* 1. get the line numbers' font from the config files */
	QFont oLineDef("Fixed", 10);
	oNumberFont = *(EditProperties::pMain->getFontValue("buffer.linenumber.font", &oLineDef));
	oNumberFontMt = QFontMetrics(oNumberFont);

	setFont(oNumberFont);
	fontascent = oNumberFontMt.ascent();

	/* bar width calculation */
	if (buffer->text)
	{
		QString sZeros;
		int iBufLen = buffer->text->lines.size() + buffer->text->nLines;
		for(int i = iBufLen; i!= 0; i /= 10) sZeros += "0";
		if (sZeros.length() < 2) sZeros="00";
		iBarWidth = oNumberFontMt.width(sZeros) + 10;
		setMinimumWidth(iBarWidth);
	}
}

#include <qpushbutton.h>
#include <qlayout.h>
EditBuffer::EditBuffer( QWidget *parent, const QString *name )
		: QVBox( parent, *name )
{
	QHBox *pHbox;

	oUndo.pBuffer = this;

	this->pSession = (EditSession *)parent;
	this->name = new QString(*name);
	this->file = NULL;

	bSyntaxHighlighting = true;

	pHbox = new QHBox( this );
	pHbox->setFrameStyle( StyledPanel );
	pHbox->setLineWidth( 2 );

	text = NULL;

	linebar = new EditBufferLineBar( pHbox, this );

	bufferview_v = new QVBox( pHbox );
	bufferview = new QHBox( bufferview_v );

	hscroll = vscroll = NULL;
	hscrollValue = 0;

	text = new EditBufferText( this, bufferview );
	childs.push_back(text);

	QVBox *pMapHolder = new QVBox( bufferview );
	pMapHolder->setSizePolicy( QSizePolicy( QSizePolicy::Fixed, QSizePolicy::Preferred ) );
	pMapHolder->setMaximumWidth( 0 );
	pMapHolder->show();
	childs.push_back(pMapHolder);

	bFileModified = false;
	bLoading = false;
	bUTF8 = true;

	tabsize = text->properties->getNumericalValue("buffer.tab.size", 8);

	/* Status bar */
	QHBox *hbox = new QHBox( this );
	hbox->setSizePolicy( *new QSizePolicy ( 
			QSizePolicy::Preferred, QSizePolicy::Fixed ) );
	hbox->setFrameStyle( StyledPanel );
	hbox->setLineWidth( 1 );

	/* create |_x_|_y_| cursor place displayer widget */
	position = new QHBox(hbox);
	position->setFrameStyle( Box );
	position->setLineWidth( 1 );
	position->setMidLineWidth( 2 );
	position->setMargin( 2 );

	(new QLabel("     ", hbox, "     "))->show();

	pFileTypeCombo = new QComboBox( hbox );
	pFileTypeCombo->insertItem( QString("DOS"), (int)(EditBuffer::FILETYPE_DOS) );
	pFileTypeCombo->insertItem( QString("UNIX"), (int)(EditBuffer::FILETYPE_UNIX) );
	pTabCombo = new QComboBox( hbox );
	pTabCombo->insertItem( QString("Tab"), 0 );
	pTabCombo->insertItem( QString("1"), 1 );
	pTabCombo->insertItem( QString("2"), 2 );
	pTabCombo->insertItem( QString("3"), 3 );
	pTabCombo->insertItem( QString("4"), 4 );
	pTabCombo->insertItem( QString("8"), 8 );
	pUTFCombo = new QComboBox( hbox );
	pUTFCombo->insertItem( QString("UTF-8"), 0 );
	pUTFCombo->insertItem( QString("Latin1"), 1 );

	(new QLabel("     ", hbox, "     "))->show();
	QLabel *pCommandLabel = new QLabel("  Command: ", hbox, "  Command: ");
	pCommandLabel->show();
	Editor::editor->com->addLabel( pCommandLabel, new QString("  Command: ") );
	pCommandTool = new QLineEdit( hbox );
	pCommandTool->setMaximumWidth( 150 );
	QObject::connect( pCommandTool, SIGNAL(returnPressed()), 
					  this, SLOT( commandToolEnter() ) );
	position_x = new QLabel ( "", position, "");
	position_x->setFrameStyle( StyledPanel | Sunken ); 
	position_x->setMargin(1);
	position_y = new QLabel ( "", position, "" );
	position_y->setFrameStyle( StyledPanel | Sunken ); 
	position_y->setMargin(1);
	statusbar = new QStatusBar( hbox );
	statusbar->setSizeGripEnabled( false );
	statusbar->setMinimumHeight(statusbar->height() + 4);
	findbox = NULL;
	QFontMetrics oFontMt = QFontMetrics(position_x->font());
	position_x->setMinimumWidth(oFontMt.width("0000") + 10);
	position_y->setMinimumWidth(oFontMt.width("0000") + 10);

	linebar->reconfigure();
}

// EditBuffer :: reconfigure()
/**  
 * @brief reconfigures the EditBuffer object
 */
void EditBuffer::reconfigure( void )
{
	tabsize = text->properties->getNumericalValue("buffer.tab.size", 8);
	text->reconfigure();
	linebar->reconfigure();

	QPalette *palette=new QPalette( position->palette() );
	QColorGroup *qg=new QColorGroup( palette->active() );
	QColor oColorGray = QColor(gray);
	QColor oColorLightGray = QColor(210, 210, 210);
	const QColor *pFrameColor =
		text->properties->getColorValue( "buffer.positioncounter.frame.color", &oColorGray, true);
	//QColor *pFrameColor = new QColor(100, 100, 100);
	qg->setColor( QColorGroup::Dark, *pFrameColor );
	qg->setColor( QColorGroup::Light, *pFrameColor );
	qg->setColor( QColorGroup::Background, *text->properties->getColorValue( 
		                  "buffer.positioncounter.bg.color", &oColorLightGray, true));

	palette->setActive( *qg );
	palette->setInactive( *qg );
	position->setPalette( *palette );

	position_x->setPalette( *palette );
	position_y->setPalette( *palette );

	update();
}

/** Kaki
  * @callgraph */
void EditBuffer::setModified( bool bModified )
{
	/* @callgraph */

	if (bModified != bFileModified)
	{
		if (pSession == NULL) return;
		QFile *pFile = getFile();
		QString sLabel;
		if (pFile != NULL)
		{
			QFileInfo fi(*getFile());
			sLabel = fi.fileName();
		}
		else sLabel = "Untitled";
		if (bModified) sLabel.append(" *");
		pSession->setTabLabel(this, sLabel);
		bFileModified = bModified;
	}
}

#include <qmessagebox.h>
void EditBuffer::loadFile( QFile &file, int iIsUTF8 )
{
	char s[501];
	bool bAppend = false;
	EditLine *pCurLine = NULL;
                           
	text->lines.clear();                // clear loaded text
	hscrollValue = 0;

	this->file=new QFile(file.name());  // load file
	eFileType = FILETYPE_UNIX;

	/* check if the file can be a UTF8 file */
	if (iIsUTF8 == -1)
	{
		file.open( IO_ReadOnly );
		while( !file.atEnd() )
		{
			/* bug here V in the case of broken UTF sequences */
			file.readLine( s, 500 );
			s[500] = '\0';
			if (!EditLine::canBeUTF8((unsigned char *)&s[0]))
			{
				bUTF8 = false;
				break;
			}
		}
		file.close();
	} else
	bUTF8 = (iIsUTF8 == 1);

	if ((bUTF8) &&
	    !EditBufferText::properties->getBoolValue("editor.openfile.utfbydefault", true))
	{
		bLoading = true;
		int iWhatto = QMessageBox::information(NULL, "Freddy", 
		           "This file seems like an UTF-8 file. "
		           "Open it as an <b>UTF-8</b> file?", "UTF-8",
		               "No", "Cancel", 0, 2);
		bLoading = false;
		if (iWhatto == 1) bUTF8 = false;
		if (iWhatto == 2) return;
	}
	pUTFCombo->setCurrentItem((bUTF8 ? 0 : 1));
	file.open( IO_ReadOnly );
	while( !file.atEnd() )
	{
		file.readLine( s, 500 );
		s[500] = '\0';
		if ((pCurLine == NULL) || (bAppend = false))
			pCurLine = new EditLine( "" );

		pCurLine->appendString( (unsigned char *)&s[0], bUTF8 );

		text->lines.push_back( pCurLine );
		for(int i = pCurLine->length() - 1; i >= 0; i--)
		{
			if (pCurLine->at(i) == '\r')
			{
				eFileType = FILETYPE_DOS;
				pCurLine->remove(i, 1);
			}
			else if (pCurLine->at(i) == '\n')
				pCurLine->remove(i, 1);
			else break;
		}
		pCurLine = NULL;
	}
	file.close();

	reconfigure();
	pFileTypeCombo->setCurrentItem(eFileType);

	cmd_scrollBar();                    // update scrollbar

	QFileInfo fi(file);                 // update file date information
	this->file_lastmodified = fi.lastModified();
}

void EditBuffer::saveFile( )
{
	bool bRETURN = false;

	/* open the file for writing 
		 - if we can not create this file -> warning message (looped) */
	while(true) {
		QFile *pFile = getFile();
		if (!pFile)
		{
			bRETURN = false;
			break;
		}

		if (pFile->open(IO_WriteOnly)) break;

		/* 1n.2n. ask the user if he/she wants to create the file */
		QString sCannotWrite;
		QFileInfo oFileInf(*pFile);
		sCannotWrite.sprintf("Cannot open '%s' for writing! Probably this is a read-only file.", oFileInf.absFilePath().latin1());
		int iWriteOrNot = QMessageBox::warning(NULL, "Freddy", 
		               sCannotWrite, QMessageBox::Retry, QMessageBox::Cancel, QMessageBox::NoButton);
		if (iWriteOrNot == QMessageBox::Cancel)
		{
			bRETURN = true;
			break;
		}
	}

	if (!bRETURN)
	{
		file->open( IO_WriteOnly );         // load file

		eFileType = (EditBuffer::FileType) pFileTypeCombo->currentItem();

		for(unsigned int i=0; i<text->lines.size(); i++)
		{
			if (!pUTFCombo->currentItem()==1)      /* UTF-8 write mode */
			{
				QCString sUTFStr = text->lines[i]->utf8();
				file->writeBlock( (const char *)sUTFStr, sUTFStr.length());
			} else                         /* Latin1 write mode */
			{
				file->writeBlock( text->lines[i]->latin1(), 
								  text->lines[i]->length() );
			}
			if (eFileType == FILETYPE_DOS) file->putch( '\r' );
			file->putch( '\n' );
		}
		file->close();

		oUndo.setSavedPosition();

		QFileInfo fi(*file);                // update file date information
		this->file_lastmodified = fi.lastModified();
	}
}

/*
 * Events
 */
bool EditBuffer::event( QEvent *e )
{
	dmesg("6:buffer", "EVENT: EditBufferText type:%d", e->type()); 

	fflush(stdout);

	if (e->type() == QEvent::Show)      // Set caption of the main window
	{
		char caption[100], cpos=0;
		const QString *format = EditBufferText::properties->getValue("editor.caption");
		if (format != NULL)
		{
			for(int pos = 0; pos < (int)format->length(); pos++)
			{
				if (format->at(pos)=='%')
				{
					if (format->find("sessionname",pos+1) == pos+1)
					{	// session's name
						EditSession *sess = (EditSession *)Editor::editor->getStack()->visibleWidget();
						memcpy(caption+cpos, sess->name.latin1(), sess->name.length());
						cpos += sess->name.length();
						pos += sizeof("sessionname")-1;
					}
					else if (format->find("filename",pos+1) == pos+1)
					{   // buffer's name
						memcpy(caption+cpos, name->latin1(), name->length());
						cpos += name->length();
						pos += sizeof("filename")-1;
					}
					else if (format->find("freddyversion",pos+1) == pos+1)
					{   // freddy's version
						memcpy(caption+cpos, FREDDY_VERSION, strlen(FREDDY_VERSION));
						cpos += strlen(FREDDY_VERSION);
						pos += sizeof("freddyversion")-1;
					}
					else if (format->find("freddybuild",pos+1) == pos+1)
					{   // freddy's build number
						memcpy(caption+cpos, FREDDY_BUILD, strlen(FREDDY_BUILD));
						cpos += strlen(FREDDY_BUILD);
						pos += sizeof("freddybuild")-1;
					}
				} else caption[cpos++] = format->at(pos).latin1();
			}
			caption[cpos] = '\0';
			topLevelWidget()->setCaption(caption);
		}
	}

	return QWidget::event( e );
}

void EditBuffer::hscrollEvent( int n )
{
//	if ( ! scrolltimer.isActive() )
//		scrolltimer.singleShot( 50, this, SLOT(scrollRelease(void)) );
	hscrollValue = n;
	text->repaint(false);
}

void EditBuffer::scrollEvent( int n )
{
//	if ( ! scrolltimer.isActive() )
//		scrolltimer.singleShot( 50, this, SLOT(scrollRelease(void)) );
	scrollRelease();
}

void EditBuffer::commandToolEnter( )
{
	Editor::editor->com->command( pCommandTool->text().latin1() );
}

void EditBuffer::scrollRelease( void )
{
	if (vscroll) cmd_scrollTo( vscroll->value() );
}

/*
 * Commands
 */

#include <qapplication.h>
#include <qclipboard.h>

// * COMMAND: scrollBar ***
void EditBuffer::cmd_scrollBar( void )
{
	if ((text == NULL) || (text->nLines == 0)) return;  // Uninitalized widget

	if ((text->iMaxLineLength >= 0) &&
	    (text->iMaxLineLength > text->size().width()/text->fontwidth))
	{
		if (hscroll == NULL)
		{
			hscroll = new QScrollBar( 0, text->iMaxLineLength, 1,
			                          text->size().width() / text->fontwidth, 0,
			                          QScrollBar::Horizontal, bufferview_v);
			QObject::connect( hscroll, SIGNAL(valueChanged(int)), 
					    this, SLOT( hscrollEvent(int) ) );
			hscroll->show();
		} else
		{
			int iPageSize = text->size().width() / text->fontwidth;
			hscroll->setMaxValue(text->iMaxLineLength - iPageSize);
			hscroll->setPageStep(iPageSize);
		}
	} else if ((hscroll != NULL) && (text->iMaxLineLength == 0))
	{
		/*childs.remove(hscroll);
		hscroll->close(true);
		hscroll = NULL;*/
	}

	if ( text->lines.size() > (unsigned int)text->nLines ) 
	{
		if ( vscroll == NULL )
		{
			vscroll =  new QScrollBar(
				0, text->lines.size()-text->nLines, 1, text->nLines-1, 0,
				QScrollBar::Vertical, bufferview);
			childs.push_back(vscroll);
			QObject::connect( vscroll, SIGNAL(valueChanged(int)), 
					    this, SLOT( scrollEvent(int) ) );
			QObject::connect( vscroll, SIGNAL(sliderReleased( void )), 
					    this, SLOT( scrollRelease( void ) ) );
		} else
		{
			vscroll->setMaxValue(text->lines.size());
			vscroll->setPageStep(text->nLines);
		}
		vscroll->show();
	} else if (vscroll != NULL) {
		/* remove the vertical scroll bar */
		vector<QWidget *, allocator<QWidget *> >::iterator oItem;
		for(oItem = childs.begin(); oItem != childs.end(); oItem ++)
			if ( *oItem == vscroll )
			{
				childs.erase( oItem );
				break;
			}
		vscroll->close(true);
		vscroll=NULL;
	}
}

// * COMMAND: scrollTo ***
// Scroll to a specified line number
void EditBuffer::cmd_scrollTo( int first, int iPlacement, bool optimizing )
{
	if (iPlacement != -1)
		first -= (text->nLines * iPlacement) / 100;
	first=max( 0, first );
	first=min( first, (int)text->lines.size()-1 );
	if (optimizing) text->scroll( first );
	else text->repaint();
	linebar->repaint(false);
}

// * COMMAND: scrollToCursor ***
void EditBuffer::cmd_scrollToCursor(int iPlacement, bool optimizing )
{
	/*calculate the new position of the cursor */
	int first = text->firstVisible;
	bool scrolled = false;
	if ( text->cursor.y()-first >= text->nLines ) {
		first = text->cursor.y()-text->nLines+1;
		text->repaintfrom=text->repaintto=-1;
		if (vscroll) vscroll->setValue( first );
		scrolled = true;
	}
	if ( text->cursor.y()<first ) {
		first = text->cursor.y();
		text->repaintfrom=text->repaintto=-1;
		if (vscroll) vscroll->setValue( first );
		scrolled = true;
	}
	if ( text->cursor.x() < hscrollValue )
	{
		hscrollValue = text->cursor.x();
		text->repaintfrom=text->repaintto=-1;
		if (hscroll) hscroll->setValue( hscrollValue );
	}
	if ( hscrollValue + text->size().width() / text->fontwidth < (text->cursor.x() + 1))
	{
		hscrollValue = text->cursor.x() - text->size().width() / text->fontwidth + 1;
		text->repaintfrom=text->repaintto=-1;
		if (hscroll) hscroll->setMaxValue( hscrollValue );
		if (hscroll) hscroll->setValue( hscrollValue );
	}

	if (optimizing) text->scroll( first );
	else text->repaint();

	if (scrolled) linebar->repaint(false);
}

// * COMMAND: setCursor ***
// Sets the cursor's coordinates
void EditBuffer::cmd_setCursor( int nx, int ny )
{
	text->cursor.setX( (nx-5)/text->fontwidth + hscrollValue );
	text->repaintfrom = text->cursor.y();
	text->cursor.setY( ny/text->fontheight+text->firstVisible );
	text->repaintto = text->cursor.y();
	if (text->repaintto < text->repaintfrom)
	{
		ny = text->repaintto;
		text->repaintto = text->repaintfrom;
		text->repaintfrom = ny;
	}
	text->repaint(false);
}

// * COMMAND: selection_copy ***
// selection: copy to the system's clipboard
void EditBuffer::cmd_selection_copy( QString *pCopiedString )
{
	int sel1x = text->select[text->selectsmaller].x(),
		sel1y = text->select[text->selectsmaller].y(),
		sel2x = text->select[1-text->selectsmaller].x(),
		sel2y = text->select[1-text->selectsmaller].y();
	bool bReturn = false;
	EditLine *sLine;
	int i;

	bool bInternalCopiedString = (pCopiedString == NULL);
	if (bInternalCopiedString) pCopiedString = new QString();

	/* 1. create new operation */
	oUndo.createOperation("SELECTION_COPY");

	/* 2. store the current selection */
	oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *(text->sCurrentSelectionColorName),
		text->select[0].x(), text->select[0].y(), text->select[1].x(), text->select[1].y());

	if ((sel1y==-1) || (sel2y==-1) || (sel1x==-1) || (sel2x==-1))
		bReturn = true;

	if (!bReturn && (sel1x == sel2x) && (sel1y == sel2y) )
		bReturn = true;

	if (!bReturn)
		for( sLine=text->lines[sel1y], i=sel1y; 
					(i<(int)text->lines.size()) && (i<=sel2y); 
					i++, sLine=text->lines[i] ) 
		{
			int n1, n2;
			if ( i==sel1y ) n1=sLine->tabbedPosition( sel1x, tabsize );
			else n1 = 0;
			if ( i==sel2y ) n2=sLine->tabbedPosition( sel2x, tabsize );
			else n2 = sLine->length();
			pCopiedString->append( sLine->mid( n1, n2-n1 ) );
#ifdef _SYSTEM_LINUX_
			if ( i<sel2y ) pCopiedString->append( "\n" );
#elif defined(_SYSTEM_WINDOWS_)
			if ( i<sel2y ) pCopiedString->append( "\r\n" );
#endif
		}

	if (!bReturn)
	{
		QApplication::clipboard()->setText( *pCopiedString );

		/* change the color of the selection: copy is done */
		oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, text->sSelectionCopiedColorName,
			text->select[0].x(), text->select[0].y(), text->select[1].x(), text->select[1].y());
	}

	/* n. close operation */
	if (!bReturn)
		oUndo.closeOperation();

	/* cleanup */
	if (bInternalCopiedString) delete(pCopiedString);
}

// * COMMAND: selection_paste ***
// selection: paste from the system's clipboard
void EditBuffer::cmd_selection_paste( bool bClipBoard )
{
	QString clip; /*!*/

	/* 1. create new operation */
	oUndo.createOperation("SELECTION_PASTE");

	/* 2. fill until the cursor */
	text->cmd_text_filltocursor();
	setModified( true );

	EditLine *l = text->lines[ text->cursor.y() ];

	/* 3. store the original selection state */
	oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *text->sCurrentSelectionColorName,
		text->select[0].x(), text->select[0].y(), text->select[1].x(), text->select[1].y());

	/* 4. prepare for selecting the new text */
	oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, QString::null,
		text->cursor.x(), text->cursor.y(), text->select[1].x(), text->select[1].y());

	/* 5. paste core code */
#ifndef _SYSTEM_WINDOWS_ /*!*/
	if ((!bClipBoard) && QApplication::clipboard()->supportsSelection())
		QApplication::clipboard()->setSelectionMode(TRUE);
#endif
	clip = QApplication::clipboard()->text();
#ifndef _SYSTEM_WINDOWS_ /*!*/
	if ((!bClipBoard) && QApplication::clipboard()->supportsSelection())
		QApplication::clipboard()->setSelectionMode(FALSE);
#endif

	clip.replace(QRegExp("\r"), "");

	/* 5.1 store the left part of the actual line */
	QString se = l->mid( l->tabbedPosition(text->cursor.x(), tabsize) );
	dmesg("6:command", "selection.paste: %s\n", se.latin1() );
	/* 5.2 delete the right part of the actual line */
	oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING, *new QString(se),
		l->tabbedPosition(text->cursor.x(), tabsize), text->cursor.y());

	/* 5.3 paste the next part of the selection */
	int n, nb;
	for( n=clip.find('\n', nb=0); n!=-1;
							nb=n+1, n=clip.find('\n', nb) ) {
		/* 5.3.1 append the current segment to the current line */
		oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(clip.mid( nb, n-nb )),
			text->lines[text->cursor.y()]->length(), text->cursor.y());
		/* 5.3.2 move the cursor one line down */
		oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
			text->cursor.x(), text->cursor.y()+1, text->cursor.x(), text->cursor.y());
		/* 5.3.3 create new line */
		oUndo.doAtomicOperation(AtomicTextOperation::CREATELINE, QString::null, 0, text->cursor.y());
	}
	l=text->lines[ text->cursor.y() ];
	oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(clip.mid( nb )),
		text->lines[text->cursor.y()]->length(), text->cursor.y());
	oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
		l->nontabbedPosition( l->length(), tabsize ), text->cursor.y(), text->cursor.x(), text->cursor.y());
	oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *new QString(se),
		text->lines[text->cursor.y()]->length(), text->cursor.y());
	oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, text->sSelectionPastedColorName,
		text->select[0].x(), text->select[0].y(), text->cursor.x(), text->cursor.y());

	/* 1. close operation */
	oUndo.closeOperation();

}

/**
 * COMMAND: selection_clear  - clears the current selection
 *
 *  <string1>xxxxxxxxxxxxx
 *  zzzzzzzzzzzzzzzzzzz
 *  yyyyyyyyyy<string2>
 */
void EditBuffer::cmd_selection_clear( void )
{
	int sel1x = text->select[text->selectsmaller].x(),
		sel1y = text->select[text->selectsmaller].y(),
		sel2x = text->select[1-text->selectsmaller].x(),
		sel2y = text->select[1-text->selectsmaller].y();
	int i;
	bool bReturn = false;

	setModified( true );

	/* 1. COND is it NOT a legal selection? return */
	if ((sel1y==-1) || (sel2y==-1) || (sel1x==-1) || (sel2x==-1)
	  || ((sel1x == sel2x) && (sel1y == sel2y)))
		bReturn = true;

	/* 2. create new operation */
	if (!bReturn)
		oUndo.createOperation("SELECTION_PASTE");

	/* 3. store the actual selection */
	if (!bReturn)
		oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, *text->sCurrentSelectionColorName,
			text->select[0].x(), text->select[0].y(), text->select[1].x(), text->select[1].y());

	EditLine *sString1;
	EditLine *sXXXXXXX;
	if (!bReturn)
	{
	/* 4. cut <string1> (delete xxxxxxxxxxxx) */
		sString1 = text->lines[ sel1y ];
		sXXXXXXX = new EditLine(*sString1);
		sXXXXXXX->remove(0, sString1->tabbedPosition(sel1x, tabsize));

	/* 5. append <string2> to <string1> (not including yyyyyyyyyyyy) */
		EditLine *sString2 = new EditLine(text->lines[sel2y]->mid(
			text->lines[sel2y]->tabbedPosition( sel2x, tabsize )));
		oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING, *sXXXXXXX,
			sString1->tabbedPosition(sel1x, tabsize), sel1y);
		oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *sString2,
			sString1->length(), sel1y);

	/* 6. remove the line "yyyyyyyyyy<string2>" if it is not a one-line-selection */
		if (sel1y != sel2y)
			oUndo.doAtomicOperation(AtomicTextOperation::DELETELINE, *new EditLine(*text->lines[sel2y]),
				0, sel2y);

	/* 7. ITER remove the "zzzzzzzzzzzzzzzzzzz" lines */
		for( i=sel1y+1; (sel1y+1<(int)text->lines.size()) && (i<sel2y); i++)
			oUndo.doAtomicOperation(AtomicTextOperation::DELETELINE,
			*new EditLine(*text->lines[sel1y+1]), 0, sel1y+1);

		/*? if ( *s == "" ) text->lines.remove( sel1y );*/

	/* 8. store the new selection */
		oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, text->sSelectionCopiedColorName,
			-1, -1, -1, -1);

	/* 9. move the cursor to the beginning of the selection */
		oUndo.doAtomicOperation(AtomicTextOperation::MOVECURSOR, QString::null,
			sel1x, sel1y, text->cursor.x(), text->cursor.y());
	}

	/*? 10. repaint (is it necessary?) */
	if (!bReturn)
	{
		text->selecting = false;

		text->repaintfrom = text->repaintto = sel1y;
		if (sel1y != sel2y) text->repaintto = -1;

		text->repaint(false);
	}
}

void EditBuffer::cmd_selection_wrap( const char *sType )
{
	bool bReturn = false;
	QString sCopiedLines;

	/* 1. get the wrap value */
	int iWrap = EditBufferText::properties->getNumericalValue("buffer.preferred_width", -1);
	if (iWrap == -1) bReturn = true;

	bool iCursXZero = (text->cursor.x() == 0);

	/* 2. get the text to be wrapped */
	/* 2.1 copy the selected text to a buffer */
	cmd_selection_copy(&sCopiedLines);
	/* 2.2 simplify the text (remove redundant spaces) */
	QString oLines = sCopiedLines.simplifyWhiteSpace();
	/* 2.3 cut out the selection from the buffer */
	cmd_selection_clear();

	if (iCursXZero) text->lines.insert(text->lines.begin() + text->cursor.y(),
	                          new EditLine(""));

	text->select[0].setX( text->cursor.x() );
	text->select[0].setY( text->cursor.y() );

	/* 3. remove EOL's from the selected text; insert spaces if words are sticked together */
	for(int ci = 0; ci != -1;)
	{
		ci = oLines.find('\n', ci);
		if (ci == -1) ci = oLines.find('\r', ci);
		if (ci != -1)
		{
			oLines.remove(ci, 1);
			if ((!oLines.at(ci).isSpace()) &&
			    ((ci >= (int)oLines.length() - 1) ||
			     (!oLines.at(ci + 1).isSpace())) &&
			    ((ci == 0) || (!oLines.at(ci - 1).isSpace())))
			oLines.insert(ci, " ");
		}
	}

	for(int ci2 = 0; ci2 < (int)oLines.length(); ci2++)
	{
		const QString &oChar = oLines.mid(ci2, 1);
		text->cmd_text_insertchar(oChar, true);
	}
//		text->cmd_text_insertchar(QString(oLines.at(ci2)), true);

	text->select[1].setX( text->cursor.x() );
	text->select[1].setY( text->cursor.y() );

	if (!strcmp(sType, "justify"))
		for(int iL = text->select[0].y(), iC = text->select[0].x();
		        iL < (int)text->select[1].y(); iL++, iC = 0)
		{
			EditLine *pLine = text->lines[iL];
			float fQuot = iWrap;
			fQuot /= pLine->nontabbedPosition(pLine->length());
			if ( fQuot > 1)
			{
				for(int iS = 0; iC < (int)pLine->length() - 1; iC++)
				{
					if (pLine->at(iC).isSpace() &&
					      !pLine->at(iC+1).isSpace() &&
					      ((iC - iS) * fQuot > iC))
					{
						pLine->insert(iC, " ");
						iS++;
					}
				}
			}
		}
	else if (!strcmp(sType, "right"))
	{
		for(int iL = text->select[0].y(), iC = text->select[0].x();
		        iL <= (int)text->select[1].y(); iL++, iC = 0)
		{
			EditLine *pLine = text->lines[iL];
			QString oNew = pLine->rightJustify(iWrap);
			(*(QString *)pLine).operator=(oNew);
		}
		text->select[1].setX(iWrap);
		text->cursor.setX(iWrap);
	}

	oUndo.doAtomicOperation(AtomicTextOperation::SETSELECTION, text->sSelectionCopiedColorName,
		text->select[0].x(), text->select[0].y(), text->select[1].x(), text->select[1].y());
}

/* * COMMAND: find(.) ***
 * this is the standard find command
 * parameteres:
 *   bOpen: 1 - create or activate a find box
 *          0 - destroy the pointer to the box
 */
void EditBuffer::cmd_find( bool bOpen )
{
	if (bOpen)
	{
		if (findbox == NULL)
		{      /* there is no find box in the buffer */
			findbox = (QWidget *)new EditBufferFind( text, statusbar );
		} else
		{      /* a find box (probably frame) is present */
			dwarn("buffer", "Weird event - buffer has focus while searching");
			findbox->setFocus();
		}
	} else findbox = NULL;
}

#include <qmessagebox.h>
// * COMMAND: cmd_checkFile *** 
/**
 *  @brief check the file in the disk if it is newer than the loaded file
 */
void EditBuffer::cmd_checkFile( void )
{
	static QMessageBox *pCheckWindow = NULL;
	if (bLoading) return;
	if (file == NULL) return;
	QFileInfo fi( *file );
	if (pCheckWindow != NULL) return;
	if ((fi.lastModified() > file_lastmodified) && (pCheckWindow == NULL) )
	{
		text->repaintfrom = text->repaintto = -1;
		text->paintEvent( NULL );
		pCheckWindow = new QMessageBox("Freddy", tr("File is modified. <b>Reload</b> it?"),
		               QMessageBox::Warning, QMessageBox::Yes|QMessageBox::Default,
		               QMessageBox::No|QMessageBox::Escape, QMessageBox::NoButton,
//		               this, tr("File modified"), true);
		               NULL, tr("File modified"), true);
	}
	if (pCheckWindow != NULL)
	{
		pCheckWindow->setFocus();
		if( pCheckWindow->exec() == QMessageBox::Yes )
		{
			command("reload", NULL, NULL, NULL);

			if (getFile())
				loadFile(*getFile(), false);
			else
				derr("io", "Unable to access the current file");
		}
		pCheckWindow = NULL;
	}
	file_lastmodified = fi.lastModified();
}

// * COMMAND: cmd_gotoLine *** 
/**
 *  @brief Ask a number from the user and jumps to the appropriate line
 */
void EditBuffer::cmd_gotoLine()
{
	/* void yet */
}

// COMMAND: cmd_undo
/**
 *  @brief Undo the previous text operation
 */
void EditBuffer::cmd_undo(bool bMicro)
{
	oUndo.undoLastOperation(bMicro);
}
void EditBuffer::cmd_redo(bool bMicro)
{
	oUndo.redoLastOperation(bMicro);
}

// ATOMIC_OPERATION: atomic_insertString(...)
/**
 *  @brief Insert a string into the appropriate line
 */
void EditBuffer::atomic_insertString(QString *sString, int iColumn, int iLine)
{
	bool bReturn = false;

	/* ERROR_HANDLE: not valid line number */
	if ((iLine < 0 ) || (iLine >= (int)(text->lines.size())))
	{
		derr("buffer", "insertString %d is outside the borders (lines in the buffer: %d)",
			iLine, text->lines.size());
		bReturn = true;
	}

	/* insert the string into the line */
	if (!bReturn)
	{
		EditLine *pLine = text->lines[iLine];
		if (pLine)
			pLine->insert(iColumn, *sString);
	}
}

// ATOMIC_OPERATION: atomic_deleteString(...)
/**
 *  @brief Deletes a substring
 */
void EditBuffer::atomic_deleteString(int iColumn, int iLine, int iLength)
{
	bool bReturn = false;

	/* ERROR_HANDLE: not valid line number */              	
	if ((iLine < 0 ) || (iLine >= (int)(text->lines.size())))
	{
		derr("buffer", "deleteString %d is outside the borders (lines in the buffer: %d)",
			iLine, text->lines.size());
		bReturn = true;
	}

	/* remove the string from the line */
	if (!bReturn)
	{
		EditLine *pLine = text->lines[iLine];
		if (pLine)
			pLine->remove(iColumn, iLength);
	}
}

// ATOMIC_OPERATION: atomic_createLine(.)
/**
 *  @brief Creates a new line
 */
void EditBuffer::atomic_createLine(int iLine)
{
	bool bReturn = false;

	/* 1. CHECK (iLine < 0)? -> return */
	if (iLine < 0)
	{
		derr("buffer", "EditBuffer::createLine %d is has illegal parameter (below 0)!", iLine);
		bReturn = true;
	}

	/* 2. if we are before the end of the file, instert the line into the file */
	if ((!bReturn) && (iLine < (int)(text->lines.size())))
	{
		text->lines.insert(text->lines.begin() + iLine, new EditLine(QString::null));
		bReturn = true;
	}

	/* 3. if we are at or after the end of the file, append the line to the file */
	if (!bReturn)
	{
		while(iLine >= (int)(text->lines.size()))
			text->lines.push_back(new EditLine(""));
	}
}

// ATOMIC_OPERATION: atomic_deleteLine(.)
/**
 *  @brief Deletes a line
 */
void EditBuffer::atomic_deleteLine(int iLine)
{
	bool bReturn = false;

	if (iLine < 0)
	{
		
		derr("buffer", "EditBuffer::deleteLine %d is has illegal parameter (below 0)!", iLine);
		bReturn = true;
	}

	if ((!bReturn) && (iLine < (int)(text->lines.size())))
	{
		text->lines.erase(text->lines.begin() + iLine);
		bReturn = true;
	}
}

// ATOMIC_OPERATION: atomic_moveCursor(..)
/**
 *  @brief Moves the cursor to a new location
 */
void EditBuffer::atomic_moveCursor(int iColumn, int iLine)
{
	text->cursor.setX(iColumn);
	text->cursor.setY(iLine);
}

// ATOMIC_OPERATION: atomic_setSelection(....)
/**
 *  @brief Sets the selected text in the buffer
 *  It is assumed that there can be only one selected text in the buffer
 */
void EditBuffer::atomic_setSelection(QString &sSelectionName, int iColumn1, int iLine1, int iColumn2, int iLine2)
{
	text->sCurrentSelectionColorName = &sSelectionName;
	text->select[0].setX(iColumn1);
	text->select[0].setY(iLine1);
	text->select[1].setX(iColumn2);
	text->select[1].setY(iLine2);
}

#include <qfontdialog.h>
char *EditBuffer::command(const char *commandstr,
			 const char *arg1, const char *arg2, const char *arg3 )
{
	char **command;
	int comnum;

	if (strstr(commandstr, "CHECK ") == commandstr)
		return (char *)1;

	enum { CM_SCROLLTO, CM_SCROLLBAR, CM_SCROLLBY, CM_SCROLLTOCURSOR,
		   CM_SETCURSOR, 
		   CM_SELECTCOPY, CM_SELECTPASTE, CM_SELECTCLEAR,
		   CM_SELECTWRAP,
	       CM_CURRENTWORD,
	       CM_UNDO, CM_REDO,
		   CM_FIND, CM_CHECKFILE, CM_GOTOLINE, CM_GETFONT,
		   CM_RELOAD, CM_CONVERTTABS };
	static char *comcache[]= {
	       "scrollto", "scrollbar", "scrollby", "scrolltocursor",
	       "setcursor",
	       "selectcopy", "selectpaste", "selectclear",
	       "selectwrap",
	       "currentword",
	       "undo", "redo",
	       "find", "checkfile", "gotoline", "getfont",
	       "reload", "converttabs", NULL };

	dmesg("4:buffer", "buffer command: %s(%s,%s,%s)", commandstr, arg1, arg2, arg3);

	for(command=comcache, comnum=0; (*command)!=NULL; command++, comnum++)
		if (0 == strcmp( commandstr, *command )) break;

	if (*command == NULL) 
	  { dwarn("buffer", "Invalid command: %s!", commandstr); return NULL; }

	switch( comnum )
	{
		case CM_SCROLLTO:
			cmd_scrollTo( atoi(arg1) );
			break;
		case CM_SCROLLBAR:
			cmd_scrollBar();
			break;
		case CM_SCROLLTOCURSOR:
			cmd_scrollToCursor();
			break;
		case CM_SETCURSOR:
			cmd_setCursor( atoi(arg1), atoi(arg2) );
			break;
		case CM_SELECTCOPY:
			cmd_selection_copy();
			cmd_scrollBar();
			text->repaint();
			break;
		case CM_SELECTPASTE:
			cmd_selection_paste((arg1==NULL)||(strcmp(arg1,"selection")) ? TRUE : FALSE);
			cmd_scrollBar();
			text->repaint();
			break;
		case CM_SELECTCLEAR:
			cmd_selection_clear();
			cmd_scrollBar();
			text->repaint();
			break;
		case CM_SELECTWRAP:
		    cmd_selection_wrap(arg1);
			text->repaint();
		    break;
		case CM_CURRENTWORD:            /* returns with the word under the cursor */
			{
				int i1, i2;
				if (text->cursor.y() < 0) return NULL;
				if (text->cursor.y() >= (int)text->lines.size()) return NULL;
				EditLine *line = text->lines[text->cursor.y()];
				if (line == NULL) break;
				int posx = line->tabbedPosition(text->cursor.x(), tabsize, false);
				for(i1 = posx; i1 >= 0; i1--)
				{
					if (!line->at(i1).isLetter() && (line->at(i1) != '_')) break;
				}
				for(i2 = posx; i2 < (int)line->length(); i2++)
				{
					if (!line->at(i2).isLetter() && (line->at(i2) != '_')) break;
				}
				i1++; i2--;

				const QString &sRet = line->mid(i1, i2-i1+1);
				return (!sRet.isNull() ? strdup(sRet.latin1()) : NULL);
			}
		case CM_UNDO:
			{
				bool bMicro = false;
				if (arg1 != NULL) bMicro = !strcmp(arg1, "micro");
				cmd_undo(bMicro);
			}
			break;
		case CM_REDO:
			{
				bool bMicro = false;
				if (arg1 != NULL) bMicro = !strcmp(arg1, "micro");
				cmd_redo(bMicro);
			}
			break;
		case CM_FIND:
			cmd_find();
			break;
		case CM_CHECKFILE:
			cmd_checkFile();
			break;
		case CM_GOTOLINE:
			cmd_gotoLine();
			break;
		case CM_GETFONT:
			{
				bool ok;
				QFont fnt=QFontDialog::getFont(&ok);
				if (ok)
				{
					fnt.setFixedPitch( true );
					QFontMetrics *fm = new QFontMetrics( fnt );
					text->fontwidth = fm->width('M');
					text->fontheight = fm->height();
					text->fontheight +=
					  text->properties->getNumericalValue("buffer.font.gap", 2);
					text->fontascent = fm->ascent();
					text->nLines = text->size().height() / text->fontheight;
					text->setFont( fnt );
				}
			}
			break;
		case CM_RELOAD:
			{
				QFile *pFile = getFile();

				if (pFile)
				{
					loadFile(*getFile(), pFileTypeCombo->currentItem());
					oUndo.setSavedPosition();
				}
				else
					derr("io", "Unable to access the current file (this file was not saved)");

				break;
			}
		case CM_CONVERTTABS:
			{
				/* get the input parameter - size of the tabs (in spaces) */
				bool bOK;
				QString sNum(arg1), sTabSpaces;
				int iTabSizenum = sNum.toInt(&bOK);
				if (bOK)
				{
				printf("converttabs: %d %d '%s'\n", iTabSizenum, bOK, sTabSpaces.latin1());

					oUndo.createOperation("TRANSLATE_TABS");
					bool bTabMode;
					for(int iLine = 0; iLine < text->lines.size(); iLine++)
					{
						EditLine *s = new EditLine(*(text->lines[ iLine ]));
						for(int j = 0, bTabMode = false; j < s->length(); j++)
						{
							if ((j % iTabSizenum) == 0) bTabMode = false;
							if (((*s)[j] == '\t') && !bTabMode)
							{
								bTabMode = true;
								s->remove(j, 1);
							}
							if (bTabMode) s->insert(j, " ");
						}
						//s->replace("\t", sTabSpaces);
						oUndo.doAtomicOperation(AtomicTextOperation::DELETESTRING,
							*new EditLine(*(text->lines[ iLine ])), 0, iLine);
						oUndo.doAtomicOperation(AtomicTextOperation::INSERTSTRING, *s, 0, iLine);
					}
					oUndo.closeOperation();
				}
			}

	}

	if ((comnum >= CM_SELECTPASTE) && (comnum <= CM_SELECTCLEAR))
		reconfigure();

	return NULL;
}

