/*
 * EditLine.cpp
 *
 * Line with wrapping capability
 * 
 * Author: Ferenc Moricz
 * $Revision: 1.26 $
 * $Date: 2004/02/22 00:12:50 $
 *
 */

#include "Debug.h"
#include "EditLine.h"

#include <stdlib.h>

EditLine::EditLine(const QString &text) : QString( text )
{
	if ( at(length()-1)=='\n' ) truncate(length()-1);
	forward = -1;
}
 
/*
 * Example:
 *     tabbedPosition("\t\tapple", 2, 4, false) -> 0
 */
int EditLine::tabbedPosition( QString &sStr, int query, int tabsize, bool jumpthrough )
{
	int position, ch;

	for(ch=position=0; ch<(int)sStr.length(); ch++)
	{
		int jump=(sStr.at(ch)=='\t' ? (position/tabsize+1)*tabsize-position : 1);
		if (position+jump <= query) position+=jump;
		else if ( (sStr.at(ch)=='\t') && (position<query) )  
			{ position+=jump; ch+=jumpthrough; break; } // Jump <tab>
		else break;
	}
	if (ch>=(int)sStr.length()) ch+=query-position;
	return ch;
}

int EditLine::tabbedPosition( int query, int tabsize, bool jumpthrough )
{
	return tabbedPosition(*this, query, tabsize, jumpthrough);
}


/*
 * Example:
 *     nontabbedPosition("\t\tapple", 2, 4) -> 8
 */
int EditLine::nontabbedPosition( QString &sStr, int query, int tabsize )
{
	int position, ch;
	for(ch=position=0; (ch<(int)sStr.length()) && (ch<query); ch++)
		position=( sStr.at(ch)=='\t' ? (position/tabsize+1)*tabsize:position+1 );
	return position;
}

int EditLine::nontabbedPosition( int query, int tabsize )
{
	return nontabbedPosition( *this, query, tabsize );
}

EditLine *EditLine::splitLine( int iPosition, int iTabSize )
{
	int iNTPos = nontabbedPosition(iPosition);

	EditLine *pNewLine = new EditLine( mid( iNTPos ) );
	truncate( iNTPos );

	return pNewLine;
}

// .gotoWord(..)
/**
 * Returns the next word boundary
 *
 * @param iPos the current x position
 *           (if -1 then this is a continued search from the previously processed line)
 * @param bRight the direction of the step (true - forward, false - backward)
 * @param bWordEndAlso jump to the word endings also
 *           (the default is to jump only to the first character of a word)
 * @return
 *         - the next (previous) word boundary
 *         - -1 if there are no more word boundaries in this line
 */
int EditLine::gotoWord(int iPos, bool bRight, bool bWordEndAlso, bool bSplitAtSign)
{
	bool bFOUND = false;       ///< true if we have found the next position
	bool bRETURN = false;      ///< true if there was an error, and we have to return
	bool bFirstCycle = true, bWasSpace, bIsSpace;

	/* >>>FORWARD>>> beginning conditionals */
	if (!bRETURN && !bFOUND && bRight)
	{
		/* it is a continued search */
		if (iPos == -1)
		{
			iPos = 0;
			bFOUND = true;
		}
		/* we are already after the end of the line -> there is no more word */
		else if (iPos >= ((int)length())) bRETURN = true;
	}

	/* <<<BACKWARD<<< beginning conditionals */
	if (!bRETURN && !bFOUND && !bRight)
	{
		/* it is a continued search */
		if ((iPos == -1) || (iPos > ((int)length())))
		{
			iPos = length() - 1;
			bFOUND = true;
		}
		/* we are already at the beginning of the line -> there is no more word */
		else if (iPos == 0) bRETURN = true;
		else iPos --;
	}

	/* <<<BACKWARD<<< and we are at the line end */
	if (!bRETURN && !bFOUND && !bRight && (iPos == ((int)length()))) iPos --;

	/* going forward or backward from char to char 
	   until a <space><letter> or <letter><space> sequence is not found */
	while(!bRETURN && !bFOUND && (iPos >= 0) && (iPos < (int)(length())))
	{
		/* We are at 'iPos'. Are we at a border? */
		if (bSplitAtSign)
			bIsSpace = !(at(iPos).isLetter() || at(iPos).isDigit() || (at(iPos) == '_'));
		else
			bIsSpace = at(iPos).isSpace();

		if (!bFirstCycle)
		{
			if ( (bWasSpace && !bIsSpace && (bRight || bWordEndAlso)) ||
			     (!bWasSpace && bIsSpace && (!bRight || bWordEndAlso)) )
			{
				bFOUND = true;
				break;
			}
		} else bFirstCycle = false;

		if (bRight) iPos ++; else iPos --;
		bWasSpace = bIsSpace;
	}

	if (bRETURN) return -1;
	else if (bFOUND) return (bRight ? iPos : iPos + 1);
	else if (iPos == -1) return 0;
	else return (bRight ? length() : 0);
}


// . appendString (..)
/**
 * Append a string to the current line. It can be an UTF-8 string also
 */
void EditLine::appendString( unsigned char *sStr, bool bUTF8 )
{
	if (! bUTF8)
		append((char *)sStr);
	else append(QString::fromUtf8((const char *)sStr));
#if 0
	else
	for(int iPos = 0; (iOctet = sStr[iPos]) != '\0'; iPos++)
	{
		if ((iOctet & 0xe0) == 0xc0)   /* continuing octet */
		{
			iOctLen --;
			if (iOctLen >= 0)
			{
				iOctSeq <<= 6;
				iOctSeq |= (iOctet & 0x3f);
			}
		}
		else
		{
			if (iOctLen != 0)          /* unclosed sequence -> error */
			{ 
				bReturn = false;
				break;
			}
			if (iOctet & 0x80)
			{                          /* first octet of an octet sequence */
				/* calculate the length of the octet sequence */
				for(; iOctet & 0x80; iOctLen ++, iOctet <<= 1);
				/* this is the first octet; so we decrement the counter */
				iOctLen --;
			}
		}
	}

	if (iOctLen != 0 ) bReturn = false;

	return bReturn;
#endif
}

// . canBeUTF8 (.)
/**
 * Checks a string if it can be an UTF-8 codec string or not. (syntactical check)
 */
bool EditLine::canBeUTF8(unsigned char *sStr)
{
	int iOctLen = 0;
	bool bReturn = true;
	int iOctet;

	for(int iPos = 0; (iOctet = sStr[iPos]) != '\0'; iPos++)
	{
		if ((iOctet & 0xc0) == 0x80)   /* continuing octet */
		{
			iOctLen --;
			if (iOctLen < 0)           /* too long sequence */
			{
				dmesg("buffer", "Too long UTF octet sequence about %c%c%c\n\tline: %s",
					sStr[iPos],sStr[iPos+1],sStr[iPos+2], sStr);
				bReturn = false;
				break;
			}
		}
		else
		{
			if (iOctLen != 0)          /* unclosed sequence -> error */
			{ 
				dmesg("buffer", "Unclosed UTF tag about %c%c%c\n\tline: %s",
					sStr[iPos-1], sStr[iPos], sStr[iPos+1], sStr);
				bReturn = false;
				break;
			}
			if (iOctet & 0x80)
			{                          /* first octet of an octet sequence */
				/* calculate the length of the octet sequence */
				for(; iOctet & 0x80; iOctLen ++, iOctet <<= 1);
				/* this is the first octet; so we decrement the counter */
				iOctLen --;
			}
		}
	}

	if (iOctLen != 0 ) bReturn = false;

	return bReturn;
}


