/**
 * @file EditProperties.cpp
 * @brief Property management system, implementation
 * 
 * @author Ferenc Moricz
 * $Revision: 1.37 $
 * $Date: 2004/02/22 00:12:50 $
 *
 */

/**
 * @class EditProperties
 * @brief The class, which stores the program properties in the 
 * memory and loads them and saves them to the config file.
 *
 * The class stores the key/valuelist pairs in its QDict structure. 
 * The user can look at the value of a propertytag by calling the 
 * \e keyValue() function. There are functions to provide automatic 
 * conversions to \e bool, \e int and \e QColor types
 * (\e getBoolValue(), \e getNumericalValue(), \e getColorValue() ).\n
 * The properties can be load from a file with \e loadConfig()
 * and they can be saved in a file with \e saveConfig().
 *
 * <b>Config files</b>
 *
 * The config files has a specific format. It mainly consist of 
 * variables and their values.
 *
 * Special characters:
 *   - '#' (comment) - the text after it is a comment
 *   - '\' (linebreak) - the line continues in the next line.
 *         Anything can be after this, it will not be mentioned.\n
 *   - '=' (set) - setting a variable's value
 */

#include "EditProperties.h"
#include "Editor.h"

#include <qcolordialog.h>
#include <qfile.h>
#include <qtooltip.h>
#include <qwhatsthis.h>

EditProperties *EditProperties::pMain;

/**
 * Constructor - saves the address of the main properties item
 */
EditProperties::EditProperties()
{
	pMain = this;
}

// HELPER: token() !DC
/**
 * @brief Gets the 1st, 2nd, ..., n-th word from a string
 */
void EditProperties::token(char *string, int index, int &begin, int &end)
{
	bool quote = false;

	for(begin = index; string[begin]!='\0'; begin++)    // search for beginning
		if ((string[begin] != ' ') && (string[begin] != '\t')) break;

	if (string[begin] == '\0') begin = -1;
	else                                // search for ending
	{
		if (string[begin] == '"')       // "
		{
			quote = true;
			begin++;
		}
		for(end = begin; string[end]!='\0'; end++)
		{
			if ( quote )
			{
				if ((end > begin) && (string[end] == '"'))  //"
				{
					string[end--] = ' ';
					break;
				}
			} else
			{
				if ((string[end] == ' ') || (string[end] == '\t'))
				{
					end --;
					break;
				}
				if ((string[end] == '=') || (string[end] == '\\'))
				{
					if (begin != end) end --;
					break;
				}
			}
		}
	}
}

//
// EditProperties :: loadConfig(.)
//
/**  @brief Load configuration variables from a file
 *   @param configfilename the file to load from
 *
 * This function does opens the file, reads it by line to line,
 * recognizes the tags within it, and inserts the parsed tags
 * into its stucture.
 *
 */
bool EditProperties::loadConfig( const char *configfilename, bool bReload )
{
	if ( configfilename == NULL ) return false;
	config = new QString( configfilename );

	QFile *file = new QFile( configfilename );
	if ( !file->exists() ) return false;
	if (!file->open( IO_ReadOnly ))
	{
		derr("config", "Open config file - could not open for read: %s", configfilename);
		return false;
	}

	// <***DC***>
	char buffer[200];
	QFileInfo fi(*file);
	bool bComment = false;
	QString *pComment = NULL;
	dmesg("4:config", "Config file: %s", fi.absFilePath().latin1());

	/* Loading variables */
	while( !file->atEnd() )
	{
		int length, begpos, endpos, word;
		char name[80], value[150];

		/* Parsing the variables from the line */
		for(word=0, endpos=0; ; )
		{
			if ( pComment == NULL ) pComment = new QString();
			/* Read a new line */
			if ((word == 0) ||
//				((buffer[begpos] == '\\')&&
//				 ((buffer[begpos+1]=='\r')||(buffer[begpos+1]=='\n'))))
				(buffer[begpos] == '\\'))
			{
				length=file->readLine( buffer, 199 );   // read the line
				if (file->atEnd()) break;

				if ((buffer[0] == '#') && (buffer[1] != '#'))     // process comments
				{
					for(endpos = 1; buffer[endpos] != '\0'; endpos ++)
						if  ((buffer[endpos] != ' ') && (buffer[endpos] != '\t')) break;
					if (buffer[endpos] != '\0')
					{
						bComment = true;
						pComment->append(buffer + endpos);
					}
					continue;
				}

				while((length>0) && (buffer[length-1]==10)||(buffer[length-1]==13))
					buffer[length-1] = '\0';            // skip EOLs
				endpos = 0;
			}

			fflush(stdout);
			token(buffer, endpos, begpos, endpos);  // get the next word

			if (begpos < 0) break;      // this is the end of the line

//			if ((buffer[begpos] == '\\')&&
//				((buffer[begpos+1]=='\r')||(buffer[begpos+1]=='\n')))
			if (buffer[begpos] == '\\')
				continue;       // newline character - read a new line

			/* process the word */
			if (word == 0)      // first word - it's the name of the tag
			{
				memcpy(name, buffer+begpos, endpos-begpos+1);
				name[endpos-begpos+1] = '\0';

				//vector<PropertyItem,allocator<PropertyItem> >::iterator p = begin();
				if ((find(name) == end()) || bReload)
					(*this)[QString(name)] = *new vector<PropertyItem>();
					//r insert( begin(), *new QString(name), new vector<PropertyItem>() );
				word++;
				dmesg("6:config", "  * %s", name);
			} else              // second and other words - it's the value
			{
				memcpy(value, buffer+begpos, endpos-begpos+1);
				value[endpos-begpos+1] = '\0';

				PropertyItem *pPropItem = new PropertyItem(value);
				if (bComment)
				{
					while((pComment->length() > 0) &&
#if QT_VERSION > 280
					      ((pComment->endsWith("\n")) || (pComment->endsWith("\r"))))
#else
						  ((pComment->at(pComment->length()-1)== '\n') ||
						  ((pComment->at(pComment->length()-1)== '\r'))))
#endif
					{
						pComment->remove( pComment->length() - 1, 1 );
					}
					dmesg("7:config", "comment: - %s:%s\n", name, pComment->latin1());
					pPropItem->pComment = pComment;
					pComment = NULL;
					bComment = false;
				}
				if ( find(name) != end())
				{
					(*this)[name].push_back( *pPropItem );
					word++;
					dmesg("7:config", "      - %s", value);
				}
			}
			endpos++;
		}
	}
	// </***DC***>

	file->close(); 
	return true;
}

//! map< QString, less<QString>,
//!              allocator<vector<PropertyItem,allocator<PropertyItem> > > > storecucc;

//
// EditProperties :: saveConfig(...)
//
/**  @brief Save configuration variables to the specified a file
 *   @param configfilename the file to save to
 *
 */
bool EditProperties::saveConfig( const char *sFileName, vector<QString> *pItems, bool bFlush )
{
	QString savebuffer = "", sOrigBuffer = "";
	QCString sTemp;
	int iSaveTag = 0;

	/* open the config file */
	QFileInfo oFI;
	oFI.setFile(Editor::editor->configdir_user, sFileName);
	QFile *file = new QFile( oFI.absFilePath() );
	if (!file->exists()) 
	{
		oFI.setFile(Editor::editor->configdir_default, sFileName);
		file->setName( oFI.absFilePath() );
	}

	char buffer[200];
	dmesg("4:config", "Saving variables:");

	if (! bFlush)
	{
		if ( !file->exists() ) return false;

		file->open( IO_ReadOnly );
		sTemp = file->readAll();
		sOrigBuffer = sTemp;
		file->close();
		file->open( IO_ReadOnly );

		/* Modify existing config file - load it and change the appr. variables in it */
		while( !file->atEnd() )
		{
			int length, begpos, endpos, i;
			char name[80];

			/* Read a line */
			length=file->readLine( buffer, 199 );

			/* Skip EOLs */
			while((length>0) && ((buffer[length-1]==10)||(buffer[length-1]==13)))
				buffer[length-1] = '\0';

			/* Parsing the variables from the line */
			token(buffer, 0, begpos, endpos);  // get first word

			memcpy(name, buffer+begpos, endpos-begpos+1);
			name[endpos-begpos+1] = '\0';

			/* Copy comments and clean lines */
			if ((begpos < 0) || (buffer[begpos] == '#'))
				savebuffer.append(buffer);

			else
			{
				/* (?) is this property changed by the user */
				vector<PropertyItem,allocator<PropertyItem> > *variable = &(*this)[name];

				for(i = 0; i < (int)pItems->size(); i++)
					if ((*pItems)[i]==name) break;

					/* (Y) write the new value */
				if ((variable) && (i < (int)pItems->size()))
				{
					iSaveTag++;
					savebuffer.append(name);
					savebuffer.append(" ");
					for(int j = 0; j < (int)variable->size(); j++)
					{
						QString *pValue = (*variable)[j].string;
#ifdef _SYSTEM_LINUX
						if (j>0) savebuffer.append(" \\\n    ");
#else						
						if (j>0) savebuffer.append(" \\\r\n    ");
#endif
						if ((pValue->find(" ") != -1) || (pValue->find(",") != -1)
							||(pValue->find("[") != -1) || (pValue->find("(") != -1))
						{
							pValue->prepend("\"");
							pValue->append("\"");
						}
						savebuffer.append(*pValue);
					}
					(*pItems)[i].prepend("-");
				}
					/* (N) write the original line */
				else savebuffer.append(buffer);
			}
			savebuffer.append("\n");
		}
		file->close();
	} else
	{
		for(int i = 0; i < (int)pItems->size(); i++)
			if (!(*pItems)[i].startsWith("-"))
			{
				vector<PropertyItem,allocator<PropertyItem> > *variable = &(*this)[(*pItems)[i]];
				if (variable)
				{
					savebuffer.append((*pItems)[i]);
					savebuffer.append(" ");
					for(int j = 0; j < (int)variable->size(); j++)
					{
						QString *pValue = (*variable)[j].string;
						if (j>0) savebuffer.append(" \\\n    ");
						if ((pValue->find(" ") != -1) || (pValue->find(",") != -1)
							||(pValue->find("[") != -1) || (pValue->find("(") != -1))
						{
							pValue->prepend("\"");
							pValue->append("\"");
						}
						savebuffer.append(*pValue);
					}
				}
				(*pItems)[i].prepend("-");
			}
		savebuffer.append("\n");
	}

	/* write the results to the file */
	if (( !bFlush && (sOrigBuffer != savebuffer) && iSaveTag) ||
	    ( bFlush && (savebuffer.length()>1)))
	{
		oFI.setFile(Editor::editor->configdir_user, sFileName);
		file->setName( oFI.absFilePath() );
		printf("file to save: %s\n",  oFI.absFilePath().latin1());
		if (file->open( IO_WriteOnly ))
		{
			file->writeBlock( savebuffer.latin1(), savebuffer.length() );
		} else printf("Unable to write to users' directory!!!\n");
		file->close();
	}

	return true;
}

//
// PropertyItem
//
// <***DC***>
static int tonumber(char c)
{
  if ((c>='0')&&(c<='9')) return (c-'0'); 
  else if ((c>='A')&&(c<='F')) return (c-'A'+10); 
  else if ((c>='a')&&(c<='f')) return (c-'a'+10);
  else return -1;
}
// <***DC***>

const QString *EditProperties::getComment( QString key )
{
	const PropertyItem *item = getProperty( key, false );
	if ( item == NULL ) return NULL;
	return item->pComment;
}

// 
// EditProperties :: getValue(...)
//
/**  @brief Returns the QString value of the tag 
 *   @param key the QString key specifying the tag
 *   @param defvalue the return value if the tag not found
 *   @param next go to the next value or not?
 *   @return the QString value of the tag
 *
 * It finds the tag specified by \e key, and returns its QString
 * value if it is found, and return \e defvalue if not.
 *
 */
const QString *EditProperties::getValue(
			QString key, const QString *defvalue, bool bCreate )
{
	PropertyItem *item = getProperty( key, bCreate );
	if ( item == NULL ) return defvalue;

	if ((item->bIsDefault) || (!item->string))
		if (defvalue) item->setValue(defvalue->latin1());
	//! QT if (next) list->next();
	return item->string;
}

/* 
 * EditProperties :: getColorValue(...)
 */
/**  @fn getColorValue
 *   @brief Returns the QColor value of the tag 
 *   @param key the QString key specifying the tag
 *   @param defvalue the return value if the tag is not found
 *   @param next go to the next value or not?
 *   @return the QColor value of the tag
 *
 * It finds the tag specified by \e key, and returns its QColor
 * value if it is found or \e defvalue if not.
 *
 */
const QColor *EditProperties::getColorValue( 
				QString key, const QColor *defvalue, bool bCreate )
{
	const PropertyItem *item = getProperty( key, bCreate );
	//! QT if (next) list->next();
	if (item == NULL) return defvalue;
	if ((item->bIsDefault) || (!item->color))
		setColorValue(key, defvalue, false);
	return item->color;
}

/* 
 * .EditProperties :: getFontValue(...)
 */
/**  @fn getFontValue
 *   @brief Returns the QFont value of the tag 
 *   @param key the QString key specifying the tag
 *   @param defvalue the return value if the tag is not found
 *   @param next go to the next value or not?
 *   @return pointer to the QFont value of the tag
 *
 * It finds the tag specified by \e key, and returns its QColor
 * value if it is found or \e defvalue if not.
 *
 */
const QFont *EditProperties::getFontValue( 
				QString key, const QFont *defvalue, bool bCreate )
{
	const PropertyItem *item = getProperty( key, bCreate );
	//! QT if (next) list->next();
	if (item == NULL) return defvalue;
	if ((item->bIsDefault) || (!item->pFont))
		setFontValue(key, defvalue, false);
	return (item->pFont ? item->pFont : defvalue);
}

/* 
 * .EditProperties :: setFontValue(...)
 */
/**  @fn setFontValue
 *   @brief Set the tag's value to "value"
 *   @param key the QString key specifying the tag
 *   @param defvalue the tag's next value
 *   @param next go to the next value or not?
 *   @return true if succeeded
 *
 * It finds the tag specified by \e key, and if found,
 * sets its value to "value".
 *
 */
bool EditProperties::setFontValue( QString &key, const QFont *value, bool next)
{
	QString sNewValue;
	PropertyItem *item = getProperty( key, true);
	if (item == NULL) return false;

	/* setting the appropriate values */
	item->bIsDefault = false;
	if (!item->string) item->string = new QString();
	if (item->pFont == NULL) item->pFont = new QFont();
	*(item->pFont) = *value;
	item->color = NULL;
	item->number = NULL;
	item->logical = NULL;

	/* convert font -> string */

	*(item->string) = "fntfamily=";
	item->string->append(value->family());
	item->string->append(",fntsize=");
	item->string->append(QString::number(value->pointSize()));
	item->string->append(",fntweight=");
	item->string->append(QString::number(value->weight()));
	fflush(stdout);

	return true;
}

// 
// EditProperties :: getBoolValue(...)
//
/**  @brief Returns the boolean value of the tag 
 *   @param key the QString key specifying the tag
 *   @param defvalue the return value if the tag is not found
 *   @param next go to the next value or not?
 *   @return the boolean value of the tag
 *
 * It finds the tag specified by \e key, and returns its boolean
 * value if it is found or \e defvalue if not.
 *
 */

bool EditProperties::getBoolValue(
			QString key, bool defvalue, bool bCreate )
{
	const PropertyItem *item = getProperty( key, bCreate );
	if ( item == NULL ) return defvalue;
	if ((item->bIsDefault) || (!item->logical))
		setBoolValue(key, defvalue, false);
	//! QT if (next) list->next();
	return (item->logical ? *(item->logical) : defvalue);
}

// 
// EditProperties :: getNumericalValue(...)
//
/**  @brief Returns the numerical value of the tag 
 *   @param key the QString key specifying the tag
 *   @param defvalue the return value if the tag is not found
 *   @param next go to the next value or not?
 *   @return the numerical value of the tag
 *
 * It finds the tag specified by \e key, and returns its numerical
 * value if it is found or \e defvalue if not.
 *
 */
int EditProperties::getNumericalValue(
			QString key, int defvalue, bool bCreate)
{
	const PropertyItem *item = getProperty( key, bCreate );
	if ( item == NULL ) return defvalue;

	if ((item->bIsDefault) || (!item->number))
		setNumericalValue(key, defvalue, false);
	//! QT if (next) list->next();
	return *(item->number);
}

/*
 * Writing routines
 */                                      

PropertyItem *EditProperties::getProperty( QString &oKey, bool bCreate )
{
	if (find(oKey) == end())
	{
		if (!bCreate) return NULL;
		(*this)[oKey] = *new vector<PropertyItem, allocator<PropertyItem> >();
	}
	vector<PropertyItem, allocator<PropertyItem> > *pList = &(*this)[oKey];
	if (pList->size() == 0)
	{
		if (!bCreate) return NULL;
		pList->push_back( PropertyItem(NULL) );
		(*pList)[0].bIsDefault = true;
	}
	return &(*pList)[0];
}

bool EditProperties::setValue( QString key, const QString *value, bool next)
{
	PropertyItem *item = getProperty( key, true);
	if (item == NULL) return false;

	item->bIsDefault = false;
	item->number = NULL;
	item->logical = NULL;
	item->color = NULL;
	item->pFont = NULL;
	if (item->string == NULL) item->string = new QString("");
	item->setValue( value->latin1() );
	return true;
}

bool EditProperties::setColorValue( QString key, const QColor *value, bool next)
{
	char sStrValue[16];
	PropertyItem *item = getProperty( key, true);
	if (item == NULL) return false;

	item->bIsDefault = false;
	if (!item->string) item->string = new QString();
	if (item->color == NULL) item->color = new QColor();
	*(item->color) = *value;
	item->number = NULL;
	item->logical = NULL;
	item->pFont = NULL;
	sStrValue[0]='#';
	sprintf(sStrValue+1, "%02x\n", item->color->red());
	sprintf(sStrValue+3, "%02x\n", item->color->green());
	sprintf(sStrValue+5, "%02x\n", item->color->blue());
	*(item->string) = sStrValue;
	return true;
}

bool EditProperties::setNumericalValue( QString key , int value, bool next)
{
	PropertyItem *item = getProperty( key, true);
	if (item == NULL) return false;

	item->bIsDefault = false;
	if (!item->string) item->string = new QString();
	item->string->setNum( value );
	item->color = NULL;
	item->pFont = NULL;
	if (item->number == NULL) item->number = new int();
	*(item->number) = value;
	item->logical = NULL;
	return true;
}

bool EditProperties::setBoolValue( QString key, bool value, bool next)
{
	PropertyItem *pItem = getProperty( key, true );

	if (pItem != NULL)
	{
		pItem->bIsDefault = false;
		if (!pItem->string) pItem->string = new QString();
		pItem->color = NULL;
		pItem->number = NULL;
		pItem->pFont = NULL;
		if (pItem->logical == NULL) pItem->logical = new bool();
		*(pItem->string) = (value ? "true" : "false" );
		*(pItem->logical) = value;
	}
	return true;
}


// 
// EditProperties ::toFirst(.)
//
/**  @brief Go to the first value of the key
 *   @param key the QString key specifying the tag
 *
 */
void EditProperties::toFirst( QString key )
{
	find(key);
}

// 
// EditProperties :: hasMore(.)
//
/**  @brief Is there more value with the same key?
 *   @param key the QString key specifying the tag
 *   @return true if there is more value with the same key
 *
 */
bool EditProperties::hasMore( QString key )
{
	vector<PropertyItem, allocator<PropertyItem> > *list=&(*this)[key];
	if (list==NULL) return FALSE;
	return (list != NULL);
	//return ( list->current() != NULL );
}

/**
 * @class PropertyItem
 * @brief It stores a property value
 */

// 
// PropertyItem :: PropertyItem(.)
//
/**  @brief Creates a new PropertyItem with the given value
 *   @param value the initial value of the item
 *
 * It creates a new PropertyItem object, and gives a \e value initial
 * value to it.\n
 */
PropertyItem::PropertyItem( const char *value )
{
	setValue( value );
	bIsDefault = false;
}

// 
// PropertyItem :: setValue(.)
//
/*
 * This function tries to detect the type of the property's new value:
 *  - if it consists only of numbers, it can be a number.
 *  - if it is "true" or "false", it can be a boolean value.
 *  - if it begins with an asterisk, and it has 6 numbers after it,
 *    then it can be a color code.
 */
void PropertyItem::setValue( const char *value )
{
	if (value == NULL) string = new QString();
	else string = new QString(value);
	pComment = NULL;

	// Convert to QColor
	color = NULL;
	if ( (string->at(0)=='#') && (string->length()==7) )
	{
	  int r, g, b, r1, r2;
	  color = (QColor *)1;
	  r1=tonumber(string->at(1).latin1());
	  r2=tonumber(string->at(2).latin1());
	  if ((r1<0)||(r2<0)) color = NULL;
	  if (color) {
		r=(r1<<4)+r2;
		r1=tonumber(string->at(3).latin1());
		r2=tonumber(string->at(4).latin1());
		if ((r1<0)||(r2<0)) { color = NULL; }
	  }
	  if (color) {
		g=(r1<<4)+r2;
		r1=tonumber(string->at(5).latin1());
		r2=tonumber(string->at(6).latin1());
		if ((r1<0)||(r2<0)) { color = NULL; }
	  }
	  if (color) {
	    b=(r1<<4)+r2;
	    dmesg("6:config", "        Saved rgb: (R %d, G %d, B %d)", r, g, b);
		color = new QColor(r, g, b);
	  } else color = NULL;
	}

	// Convert to bool
	if ( ( string->lower() == "true" ) || ( (*string)=="1") )
		logical = new bool(TRUE);
	else if ( ( string->lower() == "false" ) || ( (*string)=="0") )
		logical = new bool(FALSE);
	else logical = NULL;

	// Convert to int
	bool ok;
	int result = string->toInt( &ok );
	number = ( ok ? new int(result) : NULL );

	// Convert to font
	pFont = NULL;
	setFontValue();
}

/* 
 * .EditProperties :: loadFontValue(.)
 */
/**  @fn loadFontValue
 *   @brief Parse the given string and create a font value
 *   @param sStr the QString key specifying the font value
 *   @return a pointer to the created QFont object
 *
 * It parses the string and creates a QFont * object:
 * Sample string: "fntfamily:Courier,fntsize:12,fntweight:1"
 *
 */
void PropertyItem::setFontValue( )
{
	bool bFont = false;
	QFont *pFont;
	QString sPart, sStr = *string;
	int iPos;

	/* go to the default state */
	pFont = new QFont();

	while((iPos = sStr.find("fnt")) != -1)
	{
		if (iPos > 0) sStr.remove(0, iPos);

		int iPartEnd = sStr.find(",");
		if (iPartEnd != -1) sPart = sStr.mid(0, iPartEnd);
		else sPart = sStr;

		if (sPart.find("fntfamily=") == 0)
		{
			bFont = true;
			sPart.remove(0, 10);
			dmesg("5:config", "       font family: %s", sPart.latin1());
			pFont->setFamily(sPart);
		}
		else if (sPart.find("fntsize=") == 0)
		{
			bFont = true;
			sPart.remove(0, 8);
			dmesg("5:config", "       font size: %d", sPart.toInt());
			pFont->setPointSize(sPart.toInt());
		}
		else if (sPart.find("fntweight=") == 0)
		{
			bFont = true;
			sPart.remove(0, 10);
			dmesg("5:config", "       font weight: %d", sPart.toInt());
			pFont->setWeight(sPart.toInt());
		}
		sStr.remove(0, 3);
	}

	if (bFont) this->pFont = pFont;
	else delete(pFont);
}


