/**
 * @file Preferences.cpp
 * @brief Preferences window and tools
 * 
 * @author Ferenc Moricz
 * $Revision: 1.14 $
 * $Date: 2004/02/22 00:12:50 $
 *
 */

#include "Debug.h"
#include "Editor.h"
#include "Preferences.h"
#include "QTExt.h"

#include <qcolordialog.h>
#include <qfontdialog.h>
#include <qlistview.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qvgroupbox.h>

#include <stdlib.h>

PreferenceItem::PreferenceItem()
{
	pColor = NULL;
	pCheck = NULL;
	pLineEdit = NULL;
	pFont = NULL;
	pSelectBox = NULL;
	bChanged = bChangedApplied = FALSE;
}

void PreferenceItem::slotItem( )
{
	switch(eType)
	{
		case COLOR:
			{
				QColor oNewColor;
				if (pColor)
				{
					oNewColor = QColorDialog::getColor(*pColor);
					delete(pColor);
					pColor = NULL;
				} else oNewColor = QColorDialog::getColor(Qt::black);
				pColor = new QColor(oNewColor);
				Preferences::drawPixMap(pButton, pColor);
				bChanged = true;
				break;
			}
		case FONT:
			{
				bool ok;
				QFont oFont = QFontDialog::getFont( &ok, *pFont );
				if (ok)
				{
					*pFont = oFont;
					pButton->setFont( *pFont );
					bChanged = true;
				}
				break;
			}
		default:
				bChanged = true;
				break;
	}
}

void PreferenceItem::textChanged(const QString &sNewText)
{
	slotItem();
}

void PreferenceItem::stateChanged(int iState)
{
	slotItem();
}

PreferenceItemContainer::PreferenceItemContainer(QWidget *pParent) : QHBox(pParent)
{
	setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
	pActiveWidget = NULL;
}

void PreferenceItemContainer::filter(QWidget *pWidget)
{
	pWidget->installEventFilter( this );
	installEventFilter( this );
}

bool PreferenceItemContainer::event(QEvent *e)
{
	eventFilter(NULL, e);
	return QHBox::event(e);
}

bool PreferenceItemContainer::eventFilter(QObject *o, QEvent *e)
{
	bool bRet = (pActiveWidget == o ? false : true);

	/* if there are no tooltip help to display, return */
	if ((sComment == NULL) || (*sComment == ""))
		return false;

	/* if this is a mousepress event, display the help */
	if (e->type() == QEvent::MouseButtonPress)
	{
		QString sHelpPage, sTemp;
		sTemp = sProperty;
#ifdef _SYSTEM_LINUX_
		sHelpPage = INSTALL_DOC "/quick_help/";
#else
		sHelpPage = INSTALL_DOC;
		sHelpPage.append("\\quick_help\\");
#endif
		if (sComment) sHelpPage.append(*sComment);

		if (pPreferences->pHelpWindow)
			pPreferences->pHelpWindow->setSource(sHelpPage);

		//return QWidget::eventFilter(o, e);
		return bRet;
	}
	
	return FALSE;
}

void PreferencePage::revert( )
{
	for(int i=0; i < (int)oItems.size(); i++)
	{
		PreferenceItem *pItem = oItems[i];
		switch(pItem->eType)
		{
			case PreferenceItem::COLOR:
				{
					if (pItem->pColor) delete(pItem->pColor);
					const QColor *pOldColor = EditBufferText::properties->getColorValue(
						*(pItem->sProperty), &Qt::white);
					pItem->pColor = new QColor(*pOldColor);
					Preferences::drawPixMap(pItem->pButton, pItem->pColor);
					break;
				}
			case PreferenceItem::TEXT:
				{
					const QString *sText = EditBufferText::properties->getValue(
						*(pItem->sProperty), NULL);
					if (sText != NULL) pItem->pLineEdit->setText(*sText);
					break;
				}
			case PreferenceItem::SELECT:
				{
					const QString *sText = EditBufferText::properties->getValue(
						*(pItem->sProperty), NULL);
#if QT_VERSION > 280
					if (sText != NULL) pItem->pSelectBox->setCurrentText(*sText);
#else
					if (sText != NULL)
					{
						int iN;
						for(iN = 0; iN < pItem->pSelectBox->count(); iN++)
						{
							if (pItem->pSelectBox->text(iN) == *sText)
								pItem->pSelectBox->setCurrentItem( iN );
//						pItem->pSelectBox->changeItem(*sText, pItem->pSelectBox->currentItem());
						}
					}
#endif
					break;
				}
			case PreferenceItem::CHECK:
				{
					pItem->pCheck->setChecked(
						EditBufferText::properties->getBoolValue(*(pItem->sProperty), false));
					break;
				}
			case PreferenceItem::FONT:
				{
					QString sQuery;
					QFont oDefFont("Fixed", 10);
					if (!pItem->pFont) pItem->pFont = new QFont();
					*(pItem->pFont) = pItem->pButton->font();
					*(pItem->pFont) = *(EditBufferText::properties->getFontValue(
					             *(pItem->sProperty), &oDefFont));
					pItem->pButton->setFont( *(pItem->pFont) );
/*
					sQuery = *(pItem->sProperty);
					sQuery.append( ".family" );
					pItem->pFont->setFamily( *EditBufferText::properties->getValue(sQuery,
						              new QString("courier")) );
					sQuery = *(pItem->sProperty);
					sQuery.append( ".size" );
					pItem->pFont->setPointSize( EditBufferText::properties->getNumericalValue(sQuery, 12) );
					pItem->pButton->setFont( *(pItem->pFont) );
*/
					break;
				}
			default: break;
		}
		pItem->bChanged = FALSE;
	}
}

void Preferences::slotPrefOK( void )
{
	slotPrefApply();
	slotPrefCancel();
}

void Preferences::slotPrefApply( void )
{
	QListViewItem *pCurPage = pCategoryTree->currentItem();
	if (pCurPage)
	{
		PreferencePage *pPrefPage = oPrefPages[ pCurPage->text(0) ];  //?exists
		vector<PreferenceItem *> *pItems = NULL;
		if (pPrefPage) pItems = &(pPrefPage->oItems);
		if (pItems) for(int i = 0; i < (int)pItems->size(); i++)
		{
			if ((! (*pItems)[i]->bChangedApplied) &&
			       (*pItems)[i]->bChanged)
			{
				/* put this property to 'oChangedItems' -> mark as changed */
				oChangedItems.push_back(*(*pItems)[i]->sProperty);
				(*pItems)[i]->bChangedApplied = TRUE;
			}
//			if (oChangedItems.find(pItems->at(i)->sProperty) != -1)
			if ((*pItems)[i]->bChangedApplied)
			{
				if ((*pItems)[i]->eType == PreferenceItem::COLOR)
				{
					EditBufferText::properties->setColorValue(
						*(*pItems)[i]->sProperty, (*pItems)[i]->pColor);
				} else
				if ((*pItems)[i]->eType == PreferenceItem::TEXT)
				{
					EditBufferText::properties->setValue(
						*(*pItems)[i]->sProperty, new QString((*pItems)[i]->pLineEdit->text()));
				} else
				if ((*pItems)[i]->eType == PreferenceItem::SELECT)
				{
					EditBufferText::properties->setValue(
						*((*pItems)[i]->sProperty),
						new QString((*pItems)[i]->pSelectBox->currentText()));
				} else
				if ((*pItems)[i]->eType == PreferenceItem::CHECK)
				{
					EditBufferText::properties->setBoolValue(
						*((*pItems)[i]->sProperty), (*pItems)[i]->pCheck->isChecked());
				} else
				if ((*pItems)[i]->eType == PreferenceItem::FONT)
				{
					if ((*pItems)[i]->pFont)
					{
						EditBufferText::properties->setFontValue(
							*((*pItems)[i]->sProperty), (*pItems)[i]->pFont);
					}
				}
			}
		}
	}

	/* reconfigure (TBD should work not only on the actual buffer */
	Editor::editor->reconfigure();
/*	EditSession *pSession = (EditSession *)(Editor::editor->getStack()->visibleWidget());
	if (pSession)
	{
		EditBuffer *pBuffer = (EditBuffer *)(pSession->currentPage());
		if (pBuffer) pBuffer->reconfigure();
	}*/
}

void Preferences::slotPrefSaveConfig( void )
{
	QFileInfo fi;
	resetForSave();

	EditBufferText::properties->saveConfig("freddy.conf",
	                                                 &oChangedItems, FALSE);
	/* if we own the main freddy window -> save the configuration */
	if (Editor::editor->pEditorServer)
	{
		fi.setFile(Editor::editor->configdir_user, "saved-state");
		EditBufferText::properties->saveConfig(fi.fileName().latin1(),
		                                                 &oChangedItems, TRUE);
		Editor::editor->saveState();
		Editor::editor->bSavedStateOpened = TRUE;
	}
}

void Preferences::slotPrefCancel()
{
	if (pPrefDialogWindow != NULL)
	{
		pPrefDialogWindow->close( false );
	}
}

void Preferences::slotPrefHelp(bool bCheck)
{
	if (bCheck && !pHelpWindow)
	{
		pHelpWindow = new QTextBrowser(pSplitter, "Help browser");
		pHelpWindow->show();
#ifdef _SYSTEM_LINUX_
		pHelpWindow->setSource(INSTALL_DOC "/quick_help/index.html");
#elif defined _SYSTEM_WINDOWS_
		pHelpWindow->setSource(QString(INSTALL_DOC)+QString("\\quick_help\\index.html"));
		QString a = INSTALL_DOC;
		a+= "\\quick_help\\index.html";
#endif
	} else if (!bCheck && pHelpWindow)
	{
		pHelpWindow->close(TRUE);
		pHelpWindow = NULL;
	}
}

void Preferences::slotTreeClicked( QListViewItem *pItem )
{
	const char *sName;
	QWidget *pNextActive;
	/* get the text for the clicked item */
	if (pItem == NULL) return;
	sName = pItem->text(0).latin1();
	/* get the page which is assigned to this text */
	if (oPrefPages.find( sName ) != oPrefPages.end())
	{
		PreferencePage *pPrefPage = oPrefPages[ sName ];
		pNextActive = pPrefPage->pPageWidget;
		/* raise this page */
		if (pNextActive) pPages->raiseWidget( pNextActive );
		pPrefPage->revert();
	}
}

void Preferences::drawPixMap(QPushButton *pButton, QColor *pColor)
{
	QPixmap pPixmap(21, 13);
	pPixmap.fill( *pColor );
	QPainter p;
	p.begin( &pPixmap );
	p.drawRect( 0, 0, 21, 13);
	p.end();
	pButton->setPixmap(pPixmap);
}

#include <qmainwindow.h>
Preferences::Preferences( void )
{
	/* initialize values */
	pHelpWindow = NULL;

  /* Create the Preferences window */
	pPrefDialogWindow = new QDialog(NULL, "Preferences", true);

	pPrefDialog = new QVBox(pPrefDialogWindow, "Preferences");
//			Qt::WStyle_Dialog | WShowModal );
	pPrefDialog->setSizePolicy(QSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum));
	pPrefDialog->setMinimumSize(650, 550);
	pPrefDialog->setMargin( 7 );
	pPrefDialog->setSpacing( 7 );
	pSplitter = new QSplitter(Qt::Vertical, pPrefDialog);
	QHBox *pBox = new QHBox(pSplitter);
	pBox->setSpacing( 7 );

	/* OK, Apply, Cancel buttons */
	QHBox *pButtonBox = new QHBox( pPrefDialog );
	pButtonBox->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
	pButtonBox->setSpacing(7);
	QCheckBox *pCheck = new QCheckBox("Help", pButtonBox, "Help");
	pCheck->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
	QLabel *pLabel = new QLabel("", pButtonBox, "");
	QPushButton *pOK = new QPushButton("OK", pButtonBox, "OK");
	QPushButton *pApply = new QPushButton("Apply", pButtonBox, "Apply");
	QPushButton *pSaveConfig = new QPushButton("Save", pButtonBox, "Save");
	QPushButton *pCancel = new QPushButton("Cancel", pButtonBox, "Cancel");
	pLabel->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
	pOK->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
	pApply->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
	pSaveConfig->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
	pCancel->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed));
	//!pOK->setMinimumWidth( pOK->width() );
	//!pApply->setMinimumWidth( pApply->width() );
	//!pSaveConfig->setMinimumWidth( pSaveConfig->width() );
	//!pCancel->setMinimumWidth( pCancel->width() );

	QObject::connect( pOK, SIGNAL(clicked( void )),
						this, SLOT( slotPrefOK( void ) ) );
	QObject::connect( pApply, SIGNAL(clicked( void )),
						this, SLOT( slotPrefApply( void ) ) );
	QObject::connect( pSaveConfig, SIGNAL(clicked( void )),
						this, SLOT( slotPrefSaveConfig( void ) ) );
	QObject::connect( pCancel, SIGNAL(clicked( void )),
						this, SLOT( slotPrefCancel( void ) ) );
	QObject::connect( pCheck, SIGNAL(toggled( bool )),
						this, SLOT( slotPrefHelp( bool ) ) );

	/* create the left side (the tree )*/
	QVBox *pTableBox = new QVBox( pBox );
	pTableBox->setSizePolicy(QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Preferred));
	pTableBox->setMinimumSize(200, 100);
	pCategoryTree = new QListView( pTableBox );
	pCategoryTree->addColumn( "Category" );
	pCategoryTree->setSorting(-1, false);
	pCategoryTree->setRootIsDecorated( TRUE );                                                
	pCategoryTree->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred));
	pCategoryTree->setMaximumSize(200, 400);
	QLabel *pLabel2 = new QLabel("", pTableBox, "");
	pLabel2->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred));
	pLabel2->setMinimumSize( 0, 0 );
	pTableBox->setStretchFactor(pCategoryTree, 500);
	pTableBox->setStretchFactor(pLabel2, 1);

	QObject::connect( pCategoryTree, SIGNAL(clicked( QListViewItem * )),
						this, SLOT( slotTreeClicked(  QListViewItem * ) ) );

	/* create the right side */
	pPages = new QWidgetStack( pBox );
	pPages->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred));
	pPages->setMaximumSize(1000, 1000);

	/* Put the elements into the window */
	if ((Editor::editor == NULL) || (Editor::editor->com == NULL))
		derr("command", "FATAL: command interface is not initialized!");
	if (Editor::editor->com->menu.find("editor.preferences") ==
	        Editor::editor->com->menu.end())
	{
		derr("config", "Unable to load the preferences config!");
		return;
	}

	vector<PropertyItem, allocator<PropertyItem> > *pPrefItems =
		&(Editor::editor->com->menu["editor.preferences"]);

	if (pPrefItems == NULL)
	{
		derr("config", "Unable to load the preferences config!");
		return;
	}

	/* collect the items for the preferences window */
	QVBox *pActPage = NULL;
	QString sActPageName;
	QGroupBox *pActGroup = NULL;
	vector <QVBox *, allocator<QVBox *> > pPageList;

	for(int i = 0; i < (int)pPrefItems->size(); i++)
	{
		QString sTemp = *(*pPrefItems)[i].getString();
		sTemp.replace(QRegExp("\r"), "");
		sTemp.replace(QRegExp("\n"), "");
		char *sTag = strdup( sTemp.latin1() );
		
		char *sType = strtok(sTag, " \t\r\n");
		if (!strcmp(sType, "category")) /* this is a category */
		{
			char *sThisName = sType;
			QListViewItem *pPrev = NULL, *pThis = NULL;
			PreferencePage *pPage;
			while((sThisName = strtok(NULL, " /\t\r\n")) != NULL)
			{
				QString sName = sThisName;
				sName = sName.replace( QRegExp("_"), " " );
				if (oPrefPages.find(sName) == oPrefPages.end())
					pThis = NULL, pPage = NULL;
				else {
					pPage = oPrefPages[sName];
					pThis = (QListViewItem *)pPage->pPageTreeItem;
				}
				/* this is a new page - create it ! */
				if (pThis == NULL)
				{
					if (pPrev == NULL)
						pThis = new QListViewItem( pCategoryTree, sName, NULL);
					else pThis = new QListViewItem( pPrev, sName, NULL);
					pActPage = new QVBox( pPages );
					sActPageName = sName;
					pPage = new PreferencePage();
					pPage->pPageWidget = pActPage;
					pPage->pPageTreeItem = pThis;
					oPrefPages[sName] = pPage;
					pPageList.push_back(pActPage);
					pActPage->setMinimumSize(100, 100);
					pActPage->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred));
					pActPage->setSpacing(10);
					sName.prepend("  ");
					QLabel *pLabel = new QLabel(sName, pActPage, sName);
					QFont oFont(pLabel->font());
					oFont.setPixelSize(20);
					pLabel->setFont(oFont);
#ifdef _SYSTEM_LINUX_
					pLabel->setPaletteBackgroundColor(QColor( 160, 160, 160 ));
#else
					pLabel->setBackgroundColor(QColor( 160, 160, 160 ));
#endif
					pLabel->setMaximumHeight( 32 );
//					pPages->raiseWidget(pActPage);
				}
				pPrev = pThis;
				pActGroup = NULL;
			}
			// TBD parsing the correct comment string
			if (pPage) pPage->sComment = new QString(strtok(NULL, " \t\r\n"));
		} else
		if (!strcmp(sType, "group")) /* this is a group */
		{
			//QString *sComment = new QString(strtok(NULL, " \t\r\n"));
			if (pActPage != NULL)
			{
				QString oName = strtok(NULL, " \t\r\n");
				QString *sComment = new QString(strtok(NULL, " \t\r\n"));
				oName = oName.replace( QRegExp("_"), " " );
				pActGroup = new QVGroupBox( oName, pActPage, oName );
				pActGroup->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
				pActGroup->setMinimumSize( 10, 10 );
				pActGroup->show();
				/* create a itemcontainer for the group */
				PreferenceItemContainer *pBox = new PreferenceItemContainer((QWidget *)pActPage);
				pBox->pPreferences = this;
				pBox->sComment = sComment;
				pBox->filter(pActGroup);
			}
		} else
		if (!strcmp(sType, "item")) /* this is an item */
		{
			sType = strtok(NULL, " \t\r\n");
			QString sName( strtok(NULL, " \t\r\n"));    /* name */
			sName = sName.replace( QRegExp("_"), " " );
			QString *sProperty = new QString(strtok(NULL, " \t\r\n")); /* property */
			QString *sComment = new QString(strtok(NULL, " \t\r\n"));

			PreferencePage *pPrefPage;
			if (oPrefPages.find( sActPageName ) != oPrefPages.end())
				pPrefPage = oPrefPages[sActPageName];
			else pPrefPage = NULL;
			vector<PreferenceItem *> *pItems = NULL;
			if (pPrefPage) pItems = &(pPrefPage->oItems);

			if (!strcmp(sType, "check"))
			{
				PreferenceItemContainer *pBox = new PreferenceItemContainer(
					(pActGroup ? (QWidget *)pActGroup : (QWidget *)pActPage));
				pBox->pPreferences = this;
				pBox->sProperty = *sProperty;
				pBox->sComment = sComment;
				QCheckBox *pCheck = (new QCheckBox(sName, pBox, sName));
				pBox->pActiveWidget = pCheck;
				pBox->filter(pCheck);
				if (pItems)
				{
					PreferenceItem *pItem = new PreferenceItem();
					pItem->eType = PreferenceItem::CHECK;
					pItem->pCheck = pCheck;
					QObject::connect( pCheck, SIGNAL(stateChanged(int)),
					                  pItem, SLOT(stateChanged(int)) ); 
					pItem->sProperty = sProperty;
					pItems->push_back(pItem);
				}
			} else
			if (!strcmp(sType, "text"))
			{
				PreferenceItemContainer *pBox = new PreferenceItemContainer(
					(pActGroup ? (QWidget *)pActGroup : (QWidget *)pActPage));
				pBox->pPreferences = this;
				pBox->sProperty = *sProperty;
				pBox->sComment = sComment;
				QLabel *pLabel = new QLabel(sName, pBox, sName);
				QLineEdit *pLineEdit = new QLineEdit(NULL, pBox, NULL);
				pBox->pActiveWidget = pLineEdit;
				pBox->filter(pLineEdit);
				pBox->filter(pLabel);
				pLineEdit->setMinimumWidth(150);
				if (pItems)
				{
					PreferenceItem *pItem = new PreferenceItem();
					pItem->eType = PreferenceItem::TEXT;
					pItem->pLineEdit = pLineEdit;
					QObject::connect( pLineEdit, SIGNAL(textChanged(const QString &)),
					                  pItem, SLOT(textChanged(const QString &)) ); 
					pItem->sProperty = sProperty;
					pItems->push_back(pItem);
				}
			} else
			if (!strcmp(sType, "color"))
			{
				PreferenceItemContainer *pBox = new PreferenceItemContainer(
					(pActGroup ? (QWidget*)pActGroup : (QWidget*)pActPage));
				pBox->pPreferences = this;
				pBox->sProperty = *sProperty;
				pBox->sComment = sComment;
				QLabel *pLabel = new QLabel(sName, pBox, sName);
				pLabel->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
				QPushButton *pButton = new QPushButton( "Color", pBox, "Color");
				pBox->pActiveWidget = pButton;
				pBox->filter(pButton);
				pBox->filter(pLabel);
				pButton->setSizePolicy( QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed) );
				pButton->setBaseSize(22, 14);
				pBox->show();
				if (pItems)
				{
					PreferenceItem *pItem = new PreferenceItem();
					pItem->eType = PreferenceItem::COLOR;
					pItem->pButton = pButton;
					QObject::connect( pButton, SIGNAL(clicked()), pItem, SLOT(slotItem()) ); 
					pItem->sProperty = sProperty;
					pItems->push_back(pItem);
				}
			} else
			if (!strcmp(sType, "font"))
			{
				PreferenceItemContainer *pBox = new PreferenceItemContainer(
					(pActGroup ? (QWidget*)pActGroup : (QWidget*)pActPage));
				pBox->pPreferences = this;
				pBox->sProperty = *sProperty;
				pBox->sComment = sComment;
				QLabel *pLabel = new QLabel(sName, pBox, sName);
				pLabel->setSizePolicy(QSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed));
				QPushButton *pButton = new QPushButton( "Font", pBox, "Font");
				pBox->pActiveWidget = pButton;
				pBox->filter(pButton);
				pBox->filter(pLabel);
				pButton->setSizePolicy( QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed) );
				pButton->setBaseSize(22, 14);
				pBox->show();
				if (pItems)
				{
					PreferenceItem *pItem = new PreferenceItem();
					pItem->eType = PreferenceItem::FONT;
					pItem->pButton = pButton;
					QObject::connect( pButton, SIGNAL(clicked()), pItem, SLOT(slotItem()) ); 
					pItem->sProperty = sProperty;
					pItems->push_back(pItem);
				}
			} else
			if (!strcmp(sType, "select"))
			{
				PreferenceItemContainer *pBox = new PreferenceItemContainer(
					(pActGroup ? (QWidget *)pActGroup : (QWidget *)pActPage));
				pBox->pPreferences = this;
				pBox->sProperty = *sProperty;
				pBox->sComment = sComment;
				QLabel *pLabel = new QLabel(sName, pBox, sName);
				pLabel->setSizePolicy(QSizePolicy(QSizePolicy::Preferred,
				                                        QSizePolicy::Fixed));
				QComboBox *pSelectBox = new QComboBox(pBox, NULL);
				pBox->pActiveWidget = pSelectBox;
				pBox->filter(pLabel);
				pBox->filter(pSelectBox);
				if (pItems)
				{
					PreferenceItem *pItem = new PreferenceItem();
					pItem->eType = PreferenceItem::SELECT;
					pItem->pSelectBox = pSelectBox;
					QObject::connect( pSelectBox, SIGNAL(activated(int)),
					                  pItem, SLOT(stateChanged(int)) ); 
					pItem->sProperty = sProperty;
					pItems->push_back(pItem);
					char *sSelectItem;
					while((sSelectItem = strtok(NULL, " \t\r\n")) != NULL)
						pSelectBox->insertItem(sSelectItem, -1);
				}
			}
		}

		free(sTag);
	}

	/* ident all config page to the top */
	vector<QVBox *, allocator<QVBox *> >::iterator oItem;
	for(oItem = pPageList.begin(); oItem != pPageList.end(); oItem ++)
	{
		QLabel *pFillLabel = new QLabel("", *oItem, "");
		pFillLabel->setSizePolicy(QSizePolicy(QSizePolicy::Preferred,
		                                      QSizePolicy::Preferred));
		pFillLabel->show();
	}
}

/*
 * Shows the dialog window
 */
void Preferences::show()
{                                        
	if (pPrefDialogWindow) pPrefDialogWindow->show();
}

/*
 * Closes the dialog window
 */
void Preferences::close()
{
	if (pPrefDialogWindow) pPrefDialogWindow->close(false);
}


void Preferences::resetForSave( void )
{
	for(int i = 0; i < (int)oChangedItems.size(); i++)
	{
		if (oChangedItems[i].startsWith("-"))
		{
			oChangedItems[i].remove(0, 1);
		}
	}
}

