#include <qglobal.h>

#if QT_VERSION > 280

#include "QTExt.h"
#include "qaccel.h"
#include "qbitmap.h"
#include "qtoolbutton.h"
#include "qtooltip.h"
#include "qapplication.h"
#include "qstyle.h"
#include "qpainter.h"
#include "qiconset.h"
#include "qcursor.h"
#include "private/qinternal_p.h"

#include <ctype.h>
#include <qdragobject.h>                /* < EXT > */


/*!
  \class ExtQTab qtabbar.h
  \ingroup advanced
  \brief The ExtQTab class provides the structures in a ExtQTabBar.


  This class is used for custom ExtQTabBar tab headings.

  \sa ExtQTabBar
*/


/*! \fn ExtQTab::ExtQTab()
  Constructs an empty tab.  All fields are set to empty.
*/

ExtQTab::ExtQTab()
    : enabled( TRUE ),
      id ( 0 ),
      iconset( 0 ),
      tb( 0 )
{
}

/*! \fn ExtQTab::ExtQTab( const QString& text )
  Constructs a tab with the text, \a text.
*/

ExtQTab::ExtQTab( const QString &text )
    : label( text ),
      enabled( TRUE ),
      id( 0 ),
      iconset( 0 ),
      tb( 0 )
{
}

/*! \fn ExtQTab::ExtQTab( const QIconSet& icon, const QString& text )
  Constructs a tab with an \a icon and the text, \a text.
*/

ExtQTab::ExtQTab( const QIconSet& icon, const QString& text )
    : label( text ),
      enabled( TRUE ),
      id( 0 ),
      iconset( new QIconSet(icon) ),
      tb( 0 )
{
}

/*!
  \fn QString ExtQTab::text() const
  Return the text of the ExtQTab label.
*/

/*!
  \fn QIconSet ExtQTab::iconSet() const
  Return the QIconSet of the ExtQTab.
*/

/*!
  \fn void ExtQTab::setRect( const QRect &rect )
  Set the ExtQTab QRect to \a rect.
*/

/*!
  \fn QRect ExtQTab::rect() const
  Return the QRect for the ExtQTab.
*/

/*!
  \fn void ExtQTab::setEnabled( bool enable )
  If \a enable is TRUE enable the ExtQTab, otherwise disable it.
*/

/*!
  \fn bool ExtQTab::isEnabled() const
  Returns TRUE if the ExtQTab is enabled, otherwise return FALSE.
*/

/*!
  \fn void ExtQTab::setIdentifier( int i )
  Set the identifier for the ExtQTab to \a i.
  Each identifier for a ExtQTabBar must be unique
*/

/*!
  \fn int ExtQTab::identifier() const
  Return the identifier for the ExtQTab.
*/



/*! Destroys the tab and frees up all allocated resources */

ExtQTab::~ExtQTab()
{
    delete iconset;
    tb = 0;
}

/*!
  \class ExtQTabBar qtabbar.h

  \brief The ExtQTabBar class provides a tab bar, e.g. for use in tabbed dialogs.

  \ingroup advanced

  ExtQTabBar is straightforward to use; it draws the tabs using one of
  the predefined \link ExtQTabBar::Shape shapes\endlink, and emits a
  signal when a tab is selected.  It can be subclassed to tailor the
  look and feel.

  The choice of tab shape is a matter of taste, although tab dialogs
  (preferences and the like) invariably use \c RoundedAbove, and
  nobody uses \c TriangularAbove.  Tab controls in windows other than
  dialogs almost always use either \c RoundedBelow or \c
  TriangularBelow.  Many spreadsheets and other tab controls in which
  all the pages are essentially similar use \c TriangularBelow,
  whereas \c RoundedBelow is used mostly when the pages are different
  (e.g. a multi-page tool palette).

  The most important part of ExtQTabBar's API is the signal selected().
  This is emitted whenever the selected page changes (even at startup,
  when the selected page changes from 'none').  There is also a slot,
  setCurrentTab(), which can be used to select a page
  programmatically.

  ExtQTabBar creates automatic accelerator keys in the manner of QButton;
  e.g. if a tab's label is "\&Graphics", Alt+G becomes an accelerator
  key for switching to that tab.

  The following virtual functions may need to be reimplemented:
  \list
  \i paint() paints a single tab.  paintEvent() calls paint() for
  each tab so that any overlap will look right.
  \i addTab() creates a new tab and adds it to the bar.
  \i selectTab() decides which tab, if any, the user selects with the mouse.
  \endlist

  The index of the current tab is returned by currentTab(). The tab
  with a particular index is returned by tabAt(), the tab with a
  particular id is returned by tab(). The index of a tab is returned
  by indexOf(). The current tab can be set by index or tab pointer
  using one of the setCurrentTab() functions.

  <img src=qtabbar-m.png> <img src=qtabbar-w.png>
*/

/*! \enum ExtQTabBar::Shape
  This enum type lists the built-in shapes supported by ExtQTabBar:

  \value RoundedAbove  the normal rounded look above the pages

  \value RoundedBelow  the normal rounded look below the pages

  \value TriangularAbove  triangular tabs above the pages (very
  unusual; included for completeness)

  \value TriangularBelow  triangular tabs similar to those used in
  the spreadsheet Excel, for example
*/

class ExtQTabBarToolTip;

struct ExtQTabPrivate {
    int id;
    int focus;
#ifndef QT_NO_ACCEL
    QAccel * a;
#endif
    ExtQTab *pressed;
    ExtQTabBar::Shape s;
    QToolButton* rightB;
    QToolButton* leftB;
    bool  scrolls;
    ExtQTabBarToolTip * toolTips;
};

#ifndef QT_NO_TOOLTIP
/* \internal
*/
class ExtQTabBarToolTip : public QToolTip
{
public:
    ExtQTabBarToolTip( QWidget * parent )
	: QToolTip( parent ) {}
    virtual ~ExtQTabBarToolTip() {}

    void add( ExtQTab * tab, const QString & tip )
    {
	tabTips.replace( tab, tip );
    }

    void remove( ExtQTab * tab )
    {
	tabTips.erase( tab );
    }

    QString tipForTab( ExtQTab * tab ) const
    {
	QMapConstIterator<ExtQTab *, QString> it;
	it = tabTips.find( tab );
	if ( it != tabTips.end() )
	    return it.data();
	else
	    return QString();
    }

protected:
    void maybeTip( const QPoint & p )
    {
	ExtQTabBar * tb = (ExtQTabBar *) parentWidget();
	if ( !tb )
	    return;

	// check if the scroll buttons in the tab bar are visible -
	// don't display any tips if the pointer is over one of them
	QRect rectL, rectR;
	rectL.setRect( tb->d->leftB->x(), tb->d->leftB->y(),
		       tb->d->leftB->width(), tb->d->leftB->height() );
	rectR.setRect( tb->d->rightB->x(), tb->d->rightB->y(),
		       tb->d->rightB->width(), tb->d->rightB->height() );
	if ( tb->d->scrolls && (rectL.contains( p ) || rectR.contains( p )) )
	     return;

#ifndef QT_NO_TOOLTIP
	// find and show the tool tip for the tab under the point p
	QMapIterator<ExtQTab *, QString> it;
	for ( it = tabTips.begin(); it != tabTips.end(); ++it ) {
	    if ( it.key()->rect().contains( p ) )
		tip( it.key()->rect(), it.data() );
	}
#endif
    }

private:
    QMap<ExtQTab *, QString> tabTips;
};
#endif

/*!
  \fn void ExtQTabBar::selected( int id )

  ExtQTabBar emits this signal whenever any tab is selected, whether by
  the program or by the user.  The argument \a id is the id of the tab
  as returned by addTab().

  show() is guaranteed to emit this signal; you can display
  your page in a slot connected to this signal.
*/


/*!
  Constructs a new, empty tab bar; the \a parent and \a name arguments
  are passed on to the QWidget constructor.
*/

ExtQTabBar::ExtQTabBar( QWidget * parent, const char *name )
    : QWidget( parent, name, WRepaintNoErase | WResizeNoErase |
	       WNoMousePropagation  )
{
    d = new ExtQTabPrivate;
    d->pressed = 0;
    d->id = 0;
    d->focus = 0;
    d->toolTips = 0;
#ifndef QT_NO_ACCEL
    d->a = new QAccel( this, "tab accelerators" );
    connect( d->a, SIGNAL(activated(int)), this, SLOT(setCurrentTab(int)) );
#endif
    d->s = RoundedAbove;
    d->scrolls = FALSE;
    d->leftB = new QToolButton( LeftArrow, this, "qt_left_btn" );
    connect( d->leftB, SIGNAL( clicked() ), this, SLOT( scrollTabs() ) );
    d->leftB->hide();
    d->rightB = new QToolButton( RightArrow, this, "qt_right_btn" );
    connect( d->rightB, SIGNAL( clicked() ), this, SLOT( scrollTabs() ) );
    d->rightB->hide();
    l = new QPtrList<ExtQTab>;
    lstatic = new QPtrList<ExtQTab>;
    lstatic->setAutoDelete( TRUE );
    setFocusPolicy( TabFocus );
    setSizePolicy( QSizePolicy( QSizePolicy::Preferred, QSizePolicy::Fixed ) );
    setAcceptDrops( true );             /* < EXT */
    selectreleased = true;              /* EXT > */
}


/*!
  Destroys the tab control, freeing memory used.
*/

ExtQTabBar::~ExtQTabBar()
{
#ifndef QT_NO_TOOLTIP
    if ( d->toolTips )
	delete d->toolTips;
#endif
    delete d;
    d = 0;
    delete l;
    l = 0;
    delete lstatic;
    lstatic = 0;
}


/*!
  Adds the tab, \a newTab, to the tab control.

  Sets \a newTab's id to a new id and places the tab just to the right of the
  existing tabs. If the tab's label contains an ampersand, the letter
  following the ampersand is used as an accelerator for the tab, e.g.
  if the label is "Bro&wse" then Alt+W becomes an accelerator which
  will move the focus to this tab.
  Returns the id.

  \sa insertTab()
*/

int ExtQTabBar::addTab( ExtQTab * newTab )
{
    return insertTab( newTab );
}


/*!
  Inserts the tab, \a newTab, into the tab control.

  If \a index is not specified, the tab is simply added. Otherwise
  it's inserted at the specified position.

  Sets \a newTab's id to a new id. If the tab's label contains an
  ampersand, the letter following the ampersand is used as an
  accelerator for the tab, e.g. if the label is "Bro&wse" then Alt+W
  becomes an accelerator which will move the focus to this tab.
  Returns the id.

  \sa addTab()
*/

int ExtQTabBar::insertTab( ExtQTab * newTab, int index )
{
    newTab->id = d->id++;
    newTab->setTabBar( this );
    l->insert( 0, newTab );
    if ( index < 0 || index > int(lstatic->count()) )
	lstatic->append( newTab );
    else
	lstatic->insert( index, newTab );

    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );

#ifndef QT_NO_ACCEL
    int p = QAccel::shortcutKey( newTab->label );
    if ( p )
	d->a->insertItem( p, newTab->id );
#endif

    return newTab->id;
}

/* < EXT */
void ExtQTabBar::moveTab( ExtQTab *mtab, int index )
{
    if ( lstatic->find(mtab) != -1 )
    {
        bool bAutoDel = lstatic->autoDelete();
        lstatic->setAutoDelete( false );
        lstatic->remove( mtab );
        lstatic->setAutoDelete( bAutoDel );

        if ( index < 0 || index > int(lstatic->count()) )
            lstatic->append( mtab );
        else lstatic->insert( index, mtab );

    }

    if ( l->find(mtab) != -1 )
    {
        bool bAutoDel = l->autoDelete();
        l->setAutoDelete( false );
        l->remove( mtab );
        l->setAutoDelete( bAutoDel );

        if ( index < 0 || index > int(l->count()) )
            l->append( mtab );
        else l->insert( index, mtab );
    }
	
	layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );
}
/* EXT > */


/*!
  Removes tab \a t from the tab control, and deletes the tab.
*/
void ExtQTabBar::removeTab( ExtQTab * t )
{
    //#### accelerator labels??
#ifndef QT_NO_TOOLTIP
    if ( d->toolTips )
	d->toolTips->remove( t );
#endif
#ifndef QT_NO_ACCEL
    if ( d->a )
	d->a->removeItem( t->id );
#endif
    // remove the TabBar Reference
    if(d->pressed == t)
	d->pressed = NULL;
    t->setTabBar( 0 );
    l->remove( t );
    lstatic->remove( t );
    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );
    update();
}


/*!
  Enables tab \a id if \a enabled is TRUE or disables it if \a enabled is
  FALSE.  If \a id is currently selected, setTabEnabled(FALSE) makes
  another tab selected.

  setTabEnabled() updates the display if this causes a
  change in \a id's status.

  \sa update(), isTabEnabled()
*/

void ExtQTabBar::setTabEnabled( int id, bool enabled )
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() ) {
	if ( t && t->id == id ) {
	    if ( t->enabled != enabled ) {
		t->enabled = enabled;
#ifndef QT_NO_ACCEL
		d->a->setItemEnabled( t->id, enabled );
#endif
		QRect r( t->r );
		if ( !enabled && id == currentTab() ) {
		    QPoint p1( t->r.center() ), p2;
		    int m = 2147483647;
		    int distance;
		    // look for the closest enabled tab - measure the
		    // distance between the centers of the two tabs
		    for( ExtQTab * n = l->first(); n; n = l->next() ) {
			if ( n->enabled ) {
			    p2 = n->r.center();
			    distance = (p2.x() - p1.x())*(p2.x() - p1.x()) +
				       (p2.y() - p1.y())*(p2.y() - p1.y());
			    if ( distance < m ) {
				t = n;
				m = distance;
			    }
			}
		    }
		    if ( t->enabled ) {
			r = r.unite( t->r );
			l->append( l->take( l->findRef( t ) ) );
			emit selected( t->id );
		    }
		}
		repaint( r );
	    }
	    return;
	}
    }
}


/*!
  Returns TRUE if the tab with id \a id is enabled, or FALSE if it
  is disabled or there is no such tab.

  \sa setTabEnabled()
*/

bool ExtQTabBar::isTabEnabled( int id ) const
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() ) {
	if ( t && t->id == id )
	    return t->enabled;
    }
    return FALSE;
}



/*!\reimp
*/
QSize ExtQTabBar::sizeHint() const
{
    ExtQTab * t = l->first();
    if ( t ) {
	QRect r( t->r );
	while ( (t = l->next()) != 0 )
	    r = r.unite( t->r );
	return r.size().expandedTo( QApplication::globalStrut() );
    } else {
	return QSize( 0, 0 ).expandedTo( QApplication::globalStrut() );
    }
}

/*! \reimp */

QSize ExtQTabBar::minimumSizeHint() const
{
    return QSize( d->rightB->sizeHint().width() * 2 + 75, sizeHint().height() );
}

/*!  Paints the tab \a t using painter \a p.  If and only if \a selected
  is TRUE, \a t is drawn currently selected.

  This virtual function may be reimplemented to change the look of
  ExtQTabBar.  If you decide to reimplement it, you may also need to
  reimplement sizeHint().
*/

void ExtQTabBar::paint( QPainter * p, ExtQTab * t, bool selected ) const
{
    QStyle::SFlags flags = QStyle::Style_Default;

    if (isEnabled() && t->isEnabled())
	flags |= QStyle::Style_Enabled;
    if ( selected )
	flags |= QStyle::Style_Selected;
    else if(t == d->pressed)
	flags |= QStyle::Style_Sunken;
    //selection flags
    if(t->rect().contains(mapFromGlobal(QCursor::pos())))
	flags |= QStyle::Style_MouseOver;
    style().drawControl( QStyle::CE_TabBarTab, p, this, t->rect(),
			 colorGroup(), flags, QStyleOption((QTab *)t) );

    QRect r( t->r );
    p->setFont( font() );

    int iw = 0;
    int ih = 0;
    if ( t->iconset != 0 ) {
	iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 4;
	ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
    }
    QFontMetrics fm = p->fontMetrics();
    int fw = fm.width( t->label );
    fw -= t->label.contains('&') * fm.width('&');
    fw += t->label.contains("&&") * fm.width('&');
    int w = iw + fw + 4;
    int h = QMAX(fm.height() + 4, ih );
    paintLabel( p, QRect( r.left() + (r.width()-w)/2 - 3,
			  r.top() + (r.height()-h)/2,
			  w, h ), t, t->id == keyboardFocusTab() );
}

/*!
  Paints the label of tab \a t centered in rectangle \a br using
  painter \a p. A focus indication is drawn if \a has_focus is TRUE.
*/

void ExtQTabBar::paintLabel( QPainter* p, const QRect& br,
			  ExtQTab* t, bool has_focus ) const
{
    QRect r = br;
    bool selected = currentTab() == t->id;
    if ( t->iconset) {
	// the tab has an iconset, draw it in the right mode
	QIconSet::Mode mode = (t->enabled && isEnabled())
	    ? QIconSet::Normal : QIconSet::Disabled;
	if ( mode == QIconSet::Normal && has_focus )
	    mode = QIconSet::Active;
	QPixmap pixmap = t->iconset->pixmap( QIconSet::Small, mode );
	int pixw = pixmap.width();
	int pixh = pixmap.height();
	r.setLeft( r.left() + pixw + 4 );
	r.setRight( r.right() + 2 );
	// ### the pixmap shift should probably not be hardcoded..
	p->drawPixmap( br.left() + 2 + ((selected == TRUE) ? 0 : 2),
		       br.center().y()-pixh/2 + ((selected == TRUE) ? 0 : 2),
		       pixmap );
    }

    QStyle::SFlags flags = QStyle::Style_Default;

    if (isEnabled() && t->isEnabled())
	flags |= QStyle::Style_Enabled;
    if (has_focus)
	flags |= QStyle::Style_HasFocus;
    style().drawControl( QStyle::CE_TabBarLabel, p, this, r,
			 t->isEnabled() ? colorGroup(): palette().disabled(),
			 flags, QStyleOption((QTab *)t) );
}


/*!
  Repaints the tab row.  All the painting is done by paint();
  paintEvent() only decides which tabs need painting and in what
  order. The event is passed in \a e.

  \sa paint()
*/

void ExtQTabBar::paintEvent( QPaintEvent * e )
{
    QSharedDoubleBuffer buffer( this, e->rect() );

    buffer.painter()->setBrushOrigin( rect().bottomLeft() );
    buffer.painter()->fillRect( 0, 0, width(), height(),
				colorGroup().brush( QColorGroup::Background ) );
    ExtQTab * t;
    t = l->first();
    do {
	ExtQTab * n = l->next();
	if ( t && t->r.intersects( e->rect() ) )
	    paint( buffer.painter(), t, n == 0 );
	t = n;
    } while ( t != 0 );

    if ( d->scrolls && lstatic->first()->r.left() < 0 ) {
	QPointArray a;
	int h = height();
	if ( d->s == RoundedAbove ) {
	    buffer.painter()->fillRect( 0, 3, 4, h-5,
			colorGroup().brush( QColorGroup::Background ) );
	    a.setPoints( 5,  0,2,  3,h/4, 0,h/2, 3,3*h/4, 0,h );
	} else if ( d->s == RoundedBelow ) {
	    buffer.painter()->fillRect( 0, 2, 4, h-5,
			colorGroup().brush( QColorGroup::Background ) );
	    a.setPoints( 5,  0,0,  3,h/4, 0,h/2, 3,3*h/4, 0,h-3 );
	}

	if ( !a.isEmpty() ) {
	    buffer.painter()->setPen( colorGroup().light() );
	    buffer.painter()->drawPolyline( a );
	    a.translate( 1, 0 );
	    buffer.painter()->setPen( colorGroup().midlight() );
	    buffer.painter()->drawPolyline( a );
	}
    }
}


/*!
  This virtual function is called by the mouse event handlers to
  determine which tab is pressed.  The default implementation returns
  a pointer to the tab whose bounding rectangle contains \a p, if
  exactly one tab's bounding rectangle contains \a p.  Otherwise it returns 0.

  \sa mousePressEvent() mouseReleaseEvent()
*/

ExtQTab * ExtQTabBar::selectTab( const QPoint & p ) const
{
    ExtQTab * selected = 0;
    bool moreThanOne = FALSE;

    QPtrListIterator<ExtQTab> i( *l );
    while( i.current() ) {
	ExtQTab * t = i.current();
	++i;

	if ( t && t->r.contains( p ) ) {
//	    if ( selected )
	    if ( t == tab( currentTab() ) ) /* < EXT > */
		moreThanOne = TRUE;
	    else
		selected = t;
	}
    }

    return moreThanOne ? 0 : selected;
}


/*!\reimp
*/
void ExtQTabBar::mousePressEvent( QMouseEvent * e )
{
    if ( e->button() != LeftButton ) {
	e->ignore();
	return;
    }
    ExtQTab *t = selectTab( e->pos() );
    if ( t && t->enabled ) {
	d->pressed = t;
	if(e->type() == style().styleHint( QStyle::SH_TabBar_SelectMouseType, this ))
	    setCurrentTab( t );
	else
	    repaint(t->rect(), FALSE);
    }
}

/*!\reimp
*/

void ExtQTabBar::mouseMoveEvent ( QMouseEvent *e )
{
    // mouse move : drag this tab
    if ( ( e->state() & Qt::LeftButton ) && selectreleased )
    {
        QDragObject *d = new QTextDrag( "QTWTM", this );
        d->dragCopy();
    }

    else if ( e->button() != LeftButton )
        e->ignore();
    if(style().styleHint( QStyle::SH_TabBar_SelectMouseType, this ) == QEvent::MouseButtonRelease) {
	ExtQTab *t = selectTab( e->pos() );
	if(t != d->pressed) {
	    if(d->pressed)
		repaint(d->pressed->rect(), FALSE);
	    if((d->pressed = t))
		repaint(t->rect(), FALSE);
	}
    }
}

QPopupMenu *ExtQTabBar::pTabMenu = NULL;

void ExtQTabBar::contextMenuEvent( QContextMenuEvent *pEvent )
{
	if (!pTabMenu) return;
	pTabMenu->exec(QCursor::pos());
	pEvent->accept();
	return;
}

/*!\reimp
*/

void ExtQTabBar::dragMoveEvent( QDragMoveEvent *e )
{
    QString text; 
    if ( QTextDrag::decode( e, text ) && ( text == "QTWTM" ) &&
         e->source() == this )
    {
	for( int i = 0; i < (int)lstatic->count(); i++ )
	{
	    if ( ( e->pos().x() >= lstatic->at(i)->r.x() ) && 
	         ( e->pos().x() < lstatic->at(i)->r.x() + lstatic->at(i)->r.width() ) )
	    {
		if ( indexOf( currentTab() ) != i ) e->accept(); else e->ignore();
		break;
	    }
	}
    }
}

/*!\reimp
*/
 
void ExtQTabBar::dropEvent( QDropEvent *e )
{
    QString text; 
    if ( QTextDrag::decode( e, text ) && ( text == "QTWTM" ) &&
         e->source() == this )
    {
	for( int i = 0; i < (int)lstatic->count(); i++ )
	{
	    if ( lstatic->at(i)->r.contains( e->pos() ) )
	    {
		if ( indexOf( currentTab() ) != i )
		{
		    moveTab( tab( currentTab() ), i );
		    repaint();
		    e->acceptAction();
		    emit tabOrderChanged();
		} else e->ignore();
		break;
	    }
	}
    }
}

void ExtQTabBar::mouseReleaseEvent( QMouseEvent *e )
{
    if ( e->button() != LeftButton )
	e->ignore();
    selectreleased = true;
    if(d->pressed) {
	ExtQTab *t = selectTab( e->pos() ) == d->pressed ? d->pressed : NULL;
	d->pressed = NULL;
	if(t && e->type() == style().styleHint( QStyle::SH_TabBar_SelectMouseType, this ))
	    setCurrentTab( t );
    }
}


/*!  \reimp
*/
void ExtQTabBar::show()
{
    //  ensures that one tab is selected.
    ExtQTab * t = l->last();
    QWidget::show();

    if ( t )
	emit selected( t->id );
}

/*! \property ExtQTabBar::currentTab
    \brief the id of the currently visible tab in the tab bar

  If no tab page is currently visible, -1 will be the current value
  for this property.
  Even if the property value is not -1, you cannot assume that
  the user can see the relevant page, or that the tab is enabled.
  When you need to display something the value of this property
  represents the best page to display.

  When this property is set to \e id, it will raise the tab with the
  id \e id and emit the selected() signal.

  \sa selected() isTabEnabled()
*/

int ExtQTabBar::currentTab() const
{
    const ExtQTab * t = l->getLast();

    return t ? t->id : -1;
}

void ExtQTabBar::setCurrentTab( int id )
{
    setCurrentTab( tab( id ) );
}


/*!
    \overload
    Raises \a tab and emits the selected() signal unless the tab was
  already current.

  \sa currentTab() selected()
*/

void ExtQTabBar::setCurrentTab( ExtQTab * tab )
{
    if ( tab && l ) {
	if ( l->last() == tab )
	    return;

	QRect r = l->last()->r;
	if ( l->findRef( tab ) >= 0 )
	    l->append( l->take() );

	d->focus = tab->id;

	setMicroFocusHint( tab->rect().x(), tab->rect().y(), tab->rect().width(), tab->rect().height(), FALSE );

	if ( tab->r.intersects( r ) ) {
	    repaint( r.unite( tab->r ) );
	} else {
	    repaint( r );
	    repaint( tab->r );
	}
	makeVisible( tab );
	emit selected( tab->id );
    }
}

/*! \property ExtQTabBar::keyboardFocusTab
    \brief the id of the tab that currently has the keyboard focus

  This property contains the id of the tab that currently has the
  keyboard focus. If the tab bar does not have keyboard focus, the
  value of this property will be -1.

*/

int ExtQTabBar::keyboardFocusTab() const
{
    return hasFocus() ? d->focus : -1;
}


/*!\reimp
*/
void ExtQTabBar::keyPressEvent( QKeyEvent * e )
{
    //   The right and left arrow keys move a selector, the spacebar
    //   makes the tab with the selector active.  All other keys are
    //   ignored.

    int old = d->focus;

    bool reverse = QApplication::reverseLayout();
    if ( ( !reverse && e->key() == Key_Left ) || ( reverse && e->key() == Key_Right ) ) {
	// left - skip past any disabled ones
	if ( d->focus > 0 ) {
	    ExtQTab * t = lstatic->last();
	    while ( t && t->id != d->focus )
		t = lstatic->prev();
	    do {
		t = lstatic->prev();
	    } while ( t && !t->enabled);
	    if (t)
		d->focus = t->id;
	}
	if ( d->focus < 0 )
	    d->focus = old;
    } else if ( ( !reverse && e->key() == Key_Right ) || ( reverse && e->key() == Key_Left ) ) {
	ExtQTab * t = lstatic->first();
	while ( t && t->id != d->focus )
	    t = lstatic->next();
	do {
	    t = lstatic->next();
	} while ( t && !t->enabled);

	if (t)
	    d->focus = t->id;
	if ( d->focus >= d->id )
	    d->focus = old;
    } else {
	// other keys - ignore
	e->ignore();
	return;
    }

    // if the focus moved, repaint and signal
    if ( old != d->focus ) {
	setCurrentTab( d->focus );
    }
}


/*!  Returns a pointer to the tab with id \a id or 0 if there is no
  such tab.

  \sa count()
*/

ExtQTab * ExtQTabBar::tab( int id ) const
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() )
	if ( t && t->id == id )
	    return t;
    return 0;
}

/*! Returns a pointer to the tab at the position \a index.

  \sa indexOf()
*/

ExtQTab * ExtQTabBar::tabAt( int index ) const
{
    ExtQTab * t;
    t = lstatic->at( index );
    return t;
}


/*!
  Returns the position index of the tab with id \a id.

  \sa tabAt()
 */
int ExtQTabBar::indexOf( int id ) const
{
    ExtQTab * t;
    int idx = 0;
    for( t = lstatic->first(); t; t = lstatic->next() ) {
	if ( t && t->id == id )
	    return idx;
	idx++;
    }
    return -1;
}


/*! \property ExtQTabBar::count
    \brief the number of tabs in the tab bar

  \sa tab()
*/
int ExtQTabBar::count() const
{
    return l->count();
}


/*!
  The list of ExtQTab objects in the tab bar.
*/
QPtrList<ExtQTab> * ExtQTabBar::tabList()
{
    return l;
}


/*! \property ExtQTabBar::shape
    \brief the shape of the tabs in the tab bar

    The value of this property can be one of the following:
    \c RoundedAbove (default), \c RoundedBelow, \c TriangularAbove or \c
    TriangularBelow.

    \sa Shape
*/
ExtQTabBar::Shape ExtQTabBar::shape() const
{
    return d ? d->s : RoundedAbove;
}

void ExtQTabBar::setShape( Shape s )
{
    if ( !d || d->s == s )
	return;
    //######### must recalculate heights
    d->s = s;
    update();
}

/*!
  Lays out all existing tabs according to their label and their
  iconset.
 */
void ExtQTabBar::layoutTabs()
{
    if ( lstatic->isEmpty() )
	return;

    int hframe, vframe, overlap;
    hframe  = style().pixelMetric( QStyle::PM_TabBarTabHSpace, this );
    vframe  = style().pixelMetric( QStyle::PM_TabBarTabVSpace, this );
    overlap = style().pixelMetric( QStyle::PM_TabBarTabOverlap, this );

    QFontMetrics fm = fontMetrics();
    int x = 0;
    QRect r;
    ExtQTab *t;
    bool reverse = QApplication::reverseLayout();
    if ( reverse )
	t = lstatic->last();
    else
	t = lstatic->first();
    while ( t ) {
	int lw = fm.width( t->label );
	lw -= t->label.contains('&') * fm.width('&');
	lw += t->label.contains("&&") * fm.width('&');
	int iw = 0;
	int ih = 0;
	if ( t->iconset != 0 ) {
	    iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 4;
	    ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	}
	int h = QMAX( fm.height(), ih );
	h = QMAX( h, QApplication::globalStrut().height() );

	h += vframe;
	t->r.setRect( x, 0, QMAX( lw + hframe + iw,
				  QApplication::globalStrut().width() ), h );
	x += t->r.width() - overlap;
	r = r.unite( t->r );
	if ( reverse )
	    t = lstatic->prev();
	else
	    t = lstatic->next();
    }
    for ( t = lstatic->first(); t; t = lstatic->next() )
	t->r.setHeight( r.height() );
}

/*!
  \reimp
*/

void ExtQTabBar::styleChange( QStyle& old )
{
	layoutTabs();
	QWidget::styleChange( old );
}

/*!
  \reimp
*/
void ExtQTabBar::focusInEvent( QFocusEvent * )
{
    ExtQTab *t = l->first();
    for ( ; t; t = l->next() ) {
	if ( t->id == d->focus ) {
	    QPainter p;
	    p.begin( this );
	    QRect r = t->r;
	    p.setFont( font() );

	    int iw = 0;
	    int ih = 0;
	    if ( t->iconset != 0 ) {
		iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 4;
		ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	    }
	    QFontMetrics fm = p.fontMetrics();
	    int fw = fm.width( t->label );
	    fw -= t->label.contains('&') * fm.width('&');
	    fw += t->label.contains("&&") * fm.width('&');
	    int w = iw + fw + 4;
	    int h = QMAX(fm.height() + 4, ih );
	    paintLabel( &p, QRect( r.left() + ( r.width() -w ) /2 - 3,
				   r.top() + ( r.height()-h ) / 2,
				   w, h ), t, TRUE );
	    p.end();
	}
    }
}

/*!
  \reimp
*/
void ExtQTabBar::focusOutEvent( QFocusEvent * )
{
    ExtQTab *t = l->first();
    for ( ; t; t = l->next() ) {
	if ( t->id == d->focus ) {
	    QPainter p;
	    p.begin( this );
	    p.setBrushOrigin( rect().bottomLeft() );
	    QRect r = t->r;
	    p.setFont( font() );

	    int iw = 0;
	    int ih = 0;
	    if ( t->iconset != 0 ) {
		iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 4;
		ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	    }
	    QFontMetrics fm = p.fontMetrics();
	    int fw = fm.width( t->label );
	    fw -= t->label.contains('&') * fm.width('&');
	    fw += t->label.contains("&&") * fm.width('&');
	    int w = iw + fw + 4;
	    int h = QMAX(fm.height() + 4, ih );
	    p.fillRect( QRect( r.left() + ( r.width() -w ) / 2 - 4,
				   r.top() + ( r.height()-h ) / 2 - 1,
			       w + 4, h + 2 ), colorGroup().brush(QColorGroup::Background ) );

	    QStyle::SFlags flags = QStyle::Style_Default;
	    flags |= QStyle::Style_Selected;
	    style().drawControl( QStyle::CE_TabBarTab, &p, this, t->rect(),
				 colorGroup(), flags, QStyleOption((QTab *)t) );

	    paintLabel( &p, QRect( r.left() + ( r.width() -w ) /2 - 3,
				   r.top() + ( r.height()-h ) / 2,
				   w, h ), t, FALSE );
	    p.end();
	}
    }
}

/*!
  \reimp
*/
void ExtQTabBar::resizeEvent( QResizeEvent * )
{
    const int arrowWidth = 16;
    d->rightB->setGeometry( width() - arrowWidth, 0, arrowWidth, height() );
    d->leftB->setGeometry( width() - 2*arrowWidth, 0, arrowWidth, height() );
    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ));
}

void ExtQTabBar::scrollTabs()
{
    ExtQTab* left = 0;
    ExtQTab* right = 0;
    for ( ExtQTab* t = lstatic->first(); t; t = lstatic->next() ) {
	if ( t->r.left() < 0 && t->r.right() > 0 )
	    left = t;
	if ( t->r.left() < d->leftB->x()+2 )
	    right = t;
    }

    if ( sender() == d->leftB )
	makeVisible( left );
    else  if ( sender() == d->rightB )
	makeVisible( right );
}

void ExtQTabBar::makeVisible( ExtQTab* tab  )
{
    bool tooFarLeft = ( tab && tab->r.left() < 0 );
    bool tooFarRight = ( tab && tab->r.right() >= d->leftB->x() );

    if ( !d->scrolls || ( !tooFarLeft && ! tooFarRight ) )
	return;

    layoutTabs();

    int offset = 0;

    if ( tooFarLeft )
	offset = tab == lstatic->first() ? 0 : tab->r.left() - 8;
    else if ( tooFarRight ) {
	offset = tab->r.right() - d->leftB->x() + 1;
    }

    for ( ExtQTab* t = lstatic->first(); t; t = lstatic->next() )
	t->r.moveBy( -offset, 0 );

    d->leftB->setEnabled( offset != 0 );
    d->rightB->setEnabled( lstatic->last()->r.right() >= d->leftB->x() );

    // Make sure disabled buttons pop up again
    if ( !d->leftB->isEnabled() && d->leftB->isDown() )
	d->leftB->setDown( FALSE );
    if ( !d->rightB->isEnabled() && d->rightB->isDown() )
	d->rightB->setDown( FALSE );

    update();
}

void ExtQTabBar::updateArrowButtons()
{
    bool b = lstatic->last() &&	( lstatic->last()->r.right() > width() );
    d->scrolls = b;
    if ( d->scrolls ) {
	d->leftB->setEnabled( FALSE );
	d->rightB->setEnabled( TRUE );
	d->leftB->show();
	d->rightB->show();
    } else {
	d->leftB->hide();
	d->rightB->hide();
    }
}

/*! Removes the tool tip for the tab at index \a index.
 */
void ExtQTabBar::removeToolTip( int index )
{
#ifndef QT_NO_TOOLTIP
    ExtQTab * tab = tabAt( index );
    if ( !tab || !d->toolTips )
	return;
    d->toolTips->remove( tab );
#endif
}

/*! Sets the tool tip for the tab at index \a index to \a tip.
 */
void ExtQTabBar::setToolTip( int index, const QString & tip )
{
#ifndef QT_NO_TOOLTIP
    ExtQTab * tab = tabAt( index );
    if ( !tab )
	return;
    if ( d->toolTips == 0 )
	d->toolTips = new ExtQTabBarToolTip( this );
    d->toolTips->add( tab, tip );
#endif
}

/*! Returns the tool tip for the tab at index \a index.
 */
QString ExtQTabBar::toolTip( int index ) const
{
#ifndef QT_NO_TOOLTIP
    if ( d->toolTips )
	return d->toolTips->tipForTab( tabAt( index ) );
    else
#endif
	return QString();
}

/*!
  Sets the text of the tab to \a text.
 */
void ExtQTab::setText( const QString& text )
{
    label = text;
    if ( tb ) {
#ifndef QT_NO_ACCEL
	tb->d->a->removeItem( id );
	int p = QAccel::shortcutKey( text );
	if ( p )
	    tb->d->a->insertItem( p, id );
#endif
	tb->layoutTabs();
	tb->repaint();
    }
}

/*!
  Sets the tab iconset to \a icon
*/
void ExtQTab::setIconSet( const QIconSet &icon )
{
    iconset = new QIconSet( icon );
}

// this allows us to handle accelerators that are in a ExtQTabBar.
void ExtQTab::setTabBar( ExtQTabBar *newTb )
{
    tb = newTb;
}

/*!
    \internal
*/
void ExtQTabBar::fontChange( const QFont & oldFont )
{
    layoutTabs();
    QWidget::fontChange( oldFont );
}


/* ------------------------------------------------------------------- */

#include "QTExt.h"
#include "qobjectlist.h"
#include "qapplication.h"
#include "qwidgetstack.h"
#include "qbitmap.h"
#include "qaccel.h"
#include "qstyle.h"
#include "qpainter.h"

/*!
  \class ExtQTabWidget qtabwidget.h

  \brief The ExtQTabWidget class provides a stack of tabbed widgets.

  \ingroup organizers
  \mainclass
  \ingroup advanced

  A tabbed widget is a widget that has a tab bar of tabs, and for each
  tab a "page" which is a widget. The user selects which page to see
  and use by clicking on its tab or by pressing the indicated
  Alt+\e{letter} key combination.

  ExtQTabWidget provides a single row of tabs along the top or bottom of
  the pages (see \l{TabPosition}).

  The normal way to use ExtQTabWidget is to do the following in the
  constructor:
  \list 1
  \i Create a ExtQTabWidget.
  \i Create a QWidget for each of the pages in the tab dialog, insert
  children into it, set up geometry management for it and use addTab()
  (or insertTab()) to set up a tab and keyboard accelerator for it.
  \i Connect to the signals and slots.
  \endlist

  The position of the tabs is set with setTabPosition(), their
  shape with setTabShape(), and their margin with setMargin().

  If you don't call addTab() the page you have created will not be
  visible.  Don't confuse the object name you supply to the
  QWidget constructor and the tab label you supply to addTab().
  addTab() takes a name which indicates an accelerator and is
  meaningful and descriptive to the user, whereas the widget name is
  used primarily for debugging.

  The signal currentChanged() is emitted when the user selects a page.

  The current page is available as an index position with
  currentPageIndex() or as a wiget pointer with currentPage(). You can
  retrieve a pointer to a page with a given index using page(), and
  can find the index position of a page with indexOf(). Use
  setCurrentPage() to show a particular page by index, or showPage()
  to show a page by widget pointer.

  You can change a tab's label and iconset using changeTab() or
  setTabLabel() and setTabIconSet(). A tab page can be removed with
  removePage().

  Each tab is either enabled or disabled at any given time (see
  setTabEnabled()).  If a tab is enabled, the tab text is drawn in
  black and the user can select that tab.  If it is disabled, the tab
  is drawn in a different way and the user cannot select that tab.
  Note that even if a tab is disabled, the page can still be visible,
  for example if all of the tabs happen to be disabled.

  Although tab widgets can be a very good way to split up a complex
  dialog, it's also very easy to get into a mess. See ExtQTabDialog for
  some design hints.

  Most of the functionality in ExtQTabWidget is provided by a ExtQTabBar (at
  the top, providing the tabs) and a QWidgetStack (most of the area,
  organizing the individual pages).

  <img src=qtabwidget-m.png> <img src=qtabwidget-w.png>

  \sa ExtQTabDialog
*/


/*! \enum ExtQTabWidget::TabPosition

  This enum type defines where ExtQTabWidget can draw the tab row:
  \value Top  above the pages
  \value Bottom  below the pages
*/

/*! \enum ExtQTabWidget::TabShape

  This enum type defines the shape of the tabs:
  \value Rounded  rounded look (normal)
  \value Triangular  triangular look (very unusual, included for completeness)
*/

/* undocumented now
  \obsolete

  \fn void ExtQTabWidget::selected( const QString &tabLabel );

  This signal is emitted whenever a tab is selected (raised),
  including during the first show().

  \sa raise()
*/


/*! \fn void ExtQTabWidget::currentChanged( QWidget* );

  This signal is emitted whenever the current page changes. The
  parameter is the new current page.

  \sa currentPage(), showPage(), tabLabel()
*/

class ExtQTabBarBase : public QWidget
{
public:
    ExtQTabBarBase( ExtQTabWidget* parent=0, const char* name=0 )
        : QWidget( parent, name ) {};
protected:
    void paintEvent( QPaintEvent * )
    {
        QObject * obj = parent();
        if( obj ){
            ExtQTabWidget * t = (ExtQTabWidget *) obj;
            QPainter p( this );
	    QStyle::SFlags flags = QStyle::Style_Default;

	    if ( t->tabPosition() == ExtQTabWidget::Top )
		flags |= QStyle::Style_Top;
	    if ( t->tabPosition() == ExtQTabWidget::Bottom )
		flags |= QStyle::Style_Bottom;

	    style().drawPrimitive( QStyle::PE_TabBarBase, &p, rect(),
				   colorGroup(), flags );
        }
    }
};

class ExtQTabWidgetData
{
public:
    ExtQTabWidgetData()
        : tabs(0), tabBase(0), stack(0), dirty( TRUE ),
          pos( ExtQTabWidget::Top ), shape( ExtQTabWidget::Rounded ) {};
    ~ExtQTabWidgetData(){};
    ExtQTabBar* tabs;
    ExtQTabBarBase* tabBase;
    QWidgetStack* stack;
    bool dirty;
    ExtQTabWidget::TabPosition pos;
    ExtQTabWidget::TabShape shape;
    int alignment;
};

/*!
  Constructs a tabbed widget with parent \a parent, name \a name,
  and widget flags \a f.
*/
ExtQTabWidget::ExtQTabWidget( QWidget *parent, const char *name, WFlags f )
    : QWidget( parent, name, f )
{
    d = new ExtQTabWidgetData;

    d->stack = new QWidgetStack( this, "tab pages" );
    d->stack->installEventFilter( this );
    d->tabBase = new ExtQTabBarBase( this, "tab base" );
    d->tabBase->resize( 1, 1 );
    setTabBar( new ExtQTabBar( this, "tab control" ) );

    d->stack->setFrameStyle( QFrame::TabWidgetPanel | QFrame::Raised );
    d->stack->setLineWidth( style().pixelMetric(QStyle::PM_DefaultFrameWidth, this) );

    setSizePolicy( QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding ) );
    setFocusPolicy( TabFocus );
    setFocusProxy( d->tabs );
}

/*!\reimp
*/
ExtQTabWidget::~ExtQTabWidget()
{
    delete d;
}

/*!
  Adds another tab and page to the tab view.

    The new page is \a child; the tab's label is \a label.
  Note the difference between the widget name (which you supply to
  widget constructors and to setTabEnabled(), for example) and the tab
  label. The name is internal to the program and invariant, whereas
  the label is shown on-screen and may vary according to language and
  other factors.

    If the tab's \a label contains an ampersand, the letter following
    the ampersand is used as an accelerator for the tab, e.g. if the
    label is "Bro&wse" then Alt+W becomes an accelerator which will
    move the focus to this tab.

  If you call addTab() after show() the screen will flicker and the
  user may be confused.

  \sa insertTab()
*/
void ExtQTabWidget::addTab( QWidget *child, const QString &label)
{
    ExtQTab * t = new ExtQTab();
    Q_CHECK_PTR( t );
    t->label = label;
    addTab( child, t );
}


/*!
    \overload
  Adds another tab and page to the tab view.

  This function is the same as addTab(), but with an additional
  \a iconset.
 */
void ExtQTabWidget::addTab( QWidget *child, const QIconSet& iconset, const QString &label )
{
    ExtQTab * t = new ExtQTab();
    Q_CHECK_PTR( t );
    t->label = label;
    t->iconset = new QIconSet( iconset );
    addTab( child, t );
}

/*! \overload

  This is a low-level function for adding tabs. It is useful if you
  are using setTabBar() to set a ExtQTabBar subclass with an overridden
  ExtQTabBar::paint() routine for a subclass of ExtQTab. The \a child is
  the new page and \a tab is the tab to put the \a child on.
*/
void ExtQTabWidget::addTab( QWidget *child, ExtQTab* tab )
{
    tab->enabled = TRUE;
    int id = d->tabs->addTab( tab );
    d->stack->addWidget( child, id );
    if ( d->stack->frameStyle() != ( QFrame::TabWidgetPanel | QFrame::Raised ) )
        d->stack->setFrameStyle( QFrame::TabWidgetPanel | QFrame::Raised );
    setUpLayout();
}



/*!
  Inserts another tab and page to the tab view.

    The new page is \a child; the tab's label is \a label.
  Note the difference between the widget name (which you supply to
  widget constructors and to setTabEnabled(), for example) and the tab
  label. The name is internal to the program and invariant, whereas
  the label is shown on-screen and may vary according to language and
  other factors.

    If the tab's \a label contains an ampersand, the letter following
    the ampersand is used as an accelerator for the tab, e.g. if the
    label is "Bro&wse" then Alt+W becomes an accelerator which will
    move the focus to this tab.

  If \a index is not specified, the tab is simply added. Otherwise
  it is inserted at the specified position.

  If you call insertTab() after show(), the screen will flicker and the
  user may be confused.

  \sa addTab()
*/
void ExtQTabWidget::insertTab( QWidget *child, const QString &label, int index)
{
    ExtQTab * t = new ExtQTab();
    Q_CHECK_PTR( t );
    t->label = label;
    insertTab( child, t, index );
}


/*!
    \overload
  Inserts another tab and page to the tab view.

  This function is the same as insertTab(), but with an additional
  \a iconset.
 */
void ExtQTabWidget::insertTab( QWidget *child, const QIconSet& iconset, const QString &label, int index )
{
    ExtQTab * t = new ExtQTab();
    Q_CHECK_PTR( t );
    t->label = label;
    t->iconset = new QIconSet( iconset );
    insertTab( child, t, index );
}

/*!
    \overload
  This is a lower-level method for inserting tabs, similar to the other
  insertTab() method.  It is useful if you are using setTabBar() to set a
  ExtQTabBar subclass with an overridden ExtQTabBar::paint() routine for a
  subclass of ExtQTab.
  The \a child is the new page, \a tab is the tab to put the \a child
  on and \a index is the position in the tab bar that this page should
  occupy.
*/
void ExtQTabWidget::insertTab( QWidget *child, ExtQTab* tab, int index)
{
    tab->enabled = TRUE;
    int id = d->tabs->insertTab( tab, index );
    d->stack->addWidget( child, id );
    if ( d->stack->frameStyle() != ( QFrame::TabWidgetPanel | QFrame::Raised ) )
        d->stack->setFrameStyle( QFrame::TabWidgetPanel | QFrame::Raised );
    setUpLayout();
}


/*!
  Defines a new \a label for page \a{w}'s tab.
 */
void ExtQTabWidget::changeTab( QWidget *w, const QString &label)
{
    int id = d->stack->id( w );
    if ( id < 0 )
        return;
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
        return;
    // this will update the accelerators
    t->setText( label );

    int ct = d->tabs->currentTab();
    bool block = d->tabs->signalsBlocked();
    d->tabs->blockSignals( TRUE );
    d->tabs->setCurrentTab( 0 );
    d->tabs->setCurrentTab( ct );
    d->tabs->blockSignals( block );

    d->tabs->update();
    setUpLayout();
}

/*!
    \overload
  Defines a new \a iconset and a new \a label for page \a{w}'s tab.
 */
void ExtQTabWidget::changeTab( QWidget *w, const QIconSet& iconset, const QString &label)
{
    int id = d->stack->id( w );
    if ( id < 0 )
        return;
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
        return;
    if ( t->iconset ) {
        delete t->iconset;
        t->iconset = 0;
    }
    // this will update the accelerators
    t->iconset = new QIconSet( iconset );
    t->setText( label );

    int ct = d->tabs->currentTab();
    bool block = d->tabs->signalsBlocked();
    d->tabs->blockSignals( TRUE );
    d->tabs->setCurrentTab( 0 );
    d->tabs->setCurrentTab( ct );
    d->tabs->blockSignals( block );

    d->tabs->update();
    setUpLayout();
}

void ExtQTabWidget::moveTab( QWidget *w, int index )
{
    d->tabs->moveTab( d->tabs->tab( d->stack->id( w ) ), index );
}


/*!
  Returns TRUE if the page \a w is enabled; otherwise returns FALSE.

  \sa setTabEnabled(), QWidget::isEnabled()
*/

bool ExtQTabWidget::isTabEnabled( QWidget* w ) const
{
    int id = d->stack->id( w );
    if ( id >= 0 )
        return w->isEnabled();
    else
        return FALSE;
}

/*!
    If \a enable is TRUE, page \a w is enabled; otherwise page \a w is
    disabled. The page's tab is redrawn appropriately.

  ExtQTabWidget uses QWidget::setEnabled() internally, rather than keeping a
  separate flag.

  Note that even a disabled tab/page may be visible.  If the page is
  visible already, ExtQTabWidget will not hide it; if all the pages
  are disabled, ExtQTabWidget will show one of them.

  \sa isTabEnabled(), QWidget::setEnabled()
*/

void ExtQTabWidget::setTabEnabled( QWidget* w, bool enable)
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
        w->setEnabled( enable );
        d->tabs->setTabEnabled( id, enable );
    }
}

/*!  Ensures that page \a w is shown.  This is useful mainly for accelerators.

  \warning Used carelessly, this function can easily surprise or
  confuse the user.

  \sa ExtQTabBar::setCurrentTab()
*/
void ExtQTabWidget::showPage( QWidget * w)
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
        d->stack->raiseWidget( w );
        d->tabs->setCurrentTab( id );
        // ### why overwrite the frame style?
        if ( d->stack->frameStyle() != ( QFrame::TabWidgetPanel |QFrame::Raised ) )
            d->stack->setFrameStyle( QFrame::TabWidgetPanel | QFrame::Raised );
    }
}

/*! Removes page \a w from this stack of widgets.  Does not
  delete \a w.
  \sa showPage(), QWidgetStack::removeWidget()
*/
void ExtQTabWidget::removePage( QWidget * w )
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
	d->tabs->setTabEnabled( id, FALSE );
	d->stack->removeWidget( w );
	d->tabs->removeTab( d->tabs->tab(id) );
	setUpLayout();

	if ( d->tabs->count() == 0 )
	    d->stack->setFrameStyle( QFrame::NoFrame );
    }
}

/*!  Returns the label text for the tab on page \a w.
*/

QString ExtQTabWidget::tabLabel( QWidget * w ) const
{
    ExtQTab * t = d->tabs->tab( d->stack->id( w ) );
    return t ? t->label : QString::null;
}

/*!  Sets the tab label for page \a w to \a l
 */

void ExtQTabWidget::setTabLabel( QWidget * w, const QString &l )
{
    ExtQTab * t = d->tabs->tab( d->stack->id( w ) );
    if ( t )
        t->label = l;
    d->tabs->layoutTabs();
    d->tabs->update();
    setUpLayout();
}

/*! Returns a pointer to the page currently being displayed by the
    tab dialog.  The tab dialog does its best to make sure that this value
    is never 0 (but if you try hard enough, it can be).
*/

QWidget * ExtQTabWidget::currentPage() const
{
    return d->stack->visibleWidget();
}

/*! \property ExtQTabWidget::autoMask
    \brief whether the tab widget is automatically masked

    \sa QWidget::setAutoMask()
*/

/*! \property ExtQTabWidget::currentPage
    \brief the index position of the current tab page

  \sa ExtQTabBar::currentTab()
*/

int ExtQTabWidget::currentPageIndex() const
{
    return d->tabs->indexOf( d->tabs->currentTab() );
}

void ExtQTabWidget::setCurrentPage( int index )
{
    d->tabs->setCurrentTab( d->tabs->tabAt( index ) );
    showTab( d->tabs->currentTab() );
}


/*!
  Returns the index position of page \a w, or -1 if the widget cannot
  be found.
 */
int ExtQTabWidget::indexOf( QWidget* w ) const
{
    return d->tabs->indexOf( d->stack->id( w ) );
}


/*!
  \reimp
 */
void ExtQTabWidget::resizeEvent( QResizeEvent *e )
{
    QWidget::resizeEvent( e );
    setUpLayout();
}

/*!
  Replaces the ExtQTabBar heading the dialog by the tab bar \a tb.
  Note that this must be called \e before any tabs have been added,
  or the behavior is undefined.
  \sa tabBar()
*/
void ExtQTabWidget::setTabBar( ExtQTabBar* tb)
{
    if ( tb->parentWidget() != this )
        tb->reparent( this, QPoint(0,0), TRUE );
    delete d->tabs;
    d->tabs = tb;
    setFocusProxy( d->tabs );
    connect( d->tabs, SIGNAL(selected(int)),
             this,    SLOT(showTab(int)) );
    setUpLayout();
}


/*!
  Returns the currently set ExtQTabBar.
  \sa setTabBar()
*/
ExtQTabBar* ExtQTabWidget::tabBar() const
{
    return d->tabs;
}

/*!
  Ensures that the selected tab's page is visible and appropriately sized.
*/

void ExtQTabWidget::showTab( int i )
{
    if ( d->stack->widget( i ) ) {
	d->stack->raiseWidget( i );
	emit selected( d->tabs->tab( i )->label );
	emit currentChanged( d->stack->widget( i ) );
    }
}

/*!
  Set up the layout.
 */
void ExtQTabWidget::setUpLayout( bool onlyCheck )
{
    if ( onlyCheck && !d->dirty )
	return; // nothing to do

    if ( !isVisible() ) {
	d->dirty = TRUE;
	return; // we'll do it later
    }

    QSize t( d->tabs->sizeHint() );
    if ( t.width() > width() )
	t.setWidth( width() );
    int lw = d->stack->lineWidth();
    bool reverse = QApplication::reverseLayout();
    int tabx, taby, stacky, exty, exth, overlap;

    exth = style().pixelMetric( QStyle::PM_TabBarBaseHeight, this );
    overlap = style().pixelMetric( QStyle::PM_TabBarBaseOverlap, this );

    if( reverse ) {
	tabx = QMIN( width() - t.width(), width() - t.width() - lw + 2 );
    } else {
	tabx = QMAX( 0, lw - 2 );
    }
    if ( d->pos == Bottom ) {
	taby = height() - t.height() - lw;
	stacky = 0;
	exty = taby - (exth - overlap);
    } else { // Top
	taby = 0;
	stacky = t.height()-lw + (exth - overlap), width();
	exty = taby + t.height() - overlap;
    }

    // do alignment
    int alignment = style().styleHint( QStyle::SH_TabBar_Alignment, this );
    if ( alignment != AlignLeft && t.width() < width() ) {
	if ( alignment == AlignHCenter )
	    tabx += width()/2 - t.width()/2;
	else if ( alignment == AlignRight )
	    tabx += width() - t.width();
    }

    d->tabs->setGeometry( tabx, taby, t.width(), t.height() );
    d->tabBase->setGeometry( 0, exty, width(), exth );
    if ( exth == 0 )
	d->tabBase->hide();
    else
	d->tabBase->show();

    d->stack->setGeometry( 0, stacky, width(), height() - (exth-overlap) -
			   t.height()+QMAX(0, lw-2));

    d->dirty = FALSE;
    if ( !onlyCheck )
	update();
    if ( autoMask() )
	updateMask();
}

/*!\reimp
*/
QSize ExtQTabWidget::sizeHint() const
{
    if ( !d->dirty ) {
	ExtQTabWidget *that = (ExtQTabWidget*)this;
	that->setUpLayout( TRUE );
    }
    QSize s( d->stack->sizeHint() );
    QSize t( d->tabs->sizeHint().boundedTo( QSize(200,200) ) );
    return QSize( QMAX( s.width(), t.width() ),
		  s.height() + t.height() + d->tabBase->height() );
}


/*! \reimp
  Returns a suitable minimum size for the tab widget.
*/
QSize ExtQTabWidget::minimumSizeHint() const
{
    if ( !d->dirty ) {
	ExtQTabWidget *that = (ExtQTabWidget*)this;
	that->setUpLayout( TRUE );
    }
    QSize s( d->stack->minimumSizeHint() );
    QSize t( d->tabs->minimumSizeHint() );
    return QSize( QMAX( s.width(), t.width() ),
		  s.height() + t.height() + d->tabBase->height() );
}

/*! \reimp
 */
void ExtQTabWidget::showEvent( QShowEvent * )
{
    setUpLayout();
}


/*! \property ExtQTabWidget::tabPosition
    \brief the position of the tabs in this tab widget

  Possible values for this property are ExtQTabWidget::Top and
  ExtQTabWidget::Bottom.

  \sa TabPosition
 */
ExtQTabWidget::TabPosition ExtQTabWidget::tabPosition() const
{
    return d->pos;
}

void ExtQTabWidget::setTabPosition( TabPosition pos)
{
    if (d->pos == pos)
        return;
    d->pos = pos;
    if (d->tabs->shape() == ExtQTabBar::TriangularAbove || d->tabs->shape() == ExtQTabBar::TriangularBelow ) {
        if ( pos == Bottom )
            d->tabs->setShape( ExtQTabBar::TriangularBelow );
        else
            d->tabs->setShape( ExtQTabBar::TriangularAbove );
    }
    else {
        if ( pos == Bottom )
            d->tabs->setShape( ExtQTabBar::RoundedBelow );
        else
            d->tabs->setShape( ExtQTabBar::RoundedAbove );
    }
    d->tabs->layoutTabs();
    setUpLayout();
}

/*! \property ExtQTabWidget::tabShape
    \brief the shape of the tabs in this tab widget

    Possible values for this property are ExtQTabWidget::Rounded (default) or
    ExtQTabWidget::Triangular.

    \sa TabShape
*/

ExtQTabWidget::TabShape ExtQTabWidget::tabShape() const
{
    return d->shape;
}

void ExtQTabWidget::setTabShape( TabShape s )
{
    if ( d->shape == s )
        return;
    d->shape = s;
    if ( d->pos == Top ) {
        if ( s == Rounded )
            d->tabs->setShape( ExtQTabBar::RoundedAbove );
        else
            d->tabs->setShape( ExtQTabBar::TriangularAbove );
    } else {
        if ( s == Rounded )
            d->tabs->setShape( ExtQTabBar::RoundedBelow );
        else
            d->tabs->setShape( ExtQTabBar::TriangularBelow );
    }
    d->tabs->layoutTabs();
    setUpLayout();
}


/*! \property ExtQTabWidget::margin
    \brief the margin in this tab widget

  The margin is the distance between the innermost pixel of the frame
  and the outermost pixel of the pages.
*/
int ExtQTabWidget::margin() const
{
    return d->stack->margin();
}

void ExtQTabWidget::setMargin( int w )
{
    d->stack->setMargin( w );
    setUpLayout();
}


/*! \reimp
 */
void ExtQTabWidget::styleChange( QStyle& old )
{
    d->stack->setLineWidth( style().pixelMetric(QStyle::PM_DefaultFrameWidth, this));
    setUpLayout();
    QWidget::styleChange( old );
}


/*! \reimp
 */
void ExtQTabWidget::updateMask()
{
    if ( !autoMask() )
        return;

    QRect r;
    QRegion reg( r );
    reg += QRegion( d->tabs->geometry() );
    reg += QRegion( d->stack->geometry() );
    setMask( reg );
}


/*!\reimp
 */
bool ExtQTabWidget::eventFilter( QObject *o, QEvent * e)
{
    if ( o == d->stack && e->type() == QEvent::ChildRemoved
         && ( (QChildEvent*)e )->child()->isWidgetType() ) {
        removePage( (QWidget*)  ( (QChildEvent*)e )->child() );
        return TRUE;
    }
    return FALSE;
}

/*!
  Returns the tab page at index position \a index.
*/
QWidget *ExtQTabWidget::page( int index ) const
{
    ExtQTab *t = d->tabs->tabAt(index);
    if ( t )
	return d->stack->widget( t->id );
    // else
    return 0;
}

/*!
  Returns the label of the tab at index position \a index.
*/
QString ExtQTabWidget::label( int index ) const
{
    ExtQTab *t = d->tabs->tabAt( index );
    if ( t )
 	return t->label;
    // else
    return QString::null;
}

/*! \property ExtQTabWidget::count
     \brief the number of tabs in the tab bar
*/
int ExtQTabWidget::count() const
{
    return d->tabs->count();
}

/*!
  Returns the iconset of page \a w.
*/
QIconSet ExtQTabWidget::tabIconSet( QWidget * w ) const
{
    int id = d->stack->id( w );
    if ( id < 0 )
        return QIconSet();
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
        return QIconSet();
    if ( t->iconset )
	return QIconSet( *t->iconset );
    else
	return QIconSet();
}

/*!
  Sets the iconset for page \a w to \a iconset.
*/
void ExtQTabWidget::setTabIconSet( QWidget * w, const QIconSet & iconset )
{
    int id = d->stack->id( w );
    if ( id < 0 )
        return;
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
        return;
    if ( t->iconset )
        delete t->iconset;
    t->iconset = new QIconSet( iconset );

    d->tabs->layoutTabs();

    int ct = d->tabs->currentTab();
    bool block = d->tabs->signalsBlocked();
    d->tabs->blockSignals( TRUE );
    d->tabs->setCurrentTab( 0 );
    d->tabs->setCurrentTab( ct );
    d->tabs->blockSignals( block );

    d->tabs->update();
    setUpLayout();
}

/*!
  Sets the tab tool tip for page \a w to \a tip.
  \sa removeTabToolTip(), tabToolTip()
 */
void ExtQTabWidget::setTabToolTip( QWidget * w, const QString & tip )
{
    int index = d->tabs->indexOf( d->stack->id( w ) );
    if ( index < 0 )
        return;
    d->tabs->setToolTip( index, tip );
}

/*!
  Returns the tab tool tip for page \a w.
  \sa setTabToolTip(), removeTabToolTip()
 */
QString ExtQTabWidget::tabToolTip( QWidget * w ) const
{
    int index = d->tabs->indexOf( d->stack->id( w ) );
    if ( index < 0 )
        return QString();
    return d->tabs->toolTip( index );
}

/*! Removes the tab tool tip for page \a w. If the page does not have
  a tip, nothing happens.
  \sa setTabToolTip(), tabToolTip()
 */
void ExtQTabWidget::removeTabToolTip( QWidget * w )
{
    int index = d->tabs->indexOf( d->stack->id( w ) );
    if ( index < 0 )
        return;
    d->tabs->removeToolTip( index );
}


#else

#include "QTExt.h"
#ifndef QT_NO_TABBAR
#include "qaccel.h"
#include "qbitmap.h"
#include "qtoolbutton.h"
#include "qapplication.h"

#include <ctype.h>
#include <qdragobject.h>


// NOT REVISED
/*!
  \class ExtQTab qtabbar.h
  \brief The structures in a ExtQTabBar.

  For custom ExtQTabBar tab headings.

  \sa ExtQTabBar
*/


/*! \fn ExtQTab::ExtQTab()
  Constructs an empty tab.  All fields are set to empty.
*/


/*! \fn ExtQTab::ExtQTab( const QString& text )
  Constructs a tab with a \a text.
*/


/*! \fn ExtQTab::ExtQTab( const QIconSet& icon, const QString& text )
  Constructs a tab with an \a icon and a \a text.
*/


/*! Destructs the tab and frees up all allocated resources */

ExtQTab::~ExtQTab()
{
    delete iconset;
}


/*!
  \class ExtQTabBar qtabbar.h

  \brief The ExtQTabBar class provides a tab bar, for use in e.g. tabbed
  dialogs.

  \ingroup advanced

  The class is quite simple; it draws the tabs in one of four shapes
  and emits a signal when one is selected.  It can be subclassed to
  tailor the look and feel.

  ExtQTabBar itself support four possible shapes, described in the
  ExtQTabBar::Shape documentation.

  The choice of tab shape is still a matter of taste, to a large
  degree.  Tab dialogs (preferences and the like) invariably use \c
  RoundedAbove and nobody uses \c TriangularAbove.  Tab controls in
  windows other than dialogs almost always either \c RoundedBelow or
  \c TriangularBelow.  Many spreadsheets and other tab controls where
  all the pages are essentially similar to use \c TriangularBelow,
  while \c RoundedBelow is used mostly when the pages are different
  (e.g. a multi-page tool palette).  There is no strong tradition yet,
  however, so use your taste and create the tradition.

  The most important part of ExtQTabBar's API is the signal selected().
  It's emitted whenever the selected page changes (even at startup,
  when the selected page changes from 'none').  There are also a slot,
  setCurrentTab(), which can be used to select a page
  programmatically.

  ExtQTabBar creates automatic accelerator keys in the manner of QButton;
  e.g. if a tab's label is "\&Graphics" Alt-G becomes an accelerator
  key for switching to that tab.

  The following virtual functions may need to be reimplemented: <ul>
  <li> paint() paints a single tab.  paintEvent() calls paint() for
  each tab in such a way that any overlap will look right.  <li>
  addTab() creates a new tab and adds it to the bar. <li> selectTab()
  decides which, if any, tab the user selects with the mouse. </ul>

  <img src=qtabbar-m.png> <img src=qtabbar-w.png>
*/

/*! \enum ExtQTabBar::Shape
  This enum type lists the built-in shapes supported by ExtQTabBar:<ul>

  <li> \c RoundedAbove - the normal rounded look, above the pages

  <li> \c RoundedBelow - the normal rounded look, below the pages

  <li> \c TriangularAbove - triangular tabs, above the pages (very
  unusual, included for completeness)

  <li> \c TriangularBelow - triangular tabs, similar to those used in
  e.g. the spreadsheet Excel

  </ul>
*/

struct ExtQTabPrivate {
    int id;
    int focus;
    QAccel * a;
    ExtQTabBar::Shape s;
    QToolButton* rightB;
    QToolButton* leftB;
    bool  scrolls;
};


/*!
  \fn void ExtQTabBar::selected( int id )

  ExtQTabBar emits this signal whenever any tab is selected, whether by
  the program or the user.  The argument \a id is the ID if the tab
  as returned by addTab().

  show() is guaranteed to emit this signal, so that you can display
  your page in a slot connected to this signal.
*/


/*!
  Constructs a new, empty tab bar.
*/

ExtQTabBar::ExtQTabBar( QWidget * parent, const char *name )
    : QWidget( parent, name, WRepaintNoErase | WResizeNoErase  )
{
    d = new ExtQTabPrivate;
    d->id = 0;
    d->focus = 0;
    d->a = new QAccel( this, "tab accelerators" );
    d->s = RoundedAbove;
    d->scrolls = FALSE;
    d->leftB = new QToolButton( LeftArrow, this );
    connect( d->leftB, SIGNAL( clicked() ), this, SLOT( scrollTabs() ) );
    d->leftB->hide();
    d->rightB = new QToolButton( RightArrow, this );
    connect( d->rightB, SIGNAL( clicked() ), this, SLOT( scrollTabs() ) );
    d->rightB->hide();
    l = new QList<ExtQTab>;
    lstatic = new QList<ExtQTab>;
    lstatic->setAutoDelete( TRUE );
    setFocusPolicy( TabFocus );
    setSizePolicy( QSizePolicy( QSizePolicy::Preferred, QSizePolicy::Fixed ) );
    setAcceptDrops( true );
    selectreleased = true;

    connect( d->a, SIGNAL(activated(int)), this, SLOT(setCurrentTab(int)) );
}


/*!
  Destroys the tab control, freeing memory used.
*/

ExtQTabBar::~ExtQTabBar()
{
    delete d;
    d = 0;
    delete l;
    l = 0;
    delete lstatic;
    lstatic = 0;
}


/*!
  Adds \a newTab to the tab control.

  Allocates a new id, sets \a newTab's id, locates it just to the right of the
  existing tabs, inserts an accelerator if the tab's label contains the
  string "&p" for some value of p, adds it to the bar, and returns the
  newly allocated id.
*/

int ExtQTabBar::addTab( ExtQTab * newTab )
{
    return insertTab( newTab );
}


/*!
  Inserts \a newTab to the tab control.

  If \a index is not specified, the tab is simply added. Otherwise
  it's inserted at the specified position.

  Allocates a new id, sets \a newTab's id, locates it respectively,
  inserts an accelerator if the tab's label contains the string "&p"
  for some value of p, adds it to the bar, and returns the newly
  allocated id.
*/

int ExtQTabBar::insertTab( ExtQTab * newTab, int index )
{
    newTab->id = d->id++;
    l->insert( 0, newTab );
    if ( index < 0 || index > int(lstatic->count()) )
	lstatic->append( newTab );
    else
	lstatic->insert( index, newTab );

    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );

    int p = QAccel::shortcutKey( newTab->label );
    if ( p )
	d->a->insertItem( p, newTab->id );

    return newTab->id;
}


/*!
  Removes tab \a t from the tab control.
*/
void ExtQTabBar::removeTab( ExtQTab * t )
{
    //#### accelerator labels??
    l->remove( t );
    lstatic->remove( t );
    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );
    update();
}

void ExtQTabBar::moveTab( ExtQTab *mtab, int index )
{
    if ( lstatic->find(mtab) != -1 )
    {
        bool bAutoDel = lstatic->autoDelete();
        lstatic->setAutoDelete( false );
        lstatic->remove( mtab );
        lstatic->setAutoDelete( bAutoDel );

        if ( index < 0 || index >= int(lstatic->count()) )
            lstatic->append( mtab );
        else lstatic->insert( index, mtab );

    }

    if ( l->find(mtab) != -1 )
    {
        bool bAutoDel = l->autoDelete();
        l->setAutoDelete( false );
        l->remove( mtab );
        l->setAutoDelete( bAutoDel );

        if ( index < 0 || index >= int(l->count()) )
            l->append( mtab );
        else l->insert( index, mtab );
    }

    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ) );
}

/*!
  Enable tab \a id if \a enable is TRUE, or disable it if \a enable is
  FALSE.  If \a id is currently selected, setTabEnabled() makes
  another tab selected.

  setTabEnabled() updates the display respectively if this causes a
  change in \a id's status.

  \sa update(), isTabEnabled()
*/

void ExtQTabBar::setTabEnabled( int id, bool enabled )
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() ) {
	if ( t && t->id == id ) {
	    if ( t->enabled != enabled ) {
		t->enabled = enabled;
		d->a->setItemEnabled( t->id, enabled );
		QRect r( t->r );
		if ( !enabled && id == currentTab() ) {
		    QPoint p1( t->r.center() ), p2;
		    int m = 2147483647;
		    int distance;
		    // look for the closest enabled tab - measure the
		    // distance between the centers of the two tabs
		    for( ExtQTab * n = l->first(); n; n = l->next() ) {
			if ( n->enabled ) {
			    p2 = n->r.center();
			    distance = (p2.x() - p1.x())*(p2.x() - p1.x()) +
				       (p2.y() - p1.y())*(p2.y() - p1.y());
			    if ( distance < m ) {
				t = n;
				m = distance;
			    }
			}
		    }
		    if ( t->enabled ) {
			r = r.unite( t->r );
			l->append( l->take( l->findRef( t ) ) );
			emit selected( t->id );
		    }
		}
		updateMask();
		repaint( r );
	    }
	    return;
	}
    }
}


/*!
  Returns TRUE if the tab with id \a id is enabled, or FALSE if it
  is disabled or there is no such tab.

  \sa setTabEnabled()
*/

bool ExtQTabBar::isTabEnabled( int id ) const
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() ) {
	if ( t && t->id == id )
	    return t->enabled;
    }
    return FALSE;
}



/*!\reimp
*/
QSize ExtQTabBar::sizeHint() const
{
    ExtQTab * t = l->first();
    if ( t ) {
	QRect r( t->r );
	while ( (t = l->next()) != 0 )
	    r = r.unite( t->r );
	return r.size().expandedTo( QApplication::globalStrut() );
    } else {
	return QSize( 0, 0 ).expandedTo( QApplication::globalStrut() );
    }
}

/*! \reimp */

QSize ExtQTabBar::minimumSizeHint() const
{
    return QSize( d->rightB->sizeHint().width() * 2 + 75, sizeHint().height() );
}

/*!\reimp
*/
QSizePolicy ExtQTabBar::sizePolicy() const
{
    //### removeme 3.0
    return QWidget::sizePolicy();
}


/*!  Paint the single tab \a t using \a p.  If and only if \a selected
  is TRUE, \a t is currently selected.

  This virtual function may be reimplemented to change the look of
  ExtQTabBar.  If you decide to reimplement it, you may also need to
  reimplement sizeHint().
*/

void ExtQTabBar::paint( QPainter * p, ExtQTab * t, bool selected ) const
{
    style().drawTab( p, (const QTabBar *)this, (QTab *)t, selected );

    QRect r( t->r );
    p->setFont( font() );

    int iw = 0;
    int ih = 0;
    if ( t->iconset != 0 ) {
	iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 2;
	ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
    }
    int w = iw + p->fontMetrics().width( t->label ) + 4;
    int h = QMAX(p->fontMetrics().height() + 4, ih );
    paintLabel( p, QRect( r.left() + (r.width()-w)/2 - 3,
			  r.top() + (r.height()-h)/2,
			  w, h ), t, t->id == keyboardFocusTab() );
}

/*!
  Paints the label of tab \a t centered in rectangle \a br using
  painter \a p and draws a focus indication if \a has_focus is TRUE.
*/

void ExtQTabBar::paintLabel( QPainter* p, const QRect& br,
			  ExtQTab* t, bool has_focus ) const
{

    QRect r = br;
    if ( t->iconset) {
	// the tab has an iconset, draw it in the right mode
	QIconSet::Mode mode = (t->enabled && isEnabled())
	    ? QIconSet::Normal : QIconSet::Disabled;
	if ( mode == QIconSet::Normal && has_focus )
	    mode = QIconSet::Active;
	QPixmap pixmap = t->iconset->pixmap( QIconSet::Small, mode );
	int pixw = pixmap.width();
	int pixh = pixmap.height();
	r.setLeft( r.left() + pixw + 2 );
	p->drawPixmap( br.left()+2, br.center().y()-pixh/2, pixmap );
    }

    QRect tr = r;
    if ( t->id == currentTab() )
 	tr.setBottom( tr.bottom() - style().defaultFrameWidth() );
    
    if ( t->enabled && isEnabled()  ) {
#if defined(_WS_WIN32_)
	if ( colorGroup().brush( QColorGroup::Button ) == colorGroup().brush( QColorGroup::Background ) )
	    p->setPen( colorGroup().buttonText() );
	else
	    p->setPen( colorGroup().foreground() );
#else
	p->setPen( colorGroup().foreground() );
#endif
	p->drawText( tr, AlignCenter | ShowPrefix, t->label );
    } else if ( style() == MotifStyle ) {
	p->setPen( palette().disabled().foreground() );
	p->drawText( tr, AlignCenter | ShowPrefix, t->label );
    } else { // Windows style, disabled
	p->setPen( colorGroup().light() );
	QRect wr = tr;
	wr.moveBy( 1, 1 );
	p->drawText( wr, AlignCenter | ShowPrefix, t->label );
	p->setPen( palette().disabled().foreground() );
	p->drawText( tr, AlignCenter | ShowPrefix, t->label );
    }

    if ( !has_focus )
	return;

    if ( style() == WindowsStyle )
	p->drawWinFocusRect( br, backgroundColor() );
    else // shouldn't this be black, irrespective of everything?
	p->drawRect( br );
}


/*!
  Draws the mask for this tab bar.

  \internal
  This is not totally right - a few corner pixels missing.
*/

void  ExtQTabBar::updateMask()
{
    if ( !autoMask() )
	return;
    QBitmap bm( size() );
    bm.fill( color0 );

    QPainter p;
    p.begin( &bm, this );
    p.setBrush(color1);
    p.setPen(color1);

    ExtQTab * t;
    t = l->first();
    do {
	ExtQTab * n = l->next();
	if ( t )
	    style().drawTabMask( &p, (const QTabBar *)this, (QTab *)t, n == 0 );
	t = n;
    } while ( t != 0 );


    p.end();
    setMask( bm );
}

/*!
  Repaints the tab row.  All the painting is done by paint();
  paintEvent() only decides which tabs need painting and in what
  order.

  \sa paint()
*/

void ExtQTabBar::paintEvent( QPaintEvent * e )
{
    QPainter p( this );

    if ( backgroundMode() == X11ParentRelative ) {
	erase();
    } else {
	p.setBrushOrigin( rect().bottomLeft() );
	p.fillRect( 0, 0, width(), height(),
		    QBrush( colorGroup().brush( QColorGroup::Background ) ));
    }

    ExtQTab * t;
    t = l->first();
    do {
	ExtQTab * n = l->next();
	if ( t && t->r.intersects( e->rect() ) )
	    paint( &p, t, n == 0 );
	t = n;
    } while ( t != 0 );

    if ( d->scrolls && lstatic->first()->r.left() < 0 ) {
	QPointArray a;
	int h = height();
	if ( d->s == RoundedAbove ) {
	    p.fillRect( 0, 3, 4, h-5,
			QBrush( colorGroup().brush( QColorGroup::Background ) ));
	    a.setPoints( 5,  0,2,  3,h/4, 0,h/2, 3,3*h/4, 0,h );
	} else if ( d->s == RoundedBelow ) {
	    p.fillRect( 0, 2, 4, h-5,
			QBrush( colorGroup().brush( QColorGroup::Background ) ));
	    a.setPoints( 5,  0,0,  3,h/4, 0,h/2, 3,3*h/4, 0,h-3 );
	}

	if ( !a.isEmpty() ) {
	    p.setPen( colorGroup().light() );
	    p.drawPolyline( a );
	    a.translate( 1, 0 );
	    p.setPen( colorGroup().midlight() );
	    p.drawPolyline( a );
	}
    }
}


/*!
  This virtual functions is called by the mouse event handlers to
  determine which tab is pressed.  The default implementation returns
  a pointer to the tab whose bounding rectangle contains \a p, if
  exactly one tab's bounding rectangle contains \a p.  It returns 0
  else.

  \sa mousePressEvent() mouseReleaseEvent()
*/

ExtQTab * ExtQTabBar::selectTab( const QPoint & p ) const
{
    ExtQTab * selected = 0;
    bool moreThanOne = FALSE;

    QListIterator<ExtQTab> i( *l );
    while( i.current() ) {
	ExtQTab * t = i.current();
	++i;

	if ( t && t->r.contains( p ) ) {
	    if ( selected )
//	    if ( t == tab( currentTab() ) ) /* < EXT > */
		moreThanOne = TRUE;
	    else
		selected = t;
	}
    }

    return moreThanOne ? 0 : selected;
}


/*!\reimp
*/
void ExtQTabBar::mousePressEvent( QMouseEvent * e )
{
    if ( e->button() != LeftButton )
	return;
    ExtQTab * t = selectTab( e->pos() );
    if ( t != 0 && t == selectTab( e->pos() ) && t->enabled ) {
	setCurrentTab( t );
	/*! selectreleased = false;*/
    }
}

void ExtQTabBar::mouseMoveEvent ( QMouseEvent *e )
{
    // mouse move : drag this tab
    if ( ( e->state() & Qt::LeftButton ) && selectreleased )
    {
        QDragObject *d = new QTextDrag( "QTWTM", this );
        d->dragCopy();
    }
}

/*!\reimp
*/

void ExtQTabBar::mouseReleaseEvent( QMouseEvent * )
{
    selectreleased = true;
}

void ExtQTabBar::dragMoveEvent( QDragMoveEvent *e )
{
    QString text; 
    if ( QTextDrag::decode( e, text ) && ( text == "QTWTM" ) &&
         e->source() == this )
    {
        for( int i = 0; i < (int)lstatic->count(); i++ )
	{
	    if ( ( e->pos().x() >= lstatic->at(i)->r.x() ) && 
	         ( e->pos().x() < lstatic->at(i)->r.x() + lstatic->at(i)->r.width() ) )
	    {
			if ( currentTab() != lstatic->at(i)->id ) e->accept();
			else e->ignore();
			break;
	    }
	}
    }
}

void ExtQTabBar::dropEvent( QDropEvent *e )
{
    QString text; 
    if ( QTextDrag::decode( e, text ) && ( text == "QTWTM" ) &&
         e->source() == this )
    {
	for( int i = 0; i < (int)lstatic->count(); i++ )
	{
	    if ( lstatic->at(i)->r.contains( e->pos() ) )
	    {
		if ( currentTab() != lstatic->at(i)->id )
		{
		    moveTab( tab( currentTab() ), i );
		    repaint();
		    e->acceptAction();
		    emit tabOrderChanged();
		} else e->ignore();
		break;
	    }
	}
    }
}

QPopupMenu *ExtQTabBar::pTabMenu = NULL;

/*void ExtQTabBar::contextMenuEvent( QContextMenuEvent *pEvent )
{
	if (!pTabMenu) return;
	pTabMenu->exec(QCursor::pos());
	pEvent->accept();
	return;
} */

/*!  \reimp
*/
void ExtQTabBar::show()
{
    //  ensures that one tab is selected.
    ExtQTab * t = l->last();
    QWidget::show();

    if ( t )
	emit selected( t->id );
}

/*!  If a page is currently visible, returns its ID.  If no page is
  currently visible, returns either -1 or the ID of one of the pages.

  Even if the return value is not -1, you cannot assume either that
  the user can see the relevant page, or that the tab \link
  isTabEnabled() is enabled.\endlink

  When you need to display something, the return value from this
  function represents the best page to display.  That's all.

  \sa selected()
*/

int ExtQTabBar::currentTab() const
{
    const ExtQTab * t = l->getLast();

    return t ? t->id : -1;
}


/*! Raises the tab with ID \a id and emits the selected() signal.

  \sa currentTab() selected() tab()
*/

void ExtQTabBar::setCurrentTab( int id )
{
    setCurrentTab( tab( id ) );
}


/*! Raises \a tab and emits the selected() signal unless the tab was
  already current.

  \sa currentTab() selected()
*/

void ExtQTabBar::setCurrentTab( ExtQTab * tab )
{
    if ( tab && l ) {
	if ( l->last() == tab )
	    return;

	QRect r = l->last()->r;
	if ( l->findRef( tab ) >= 0 )
	    l->append( l->take() );

	d->focus = tab->id;
	updateMask();
	if ( tab->r.intersects( r ) ) {
	    repaint( r.unite( tab->r ) );
	} else {
	    repaint( r );
	    repaint( tab->r );
	}
	makeVisible( tab );
	emit selected( tab->id );
    }
}

/*!  If this tab control has keyboard focus, returns the ID of the
  tab Space will select.  Otherwise, returns -1.
*/

int ExtQTabBar::keyboardFocusTab() const
{
    return hasFocus() ? d->focus : -1;
}


/*!\reimp
*/
void ExtQTabBar::keyPressEvent( QKeyEvent * e )
{
    //   The right and left arrow keys move a selector, the spacebar
    //   makes the tab with the selector active.  All other keys are
    //   ignored.

    int old = d->focus;

    if ( e->key() == Key_Left ) {
	// left - skip past any disabled ones
	if ( d->focus > 0 ) {
	    ExtQTab * t = lstatic->last();
	    while ( t && t->id != d->focus )
		t = lstatic->prev();
	    do {
		t = lstatic->prev();
	    } while ( t && !t->enabled);
	    if (t)
		d->focus = t->id;
	}
	if ( d->focus < 0 )
	    d->focus = old;
    } else if ( e->key() == Key_Right ) {
	ExtQTab * t = lstatic->first();
	while ( t && t->id != d->focus )
	    t = lstatic->next();
	do {
	    t = lstatic->next();
	} while ( t && !t->enabled);
    if (t)
	d->focus = t->id;
    if ( d->focus >= d->id )
	d->focus = old;
    } else {
	// other keys - ignore
	e->ignore();
	return;
    }

    // if the focus moved, repaint and signal
    if ( old != d->focus ) {
	setCurrentTab( d->focus );
    }
}


/*!  Returns a pointer to the tab with ID \a id, or 0 if there is no
  such tab.

  \sa count()
*/

ExtQTab * ExtQTabBar::tab( int id )
{
    ExtQTab * t;
    for( t = l->first(); t; t = l->next() )
	if ( t && t->id == id )
	    return t;
    return 0;
}


ExtQTab * ExtQTabBar::tabAt( int index )
{
    ExtQTab * t;
    t = lstatic->at( index );
    return t;
}

/*!
  Returns the position index of the tab with id \a id.

  \sa tabAt()
 */
int ExtQTabBar::indexOf( int id ) const
{
    ExtQTab * t;
    int idx = 0;
    for( t = lstatic->first(); t; t = lstatic->next() )
    {
        if ( t && t->id == id )
            return idx;
        idx++;
    }
    return -1;
}


/*! Returns the number of tabs in the tab bar.

  \sa tab()
*/
int ExtQTabBar::count() const
{
    return l->count();
}


/*!
  The list of ExtQTab objects added.
*/
QList<ExtQTab> * ExtQTabBar::tabList()
{
    return l;
}


/*!  Returns the shape of this tab bar. \sa setShape() */

ExtQTabBar::Shape ExtQTabBar::shape() const
{
    return d ? d->s : RoundedAbove;
}


/*!  Sets the shape of this tab bar to \a s and refreshes the bar.
*/

void ExtQTabBar::setShape( Shape s )
{
    if ( !d || d->s == s )
	return;
    //######### must recalculate heights
    d->s = s;
    updateMask();
    update();
}



/*!
  Layout all existing tabs (i.e. setting their \c r attribute) according
  to their label and their iconset.
 */
void ExtQTabBar::layoutTabs()
{
    if ( lstatic->isEmpty() )
	return;

    int hframe, vframe, overlap;
    style().tabbarMetrics( (const QTabBar *)this, hframe, vframe, overlap );
    QFontMetrics fm = fontMetrics();
    int x = 0;
    QRect r;
    ExtQTab *t;
    for ( t = lstatic->first(); t; t = lstatic->next() ) {
	int lw = fm.width( t->label );
	int iw = 0;
	int ih = 0;
	if ( t->iconset != 0 ) {
	    iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 2;
	    ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	}
	int h = QMAX( fm.height(), ih );
	h = QMAX( h, QApplication::globalStrut().height() );

	h += vframe;
	t->r.setRect( x, 0, QMAX( lw + hframe + iw,
		    QApplication::globalStrut().width() ), h );
	x += t->r.width() - overlap;
	r = r.unite( t->r );
    }
    for ( t = lstatic->first(); t; t = lstatic->next() )
	t->r.setHeight( r.height() );
}

/*!
  \reimp
*/

void ExtQTabBar::styleChange( QStyle& old )
{
	layoutTabs();
	QWidget::styleChange( old );
}

/*!
  \reimp
*/
void ExtQTabBar::focusInEvent( QFocusEvent * )
{
    ExtQTab *t = l->first();
    for ( ; t; t = l->next() ) {
	if ( t->id == d->focus ) {
	    QPainter p;
	    p.begin( this );
	    QRect r = t->r;
	    p.setFont( font() );

	    int iw = 0;
	    int ih = 0;
	    if ( t->iconset != 0 ) {
		iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 2;
		ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	    }
	    int w = iw + p.fontMetrics().width( t->label ) + 4;
	    int h = QMAX(p.fontMetrics().height() + 4, ih );
	    paintLabel( &p, QRect( r.left() + ( r.width() -w ) /2 - 3,
				   r.top() + ( r.height()-h ) / 2,
				   w, h ), t, TRUE );
	    p.end();
	}
    }
}

/*!
  \reimp
*/
void ExtQTabBar::focusOutEvent( QFocusEvent * )
{
    ExtQTab *t = l->first();
    for ( ; t; t = l->next() ) {
	if ( t->id == d->focus ) {
	    QPainter p;
	    p.begin( this );
	    p.setBrushOrigin( rect().bottomLeft() );
	    QRect r = t->r;
	    p.setFont( font() );

	    int iw = 0;
	    int ih = 0;
	    if ( t->iconset != 0 ) {
		iw = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).width() + 2;
		ih = t->iconset->pixmap( QIconSet::Small, QIconSet::Normal ).height();
	    }
	    int w = iw + p.fontMetrics().width( t->label ) + 4;
	    int h = QMAX(p.fontMetrics().height() + 4, ih );
	    p.fillRect( QRect( r.left() + ( r.width() -w ) / 2 - 4,
				   r.top() + ( r.height()-h ) / 2 - 1,
			       w + 2, h + 2 ), colorGroup().brush(QColorGroup::Background ) );
	    style().drawTab( &p, (const QTabBar *)this, (QTab *)t, TRUE );
	    paintLabel( &p, QRect( r.left() + ( r.width() -w ) /2 - 3,
				   r.top() + ( r.height()-h ) / 2,
				   w, h ), t, FALSE );
	    p.end();
	}
    }
}

/*!
  \reimp
*/
void ExtQTabBar::resizeEvent( QResizeEvent * )
{
    const int arrowWidth = 16;
    d->rightB->setGeometry( width() - arrowWidth, 0, arrowWidth, height() );
    d->leftB->setGeometry( width() - 2*arrowWidth, 0, arrowWidth, height() );
    layoutTabs();
    updateArrowButtons();
    makeVisible( tab( currentTab() ));
}

void ExtQTabBar::scrollTabs()
{
    ExtQTab* left = 0;
    ExtQTab* right = 0;
    for ( ExtQTab* t = lstatic->first(); t; t = lstatic->next() ) {
	if ( t->r.left() < 0 && t->r.right() > 0 )
	    left = t;
	if ( t->r.left() < d->leftB->x() )
	    right = t;
    }

    if ( sender() == d->leftB )
	makeVisible( left );
    else  if ( sender() == d->rightB )
	makeVisible( right );
}

void ExtQTabBar::makeVisible( ExtQTab* tab  )
{
    bool tooFarLeft = ( tab && tab->r.left() < 0 );
    bool tooFarRight = ( tab && tab->r.right() >= d->leftB->x() );

    if ( !d->scrolls || ( !tooFarLeft && ! tooFarRight ) )
	return;

    layoutTabs();

    int offset = 0;

    if ( tooFarLeft )
	offset = tab == lstatic->first() ? 0 : tab->r.left() - 8;
    else if ( tooFarRight ) {
	offset = tab->r.right() - d->leftB->x() + 1;
    }

    for ( ExtQTab* t = lstatic->first(); t; t = lstatic->next() )
	t->r.moveBy( -offset, 0 );

    d->leftB->setEnabled( offset != 0 );
    d->rightB->setEnabled( lstatic->last()->r.right() >= d->leftB->x() );


    update();
}

void ExtQTabBar::updateArrowButtons()
{
    bool b = lstatic->last() &&	( lstatic->last()->r.right() > width() );
    d->scrolls = b;
    if ( d->scrolls ) {
	d->leftB->setEnabled( FALSE );
	d->rightB->setEnabled( TRUE );
	d->leftB->show();
	d->rightB->show();
    } else {
	d->leftB->hide();
	d->rightB->hide();
    }
}
#endif

#include "QTExt.h"
#ifndef QT_NO_TABWIDGET
#include "qobjectlist.h"
#include "qobjectdict.h"
#include "qapplication.h"
#include "qwidgetstack.h"
#include "qbitmap.h"

// NOT REVISED
/*!
  \class ExtQTabWidget qtabwidget.h

  \brief The ExtQTabWidget class provides a stack of tabbed widgets.

  \ingroup organizers

  A tabbed widget is one in which several "pages" are available, and
  the user selects which page to see and use by clicking on its tab,
  or by pressing the indicated Alt-(letter) key combination.

  ExtQTabWidget does not provide more than one row of tabs, and does not
  provide tabs along the sides or bottom of the pages.

  The normal way to use ExtQTabWidget is to do the following in the
  constructor: <ol> <li> Create a ExtQTabWidget. <li> Create a QWidget
  for each of the pages in the tab dialog, insert children into it,
  set up geometry management for it, and use addTab() to set up a tab
  and keyboard accelerator for it. <li> Connect to the
  signals and slots. </ol>

  If you don't call addTab(), the page you have created will not be
  visible.  Please don't confuse the object name you supply to the
  QWidget constructor and the tab label you supply to addTab():
  addTab() takes a name which indicates an accelerator and is
  meaningful and descriptive to the user, while the widget name is
  used primarily for debugging.

  A signal currentChanged() is emitted when the user selects some
  page.

  Each tab is either enabled or disabled at any given time.  If a tab
  is enabled, the tab text is drawn in black and the user can select
  that tab.  If it is disabled, the tab is drawn in a different way
  and the user can not select that tab.  Note that even though a tab
  is disabled, the page can still be visible, for example if all of
  the tabs happen to be disabled.

  While tab widgets can be a very good way to split up a complex
  dialog, it's also very easy to make a royal mess out of it. See
  ExtQTabDialog for some design hints.

  Most of the functionality in ExtQTabWidget is provided by a ExtQTabBar (at
  the top, providing the tabs) and a QWidgetStack (most of the area,
  organizing the individual pages).

  <img src=qtabwidget-m.png> <img src=qtabwidget-w.png>

  \sa ExtQTabDialog
*/


/*! \enum ExtQTabWidget::TabPosition

  This enum type defines where ExtQTabWidget can draw the tab row: <ul>
  <li> \c Top - above the pages
  <li> \c Bottom - below the pages
  </ul>
*/

/*! \enum ExtQTabWidget::TabShape

  This enum type defines the shape of the tabs: <ul>
  <li> \c Rounded - rounded look (normal)
  <li> \c Triangular - triangular look (very unusual, included for completeness)
  </ul>
*/

/* undocumented now
  \obsolete

  \fn void ExtQTabWidget::selected( const QString &tabLabel );

  This signal is emitted whenever a tab is selected (raised),
  including during the first show().

  \sa raise()
*/


/*! \fn void ExtQTabWidget::currentChanged( QWidget* );

  This signal is emitted whenever the current page changes.

  \sa currentPage(), showPage(), tabLabel()
*/

class ExtQTabWidgetData
{
public:
    ExtQTabWidgetData()
	: tabs(0), stack(0), dirty( TRUE ), pos( ExtQTabWidget::Top ), shape( ExtQTabWidget::Rounded )
	{};
    ~ExtQTabWidgetData(){};
    ExtQTabBar* tabs;
    QWidgetStack* stack;
    bool dirty;
    ExtQTabWidget::TabPosition pos;
    ExtQTabWidget::TabShape shape;
};





/*!
  Constructs a tabbed widget with parent \a parent, name \a name
  and widget flags \a f.
*/


ExtQTabWidget::ExtQTabWidget( QWidget *parent, const char *name, WFlags f )
    : QWidget( parent, name, f )
{
    init();
}

/*!
  \overload
*/
ExtQTabWidget::ExtQTabWidget( QWidget *parent, const char *name )
    : QWidget( parent, name )
{
    init();
}

void ExtQTabWidget::init()
{
    d = new ExtQTabWidgetData;

    d->stack = new QWidgetStack( this, "tab pages" );
    d->stack->installEventFilter( this );
    setTabBar( new ExtQTabBar( this, "tab control" ) );

    d->stack->setFrameStyle( QFrame::StyledPanel | QFrame::Raised );
    d->stack->setLineWidth( style().defaultFrameWidth() );

    setSizePolicy( QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding ) );
    setFocusPolicy( TabFocus );
    setFocusProxy( d->tabs );
}

/*!
  Destructs the tab widget.
*/
ExtQTabWidget::~ExtQTabWidget()
{
    delete d;
}

/*!
  Adds another tab and page to the tab view.

  The tab will be labelled \a label and \a child constitutes the new
  page.  Note the difference between the widget name (which you supply
  to widget constructors and to e.g. setTabEnabled()) and the tab
  label: The name is internal to the program and invariant, while the
  label is shown on screen and may vary according to e.g. language.

  \a label is written in the QButton style, where &P makes Qt create
  an accelerator key on Alt-P for this page.  For example:

  \code
    td->addTab( graphicsPane, "&Graphics" );
    td->addTab( soundPane, "&Sound" );
  \endcode

  If the user presses Alt-S the sound page of the tab dialog is shown,
  if the user presses Alt-P the graphics page is shown.

  If you call addTab() after show(), the screen will flicker and the
  user will be confused.
*/
void ExtQTabWidget::addTab( QWidget *child, const QString &label)
{
    ExtQTab * t = new ExtQTab();
    CHECK_PTR( t );
    t->label = label;
    addTab( child, t );
}


/*!
  Adds another tab and page to the tab view.

  This function is the same as addTab() but with an additional
  iconset.
 */
void ExtQTabWidget::addTab( QWidget *child, const QIconSet& iconset, const QString &label)
{
    ExtQTab * t = new ExtQTab();
    CHECK_PTR( t );
    t->label = label;
    t->iconset = new QIconSet( iconset );
    addTab( child, t );
}

/*!
  This is a lower-level method for adding tabs, similar to the other
  addTab() method.  It is useful if you are using setTabBar() to set a
  ExtQTabBar subclass with an overridden ExtQTabBar::paint() routine for a
  subclass of ExtQTab.
*/
void ExtQTabWidget::addTab( QWidget *child, ExtQTab* tab)
{
    tab->enabled = TRUE;
    int id = d->tabs->addTab( tab );
    d->stack->addWidget( child, id );
    if ( d->stack->frameStyle() != ( QFrame::StyledPanel | QFrame::Raised ) )
	d->stack->setFrameStyle( QFrame::StyledPanel | QFrame::Raised );
    setUpLayout();
}



/*!
  Inserts another tab and page to the tab view.

  The tab will be labelled \a label and \a child constitutes the new
  page.  Note the difference between the widget name (which you supply
  to widget constructors and to e.g. setTabEnabled()) and the tab
  label: The name is internal to the program and invariant, while the
  label is shown on screen and may vary according to e.g. language.

  \a label is written in the QButton style, where &P makes Qt create
  an accelerator key on Alt-P for this page.  For example:

  \code
    td->insertTab( graphicsPane, "&Graphics" );
    td->insertTab( soundPane, "&Sound" );
  \endcode

  If \a index is not specified, the tab is simply added. Otherwise
  it's inserted at the specified position.

  If the user presses Alt-S the sound page of the tab dialog is shown,
  if the user presses Alt-P the graphics page is shown.

  If you call insertTab() after show(), the screen will flicker and the
  user will be confused.
*/
void ExtQTabWidget::insertTab( QWidget *child, const QString &label, int index)
{
    ExtQTab * t = new ExtQTab();
    CHECK_PTR( t );
    t->label = label;
    insertTab( child, t, index );
}


/*!
  Inserts another tab and page to the tab view.

  This function is the same as insertTab() but with an additional
  iconset.
 */
void ExtQTabWidget::insertTab( QWidget *child, const QIconSet& iconset, const QString &label, int index )
{
    ExtQTab * t = new ExtQTab();
    CHECK_PTR( t );
    t->label = label;
    t->iconset = new QIconSet( iconset );
    insertTab( child, t, index );
}

/*!
  This is a lower-level method for inserting tabs, similar to the other
  insertTab() method.  It is useful if you are using setTabBar() to set a
  ExtQTabBar subclass with an overridden ExtQTabBar::paint() routine for a
  subclass of ExtQTab.
*/
void ExtQTabWidget::insertTab( QWidget *child, ExtQTab* tab, int index)
{
    tab->enabled = TRUE;
    int id = d->tabs->insertTab( tab, index );
    d->stack->addWidget( child, id );
    if ( d->stack->frameStyle() != ( QFrame::StyledPanel | QFrame::Raised ) )
	d->stack->setFrameStyle( QFrame::StyledPanel | QFrame::Raised );
    setUpLayout();
}


/*!
  Defines a new label for the tab of page \a w
 */
void ExtQTabWidget::changeTab( QWidget *w, const QString &label)
{

    //#### accelerators
    int id = d->stack->id( w );
    if ( id < 0 )
	return;
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
	return;
    t->label = label;
    d->tabs->layoutTabs();

    int ct = d->tabs->currentTab();
    d->tabs->blockSignals( TRUE );
    d->tabs->setCurrentTab( 0 );
    d->tabs->setCurrentTab( ct );
    d->tabs->blockSignals( FALSE );

    d->tabs->update();
    setUpLayout();
}

/*!
  Defines a new \a iconset and a new \a label for the tab of page \a w
 */
void ExtQTabWidget::changeTab( QWidget *w, const QIconSet& iconset, const QString &label)
{
    //#### accelerators
    int id = d->stack->id( w );
    if ( id < 0 )
	return;
    ExtQTab* t = d->tabs->tab( id );
    if ( !t )
	return;
    if ( t->iconset )
	delete t->iconset;
    t->label = label;
    t->iconset = new QIconSet( iconset );
    d->tabs->layoutTabs();

    int ct = d->tabs->currentTab();
    d->tabs->blockSignals( TRUE );
    d->tabs->setCurrentTab( 0 );
    d->tabs->setCurrentTab( ct );
    d->tabs->blockSignals( FALSE );

    d->tabs->update();
    setUpLayout();
}

void ExtQTabWidget::moveTab( QWidget *w, int index )
{
    d->tabs->moveTab( d->tabs->tab( d->stack->id( w ) ), index );
}

/*!
  Returns TRUE if the page \a w is enabled, and
  false if it is disabled.

  \sa setTabEnabled(), QWidget::isEnabled()
*/

bool ExtQTabWidget::isTabEnabled( QWidget* w ) const
{
    int id = d->stack->id( w );
    if ( id >= 0 )
	return w->isEnabled();
    else
	return FALSE;
}

/*!
  Enables/disables page \a w according to the value of \a enable, and
  redraws the page's tab appropriately.

  ExtQTabWidget uses QWidget::setEnabled() internally, rather than keep a
  separate flag.

  Note that even a disabled tab/page may be visible.  If the page is
  visible already, ExtQTabWidget will not hide it, and if all the pages
  are disabled, ExtQTabWidget will show one of them.

  \sa isTabEnabled(), QWidget::setEnabled()
*/

void ExtQTabWidget::setTabEnabled( QWidget* w, bool enable)
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
	w->setEnabled( enable );
	d->tabs->setTabEnabled( id, enable );
    }
}

/*!  Ensures that \a w is shown.  This is useful mainly for accelerators.

  \warning Used carelessly, this function can easily surprise or
  confuse the user.

  \sa ExtQTabBar::setCurrentTab()
*/
void ExtQTabWidget::showPage( QWidget * w)
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
	d->stack->raiseWidget( w );
	d->tabs->setCurrentTab( id );
	// ### why overwrite the frame style?
	if ( d->stack->frameStyle() != ( QFrame::StyledPanel|QFrame::Raised ) )
	    d->stack->setFrameStyle( QFrame::StyledPanel | QFrame::Raised );
    }
}

/*! Removes page \a w from this stack of widgets.  Does not
  delete \a w.
  \sa showPage(), QWidgetStack::removeWidget()
*/
void ExtQTabWidget::removePage( QWidget * w )
{
    int id = d->stack->id( w );
    if ( id >= 0 ) {
	d->tabs->setTabEnabled( id, FALSE );
	d->stack->removeWidget( w );
	d->tabs->removeTab( d->tabs->tab(id) );
	setUpLayout();
	if ( !currentPage() && d->stack->frameStyle() != QFrame::NoFrame )
	    d->stack->setFrameStyle( QFrame::NoFrame );
    }
}

/*!  Returns the text in the tab for page \a w.
*/

QString ExtQTabWidget::tabLabel( QWidget * w)
{
    ExtQTab * t = d->tabs->tab( d->stack->id( w ) );
    return t ? t->label : QString::null;
}

/*!  Sets the tab label for page \a w to \a l
 */

void ExtQTabWidget::setTabLabel( QWidget * w, const QString &l )
{
    ExtQTab * t = d->tabs->tab( d->stack->id( w ) );
    if ( t )
        t->label = l;
    d->tabs->layoutTabs();
    d->tabs->update();
    setUpLayout();
}


/*!  Returns a pointer to the page currently being displayed by the
tab dialog.  The tab dialog does its best to make sure that this value
is never 0, but if you try hard enough it can be.
*/

QWidget * ExtQTabWidget::currentPage() const
{
    return d->stack->visibleWidget();
}

/*! Returns the ID of the current page.
*/

int ExtQTabWidget::currentPageIndex() const
{
    /* BUG old - buggy? return d->tabs->currentTab(); */
    return d->tabs->indexOf( d->tabs->currentTab() );
}

/*! Sets the page with index \a id as current page.

    Note that \e id is not the index that is specified when you insert a tab
*/

void ExtQTabWidget::setCurrentPage( int id )
{
    d->tabs->setCurrentTab( id );
    showTab( id );
}

void ExtQTabWidget::setCurrentPageIndex( int index )
{
    d->tabs->setCurrentTab( d->tabs->tabAt(index) );
    showTab( d->tabs->currentTab() );
}

/*!
  \reimp
 */
void ExtQTabWidget::resizeEvent( QResizeEvent * )
{
    setUpLayout();
}

/*!
  Replaces the ExtQTabBar heading the dialog by the given tab bar.
  Note that this must be called \e before any tabs have been added,
  or the behavior is undefined.
  \sa tabBar()
*/
void ExtQTabWidget::setTabBar( ExtQTabBar* tb)
{
    if ( tb->parentWidget() != this )
	tb->reparent( this, QPoint(0,0), TRUE );
    delete d->tabs;
    d->tabs = tb;
    connect( d->tabs, SIGNAL(selected(int)),
	     this,    SLOT(showTab(int)) );
    setUpLayout();
}


/*!
  Returns the currently set ExtQTabBar.
  \sa setTabBar()
*/
ExtQTabBar* ExtQTabWidget::tabBar() const
{
    return d->tabs;
}

/*!
  Ensures that the selected tab's page is visible and appropriately sized.
*/

void ExtQTabWidget::showTab( int i )
{
    if ( d->stack->widget( i ) ) {
	d->stack->raiseWidget( i );
	emit selected( d->tabs->tab( i )->label );
	emit currentChanged( d->stack->widget( i ) );
	d->stack->widget( i )->setFocus();
    }
}

/*!
  Set up the layout.
 */
void ExtQTabWidget::setUpLayout( bool onlyCheck )
{
    if ( onlyCheck && !d->dirty )
	return; // nothing to do

    if ( !isVisible() ) {
	d->dirty = TRUE;
	return; // we'll do it later
    }
    QSize t( d->tabs->sizeHint() );
    if ( t.width() > width() )
	t.setWidth( width() );
    int lw = d->stack->lineWidth();
    if ( d->pos == Bottom ) {
	d->tabs->setGeometry( QMAX(0, lw-2), height() - t.height() - lw, t.width(), t.height() );
	d->stack->setGeometry( 0, 0, width(), height()-t.height()+QMAX(0, lw-2) );
    }
    else { // Top
	d->tabs->setGeometry( QMAX(0, lw-2), 0, t.width(), t.height() );
	d->stack->setGeometry( 0, t.height()-lw, width(), height()-t.height()+QMAX(0, lw-2));
    }

    d->dirty = FALSE;
    if ( !onlyCheck )
	update();
    if ( autoMask() )
	updateMask();
}

/*!\reimp
*/
QSize ExtQTabWidget::sizeHint() const
{
    QSize s( d->stack->sizeHint() );
    QSize t( d->tabs->sizeHint() );
    return QSize( QMAX( s.width(), t.width()),
		  s.height() + t.height() );
}


/*!
  Returns a suitable minimum size for the tab widget.
*/
QSize ExtQTabWidget::minimumSizeHint() const
{
    QSize s( d->stack->minimumSizeHint() );
    QSize t( d->tabs->minimumSizeHint() );
    return QSize( QMAX( s.width(), t.width()),
		  s.height() + t.height() );
}

/*! \reimp
 */
void ExtQTabWidget::showEvent( QShowEvent * )
{
    setUpLayout( TRUE );
}


/*!
  Returns the position of the tabs.

  Possible values are ExtQTabWidget::Top and ExtQTabWidget::Bottom.
  \sa setTabPosition()
 */
ExtQTabWidget::TabPosition ExtQTabWidget::tabPosition() const
{
    return d->pos;
}

/*!
  Sets the position of the tabs to \e pos

  Possible values are ExtQTabWidget::Top and ExtQTabWidget::Bottom.
  \sa tabPosition()
 */
void ExtQTabWidget::setTabPosition( TabPosition pos)
{
    if (d->pos == pos)
	return;
    d->pos = pos;
    if (d->tabs->shape() == ExtQTabBar::TriangularAbove || d->tabs->shape() == ExtQTabBar::TriangularBelow ) {
	if ( pos == Bottom )
	    d->tabs->setShape( ExtQTabBar::TriangularBelow );
	else
	    d->tabs->setShape( ExtQTabBar::TriangularAbove );
    }
    else {
	if ( pos == Bottom )
	    d->tabs->setShape( ExtQTabBar::RoundedBelow );
	else
	    d->tabs->setShape( ExtQTabBar::RoundedAbove );
    }
    d->tabs->layoutTabs();
    setUpLayout();
}

/*!
  Returns the shape of the tabs.
  
  \sa setTabShape()
*/

ExtQTabWidget::TabShape ExtQTabWidget::tabShape() const
{
    return d->shape;
}

/*!
  Sets the shape of the tabs to \a s.
*/

void ExtQTabWidget::setTabShape( TabShape s )
{
    if ( d->shape == s )
	return;
    d->shape = s;
    if ( d->pos == Top ) {
	if ( s == Rounded )
	    d->tabs->setShape( ExtQTabBar::RoundedAbove );
	else
	    d->tabs->setShape( ExtQTabBar::TriangularAbove );
    } else {
	if ( s == Rounded )
	    d->tabs->setShape( ExtQTabBar::RoundedBelow );
	else
	    d->tabs->setShape( ExtQTabBar::TriangularBelow );
    }
    d->tabs->layoutTabs();
    setUpLayout();
}


/*!
  Returns the width of the margin. The margin is the distance between
  the innermost pixel of the frame and the outermost pixel of the
  pages.

  \sa setMargin()
*/
int ExtQTabWidget::margin() const
{
    return d->stack->margin();
}

/*!
  Sets the width of the margin to \e w.
  \sa margin()
*/
 void ExtQTabWidget::setMargin( int w )
{
    d->stack->setMargin( w );
    setUpLayout();
}


/*! \reimp
 */
void ExtQTabWidget::styleChange( QStyle& old )
{
    d->stack->setLineWidth( style().defaultFrameWidth() );
    setUpLayout();
    QWidget::styleChange( old );
}


/*! \reimp
 */
void ExtQTabWidget::updateMask()
{
    if ( !autoMask() )
	return;

    QRect r;
    QRegion reg( r );
    reg += QRegion( d->tabs->geometry() );
    reg += QRegion( d->stack->geometry() );
    setMask( reg );
}


/*!\reimp
 */
bool ExtQTabWidget::eventFilter( QObject *o, QEvent * e)
{
    if ( o == d->stack && e->type() == QEvent::ChildRemoved
	 && ( (QChildEvent*)e )->child()->isWidgetType() ) {
	removePage( (QWidget*)  ( (QChildEvent*)e )->child() );
	return TRUE;
    }
    return FALSE;
}


/*!\reimp
 */

QSizePolicy ExtQTabWidget::sizePolicy() const
{
    //### removeme 3.0
    return QWidget::sizePolicy();
}

#endif

#endif
