/**
 * @file UndoRedo.cpp
 *
 * @brief Undo/Redo function, implementation
 * 
 * @author Ferenc Moricz
 * $Revision: 1.11 $
 * $Date: 2004/02/22 00:12:50 $
 *
 */

#include "UndoRedo.h"
#include "EditBuffer.h"
#include "Debug.h"

// AtomicTextOperation::AtomicTextOperation(.....) CONSTRUCTOR
/**
 *  @brief Creates a new atomic operationc
 */
AtomicTextOperation::AtomicTextOperation(OperationCode eCode, QString sText, int iColumn, int iLine, int iAttrib1, int iAttrib2)
{
	this->eOperationCode = eCode;
	this->sText = sText;
	this->iColumn = iColumn;
	this->iLine = iLine;
	this->iAttrib1 = iAttrib1;
	this->iAttrib2 = iAttrib2;
}

// AtomicTextOperation::execDo(.)
/**
 *  @brief Executes an atomic operation on a buffer
 */
void AtomicTextOperation::execDo(EditBuffer *pBuffer)
{
	bool bReturn = false;

	if (pBuffer == NULL)
	{
		dwarn("undoredo", "Null buffer in AtomicTextOperation::execDo!");
		bReturn = true;
	}

	if (!bReturn)
	switch (eOperationCode)
	{
		case INSERTSTRING:
			pBuffer->atomic_insertString(&sText, iColumn, iLine);
			break;
		case DELETESTRING:
			pBuffer->atomic_deleteString(iColumn, iLine, sText.length());
			break;
		case CREATELINE:
			pBuffer->atomic_createLine(iLine);
			break;
		case DELETELINE:
			pBuffer->atomic_deleteLine(iLine);
			break;
		case MOVECURSOR:
			pBuffer->atomic_moveCursor(iColumn, iLine);
			break;
		case SETSELECTION:
			pBuffer->atomic_setSelection(sText, iColumn, iLine, iAttrib1, iAttrib2);
			pBuffer->text->selectcolor = pBuffer->text->properties->getColorValue (
				sText, (sText == "buffer.selection.color" ? &Qt::blue :
				         (sText == "buffer.selection.pastedcolor" ? &Qt::magenta : &Qt::green)));
			break;
		default:
			break;
	}
}

// AtomicTextOperation::execUndo(.)
/**
 *  @brief Undo an atomic operation on a buffer
 */
void AtomicTextOperation::execUndo(EditBuffer *pBuffer)
{
	bool bReturn = false;

	if (pBuffer == NULL)
	{
		dwarn("undoredo", "Null buffer in AtomicTextOperation::execDo!");
		bReturn = true;
	}

	if (!bReturn)
	switch (eOperationCode)
	{
		case INSERTSTRING:
			pBuffer->atomic_deleteString(iColumn, iLine, sText.length());
			break;
		case DELETESTRING:
			pBuffer->atomic_insertString(&sText, iColumn, iLine);
			break;
		case CREATELINE:
			pBuffer->atomic_deleteLine(iLine);
			break;
		case DELETELINE:
			pBuffer->atomic_createLine(iLine);
			pBuffer->atomic_insertString(&sText, 0, iLine);
			break;
		case MOVECURSOR:
			pBuffer->text->cursor.setX(iAttrib1);
			pBuffer->text->cursor.setY(iAttrib2);
			break;
		case SETSELECTION:
			pBuffer->atomic_setSelection(sText, iColumn, iLine, iAttrib1, iAttrib2);
			pBuffer->text->selectcolor = pBuffer->text->properties->getColorValue (
				sText, (sText == "buffer.selection.color" ? &Qt::blue :
				         (sText == "buffer.selection.pastedcolor" ? &Qt::magenta : &Qt::green)));
			break;
		default:
			break;
	}
}

// TextOperation::addAtomicOperation(.)
/**
 *  @brief Creates and stores a new atomic operation
 */
void TextOperation::addAtomicOperation(AtomicTextOperation *pOperation)
{
	ovAtomicOps.push_back(pOperation);
}

// UndoRedo() constructor
/**
 *  @brief Creates a new UndoRedo object
 */
UndoRedo::UndoRedo()
{
	bOperationOpened = false;
	pSavedPosition = NULL;
}

// UndoRedo::createOperation()
/**
 *  @brief Creates and opens a new text operation
 */
void UndoRedo::createOperation(const QString &sName)
{
	dmesg("6:undoredo", "UNDOREDO: create operation %s", sName.latin1());
	TextOperation *pOperation = new TextOperation();
	pOperation->sOperationName = new QString(sName);
	ovPrevOperationLst.push_back( pOperation);
	bOperationOpened = true;
}

void UndoRedo::continueOperation(void)
{
	bOperationOpened = true;
}

// UndoRedo::closeOperation()
/**
 *  @brief Closes an opened text operation
 */
void UndoRedo::closeOperation(void)
{
	bool bChanged = false, bReturn = false;
	TextOperation *pTextOp = NULL;

	/* 1. close the operation list */
	bOperationOpened = false;

	/* 2. get the actual operation */
	if (ovPrevOperationLst.size() > 0)
	{
		/* 2.1 */
		pTextOp = ovPrevOperationLst[ovPrevOperationLst.size()-1];
		if (pTextOp == NULL)
		{
			derr("undoredo", "Actual text operation is not found (NULL pointer)!");
			bReturn = true;
		}
	}
	else
	{
		derr("undoredo", "Trying to close an operation where there are no operation in the list!");
		bReturn = true;
	}

	/* 3. ITER check all of the atomic operations in this text operation.
	           if any of them changes the buffer the buffer then this operation changes the buffer */
	if (!bReturn)
		for(int iIdx = 0; iIdx < (int)pTextOp->ovAtomicOps.size(); iIdx ++)
		{
			switch((*(pTextOp->ovAtomicOps.begin() + iIdx))->getOperationCode())
			{
				case AtomicTextOperation::INSERTSTRING:
				case AtomicTextOperation::DELETESTRING:
				case AtomicTextOperation::CREATELINE:
				case AtomicTextOperation::DELETELINE:
						bChanged = true;
						break;
				default:
						break;
			}
		}

	/* 4. set the 'eStatus' flag for this operation */
	if (!bReturn)
		pTextOp->eStatus = ( bChanged ? TextOperation::TEXT_CHANGED : TextOperation::TEXT_UNCHANGED );

	/* 5. check the history for the saved state and refresh the qtab's label */
	if (!bReturn)
		pBuffer->setModified(isChanged());
}

// UndoRedo::doAtomicOperation(.....)
/**
 *  @brief Executes an atomic operation
 */
void UndoRedo::doAtomicOperation(AtomicTextOperation::OperationCode eCode, QString sString, int iColumn, int iLine, int iAttrib1, int iAttrib2)
{
	AtomicTextOperation *pAtomic = new AtomicTextOperation(eCode, sString, iColumn, iLine, iAttrib1, iAttrib2);

	/* if the list is not opened, add new operation */
	if (bOperationOpened == false)
	{
		dwarn("undoredo", "add atomic to a closed operation list!");
		createOperation("<NULL>");
	}

	if (ovPrevOperationLst.size() > 0)
		ovPrevOperationLst[ovPrevOperationLst.size()-1]->addAtomicOperation( pAtomic );

	/* execute the atomic operation */
	pAtomic->execDo(pBuffer);
}

// UndoRedo::execute_UndoRedoOperation()
/**
 *  @brief Undo or redo the last text operation
 *  @param bMicroUndo working in "microundo" mode or not; microundo steps only
 *         one operation at a time, "undo" executes of the operations from the same type
 */
void UndoRedo::execute_UndoRedoOperation(bool bUndo, bool bMicro)
{
	bool bFirst = true, bReturn = false;
	QString sFirstOpName, sCurOpName;
	TextOperation *pOperation = NULL;

	/* 1.a COND are we at the end of the undo list?
	       YES -> return */
	if (bUndo && (ovPrevOperationLst.size() == 0))
	{
		bReturn = true;
		dmesg("4:undoredo", "Unable to undo, we are already at the start of the history");
	} else
	/* 1.b COND are we at the end of the redo list?
	       YES -> return */
	if (!bUndo && (ovNextOperationLst.size() == 0))
	{
		bReturn = true;
		dmesg("4:undoredo", "Unable to redo, we are at the end of the history");
	} else

	/* 2. go while there is operation in the list
	      (we can stop earlier the iteration - further conditions are inside the block) */
	while (!bReturn && (bUndo ? ovPrevOperationLst.size() > 0 : ovNextOperationLst.size() > 0))
	{
		/* 2.1 get the last operation */
		if (bUndo)
			pOperation = *(ovPrevOperationLst.begin()+ovPrevOperationLst.size()-1);
		else pOperation = *(ovNextOperationLst.begin()+ovNextOperationLst.size()-1);

		/* 2.2 COND is this operation object NULL?
		        YES -> break (impossible error) */
		if (pOperation == NULL)
		{
			derr("1:undoredo", "impossible - NULL operation!");
			bReturn = true;
		}

		/* 2.3 COND is this the _first_ operation?
		        YES -> store its stripped type string */
		if (!bReturn && bFirst)
		{
			sFirstOpName = *(pOperation->sOperationName);
			stripOpName(sFirstOpName);
			bFirst = false;
		}

		/* 2.4 strip the name of sCurOpName (the actual operation)
		       to make it comparable with sFirstOpName (first processed operation) */
		if (!bReturn)
		{
			sCurOpName = *(pOperation->sOperationName);
			stripOpName(sCurOpName);
		}

		/* 2.5 COND (macro undo) is this an operation with a new type?
		   YES -> break (because we executed all of the operations with the same type) */
		if ((!bReturn) && (sFirstOpName != sCurOpName))
			bReturn = true;

		if (!bReturn)
		{
			/* 2.6 pop the last operation from the list*/
			if (bUndo)
			{
				ovPrevOperationLst.pop_back();
				ovNextOperationLst.push_back(pOperation);
			} else
			{
				ovNextOperationLst.pop_back();
				ovPrevOperationLst.push_back(pOperation);
			}

			/* 2.7 ITER go through and undo all of the atomic operations of this operation */
			for(int iIt = (bUndo ? pOperation->ovAtomicOps.size() - 1 : 0);
					(bUndo ? iIt >= 0 : iIt < (int)pOperation->ovAtomicOps.size()); (bUndo ? iIt--: iIt++))
			{
				/* 2.7.1 get and pop the last atomic operation */
				AtomicTextOperation *pAtomic = pOperation->ovAtomicOps[iIt];

				/* 2.7.2 undo the atomic operation */
				if (pAtomic != NULL)
				{
					if (bUndo)
						pAtomic->execUndo(pBuffer);
					else pAtomic->execDo(pBuffer);
				} else dwarn("2:undoredo", "Undo error - NULL atomic operation");
			}
		}

		/* 2.8 COND are we in "micro-undo" mode?
		   YES -> make end the loop (right after the first execution) */
		if (bMicro) bReturn = true;
	}

	/* 3. draw the result */
	pBuffer->cmd_scrollToCursor();
	pBuffer->repaint(false);

	/* 4. set/unset the 'modified' sign */
	pBuffer->setModified(isChanged());
}

// UndoRedo::stripOpName(.)
/**
 *  @brief Strips an operation type strings to get its main type
 *
 *  @example stripOpName("MOVE_CURSOR_(LEFT)") -> MOVE_CURSOR
 *  @param sOpName the string to strip
 */
void UndoRedo::stripOpName(QString &sOpName)
{
	/* 1. delete the string after a '(' character */
	int iTrunc = sOpName.find("(");
	if (iTrunc != -1) sOpName.truncate(iTrunc);

	/* 2. remove the ending '_' characters */
	while ((sOpName.length() >0) && (sOpName.at(sOpName.length()-1) == '_'))
		sOpName.truncate(sOpName.length() - 1);
}

// UndoRedo::setSavedPosition()
/**
 *  @brief Stores the position where the buffer was last saved
 */
void UndoRedo::setSavedPosition(void)
{
	/* create a marker position */
	createOperation("SAVE_POSITION");
	closeOperation();

	/* set the current position */
	if (ovPrevOperationLst.size() == 0) pSavedPosition = NULL;
	else pSavedPosition = *(ovPrevOperationLst.begin() + ovPrevOperationLst.size() - 1);

	if (pBuffer) pBuffer->setModified(false);
}

// UndoRedo::isAtSavedPosition()
/**
 *  @brief Examines if we are at the saved position
 *  @return true if we are at the saved position
 */
bool UndoRedo::isAtSavedPosition(void)
{
	TextOperation *pCurOper;

	/* 1. get the current position */
	if (ovPrevOperationLst.size() == 0) pCurOper = NULL;
	else pCurOper = *(ovPrevOperationLst.begin() + ovPrevOperationLst.size() - 1);

	/* 2. compare the current position with the saved position */
	return (pCurOper == pSavedPosition);
}

// UndoRedo::isChanged()
/**
 *  @brief Examines if the buffer's current state is saved
 *  @return true if it is changed after the last save
 */
bool UndoRedo::isChanged(void)
{
	bool bFOUND = false, bRETURN = false;
	bool bChanged;
	int iIdx;

	/* 1. ITER go back in the history to get the saved position */
	for(iIdx = ovPrevOperationLst.size()-1, bChanged = false; iIdx >= 0; iIdx --)
	{
		/* 1.1 COND if any of the text operation changed the text, then the file is unsaved */
		if (ovPrevOperationLst[iIdx]->eStatus == TextOperation::TEXT_CHANGED)
			bChanged = true;

		/* 1.2 COND we found the saved position or the file wasn't saved yet -> end the search*/
		if ((ovPrevOperationLst[iIdx] == pSavedPosition) ||
		    ((pSavedPosition == 0) && (iIdx == 0)))
		{
			bFOUND = true;
			break;
		}
	}

	/* 2. ITER if 1. failed then go forward in the history to get the saved position */
	if (!bFOUND)
	for(iIdx = ovNextOperationLst.size()-1, bChanged = false; iIdx >= 0; iIdx --)
	{
		/* 2.1 COND if any of the text operation changed the text, then the file is unsaved */
		if (ovNextOperationLst[iIdx]->eStatus == TextOperation::TEXT_CHANGED)
			bChanged = true;

		/* 2.2 COND we found the saved position -> end the search*/
		if (ovNextOperationLst[iIdx] == pSavedPosition)
		{
			bFOUND = true;
			break;
		}
	}

	/* 3. if we didn't find the saved state, report it (then the text is possibly changed) */
	if (!bFOUND)
	{
		derr("undoredo", "Unable to found last saved state - the text is possibly changed!");
		bRETURN = true;
	} else
	bRETURN = bChanged;

	/* 4. return with the result */
	return bRETURN;
}


